#if !defined(MATH3D_H)
/* ========================================================================
   Header File: d6math3d.h
   Description: Yet another cheesy 3-vector, 3x3-matrix library,
                transmogrified from dumb3d.h
   ======================================================================== */

/* ========================================================================
   Explicit Dependencies
   ======================================================================== */
#include <math.h>		// for sqrt
#include <float.h>      // for min/max
#include <stdlib.h>     // rand & RAND_MAX

// The default real type
#define REAL_IS_FLOAT  1
#define REAL_IS_DOUBLE 0
typedef float real;

real const REAL_MAX = FLT_MAX;
real const REAL_MIN = FLT_MIN;
real const RealPi = 3.14159265358979323846f;
real const Epsilon = 0.00001f;

/* Remove MacIntosh lameness */
#if defined(Length)
#undef Length
#endif

#pragma warning (disable:4244)		// int to float conversion

/*----------------------------------------------------------------------------

This header contains the declarations for the dumb3d functions.

*/

// forward declarations

class vector_3;
class matrix_3x3;

/*----------------------------------------------------------------------------

globally useful functions.

*/

inline real rabs( real x ) { return (real)fabs(x); }

inline void CreateOpenGLTransform( matrix_3x3 const &Orientation,
				vector_3 const &Position, float *pM);
inline void DecomposeOpenGLTransform( float const *M, matrix_3x3 &Rotation,
                                      vector_3 &Translation );

class ostream;
ostream& operator<<(ostream &outstream, vector_3 const &streamme);

bool RayIntersectsBoundingBox(vector_3 const &ray_origin,
							  vector_3 const &ray_direction,
							  vector_3 const &box_origin,
							  matrix_3x3 const &box_orientation,
							  real const width,
							  real const height,
							  real const depth,
							  real &t, vector_3 &normal);

real RayIntersectsPlane(vector_3 const &ray_origin,
						vector_3 const &ray_direction,
						vector_3 const &plane_normal,
						real const plane_d,
						real &vd);

inline vector_3 operator+( vector_3 const &Operand1,
		vector_3 const &Operand2 );

inline vector_3 operator-( vector_3 const &Operand1,
		vector_3 const &Operand2 );

inline vector_3 operator*( vector_3 const &Multiplicand,
		real const &Multiplier );

inline vector_3 operator*( real const &Multiplier,
		vector_3 const &Multiplicand );
	  
inline vector_3 operator/( vector_3 const &Quotient,
		real const &Divisor );

inline vector_3 operator-( vector_3 const &Operand1 );

inline vector_3 HadamardProduct( vector_3 const &Operand1,
		vector_3 const &Operand2 );

inline vector_3 CrossProduct( vector_3 const &Operand1,
		vector_3 const &Operand2 );

inline vector_3 Normalize( vector_3 const &A );
inline real Length( vector_3 const &A );
inline real Length2( vector_3 const &A );

matrix_3x3 Inverse( matrix_3x3 const &A );

inline vector_3 &ClipVectorMagnitude( vector_3 &A, real const MaximumLength );
inline real ClipRealMagnitude( real Value, real MaximumMagnitude );

void OrthonormalizeOrientation( matrix_3x3 &Orientation );

inline real DotProduct( vector_3 const &Operand1, vector_3 const &Operand2 );

real AngleBetween( vector_3 const &Operand1, vector_3 const &Operand2 );
real AngleBetweenUnit( vector_3 const &Unit1, vector_3 const &Unit2 );

inline real RadiansFrom( real Degrees );
inline real DegreesFrom( real Radians );

void GetDextralCoordinateSystem( vector_3 const &InputX,
				vector_3 &OutputY, vector_3 &OutputZ );

inline matrix_3x3 Transpose( matrix_3x3 const &Matrix );

matrix_3x3 operator*( matrix_3x3 const &Multiplicand,
		matrix_3x3 const &Multiplier );

vector_3 operator*( matrix_3x3 const &Multiplicand,
		vector_3 const &Multiplier );

inline bool operator==( vector_3 const &A, vector_3 const &B);

inline matrix_3x3 operator-( matrix_3x3 const &Matrix );
inline matrix_3x3 operator+( matrix_3x3 const &Operand1,
		matrix_3x3 const &Operand2 );
inline matrix_3x3 operator-( matrix_3x3 const &Operand1,
		matrix_3x3 const &Operand2 );
inline matrix_3x3 operator*( real const Value, matrix_3x3 const &Matrix );
inline matrix_3x3 operator*( matrix_3x3 const &Matrix, real const Value );

inline real Square( real a ) { return a*a; }
inline real Cube( real a ) { return Square(a)*a; }

bool Inverse( matrix_3x3 const &A, matrix_3x3 &AInverse );

inline real ContinuousMaxMagnitude( real v, real v_max );
inline vector_3 ContinuousMaxMagnitude( vector_3 const &v, real v_max );


/*----------------------------------------------------------------------------

triple.	Base class for vectors and points.

*/

class triple
{
public:

	inline real &operator()( int unsigned Row );
	inline real const &operator()( int unsigned Row ) const;

	inline real GetElement( int Row ) const;

	inline real GetX( void ) const;
	inline real GetY( void ) const;
	inline real GetZ( void ) const;

	inline void SetElement( int Row, real Value );

	inline void SetX( real Value );
	inline void SetY( real Value );
	inline void SetZ( real Value );

	inline real *GetElementsPointer(void) {return(aElements);}
	inline real *GetElements(void) {return(aElements);}
	
protected:

	inline triple( void );
	inline triple( real X, real Y, real Z );
	inline triple( triple const & );
	inline triple &operator=( triple const & );

	real aElements[3];
};


/*----------------------------------------------------------------------------

vector_3.  This class represents a 3D vector.

*/

class vector_3 :
	public triple
{
public:

	inline vector_3( void );
	inline vector_3( real const *pElements );
	inline vector_3( real X, real Y, real Z );

	inline vector_3 &operator+=(vector_3 const &addend);
	inline vector_3 &operator-=(vector_3 const &lessor);
	inline vector_3 &operator*=(real const multiplier);
	inline vector_3 &operator/=(real const divisor);

	inline bool IsZero( void ) const;
	vector_3 &Normalize( void );
	real Length( void ) const;
};



/*----------------------------------------------------------------------------

matrix_3x3.  This class represents row major 3x3 homogeneous matrices.

*/

class matrix_3x3
{
public:

	matrix_3x3( void );
	matrix_3x3( real (*pElements)[3] );
	matrix_3x3( real const pElements[9] );

	// @todo this isn't the safest ctor to have
	enum mode { SkewSymmetric, Zero, Rows, Columns, Diagonal, OuterProduct };
	inline matrix_3x3( mode MustBeZero );
	inline matrix_3x3( vector_3 const &CrossVector, mode MustBeSkewSymmetric );
	inline matrix_3x3( vector_3 const &a, vector_3 const &b,
					   vector_3 const &c, mode MustBeRowsOrColumns );
	inline matrix_3x3( real const d, mode MustBeDiagonal );
	inline matrix_3x3( vector_3 const &a, vector_3 const &b,
						mode MustBeOuterProduct );
	
	inline matrix_3x3( vector_3 const &Axis, real AngleInRadians );
	inline matrix_3x3( vector_3 const &XAxis, vector_3 const &YAxis,
					   vector_3 const &ZAxis);
	inline matrix_3x3( matrix_3x3 const &Rotation );

	matrix_3x3 &ConcatenateXRotation( real Radians );
	matrix_3x3 &ConcatenateYRotation( real Radians );
	matrix_3x3 &ConcatenateZRotation( real Radians );

	matrix_3x3 &ConcatenateXScale( real Distance );
	matrix_3x3 &ConcatenateYScale( real Distance );
	matrix_3x3 &ConcatenateZScale( real Distance );
	inline matrix_3x3 &ConcatenateScale( vector_3 &Scale );

	inline matrix_3x3 &operator+=( matrix_3x3 const &A );
	inline matrix_3x3 &operator*=( matrix_3x3 const &A ) { *this = *this * A; return *this; }
	inline matrix_3x3 &operator-=( matrix_3x3 const &A );

	inline real &operator()( int unsigned Row, int unsigned Column );
	inline real const &operator()( int unsigned Row, int unsigned Column ) const;

	inline real GetElement( int Row, int Column ) const;
	inline matrix_3x3 &SetElement( int Row, int Column, real Value );

	// @todo we need a row vector class
	inline vector_3 GetRow( int unsigned Row ) const;
	inline vector_3 GetColumn( int unsigned Column ) const;

	inline real const *GetElementsPointer( void ) const { return aElements[0]; }

	inline bool IsIdentity( void ) const;
	inline bool IsZero( void ) const;

protected:

	enum do_not_initialize { DoNotInitialize };

	inline matrix_3x3( do_not_initialize );

	void InitWith( vector_3 const &Axis, real AngleInRadians );
	void InitWith( vector_3 const &XAxis, vector_3 const &YAxis,
				   vector_3 const &ZAxis );

	real aElements[3][3];
};

/*----------------------------------------------------------------------------

inline function definitions.

*/

inline real RadiansFrom( real Degrees )
{
	return (Degrees / real(360)) * RealPi * real(2);
}

inline real DegreesFrom( real Radians )
{
	return real(360) * (Radians / (RealPi * real(2)));
}

inline vector_3 &vector_3::operator+=(vector_3 const &addend)
{
	SetX(GetX() + addend.GetX());
	SetY(GetY() + addend.GetY());
	SetZ(GetZ() + addend.GetZ());

	return(*this);
}

inline vector_3 operator+( vector_3 const &Operand1,
		vector_3 const &Operand2 )
{
	return vector_3(Operand1.GetX() + Operand2.GetX(),
					Operand1.GetY() + Operand2.GetY(),
					Operand1.GetZ() + Operand2.GetZ());
}

inline vector_3 &vector_3::operator-=(vector_3 const &lessor)
{
	SetX(GetX() - lessor.GetX());
	SetY(GetY() - lessor.GetY());
	SetZ(GetZ() - lessor.GetZ());

	return(*this);
}

inline vector_3 operator-( vector_3 const &Operand1,
		vector_3 const &Operand2 )
{
	return vector_3(Operand1.GetX() - Operand2.GetX(),
					Operand1.GetY() - Operand2.GetY(),
					Operand1.GetZ() - Operand2.GetZ());
}

inline vector_3 operator-( vector_3 const &Operand1 )
{
	return vector_3(-Operand1.GetX(),-Operand1.GetY(),-Operand1.GetZ());
}

inline vector_3 &vector_3::operator*=(real const multiplier)
{
	SetX(GetX() * multiplier);
	SetY(GetY() * multiplier);
	SetZ(GetZ() * multiplier);

	return(*this);
}

inline vector_3 operator*( vector_3 const &Multiplicand,
		real const &Multiplier )
{
	return vector_3(Multiplicand.GetX() * Multiplier,
					Multiplicand.GetY() * Multiplier,
					Multiplicand.GetZ() * Multiplier);
}

inline vector_3 operator*( real const &Multiplier,
		vector_3 const &Multiplicand ) 
{
	return Multiplicand * Multiplier;
}
	  
inline vector_3 &vector_3::operator/=(real const divisor)
{
	SetX(GetX() / divisor);
	SetY(GetY() / divisor);
	SetZ(GetZ() / divisor);

	return(*this);
}

inline vector_3 operator/( vector_3 const &Quotient,
		real const &Divisor )
{
	return Quotient * (real(1.0)/Divisor);
}

inline vector_3 CrossProduct( vector_3 const &Operand1,
		vector_3 const &Operand2 )
{
	real X = Operand1.GetY() * Operand2.GetZ() -
				Operand1.GetZ() * Operand2.GetY();
	real Y = Operand1.GetZ() * Operand2.GetX() -
				Operand1.GetX() * Operand2.GetZ();
	real Z = Operand1.GetX() * Operand2.GetY() -
				Operand1.GetY() * Operand2.GetX();

	return vector_3(X,Y,Z);
}

inline vector_3 HadamardProduct( vector_3 const &Operand1,
		vector_3 const &Operand2 )
{
	return vector_3(Operand1.GetX() * Operand2.GetX(),
					Operand1.GetY() * Operand2.GetY(),
					Operand1.GetZ() * Operand2.GetZ());
}

inline real DotProduct( vector_3 const &Operand1, vector_3 const &Operand2 )
{
	return Operand1.GetX() * Operand2.GetX() +
			Operand1.GetY() * Operand2.GetY() +
			Operand1.GetZ() * Operand2.GetZ();
}

inline real triple::GetElement( int Row ) const
{
	return aElements[Row];
}

inline real triple::GetX( void ) const
{
	return aElements[0];
}

inline real triple::GetY( void ) const
{
	return aElements[1];
}

inline real triple::GetZ( void ) const
{
	return aElements[2];
}

inline real &triple::operator()( int unsigned Row )
{
	assert(Row < 3);
	return aElements[Row];
}

inline real const &triple::operator()( int unsigned Row ) const
{
	assert(Row < 3);
	return aElements[Row];
}

inline void triple::SetElement( int Row, real Value )
{
	aElements[Row] = Value;
}

inline void triple::SetX( real Value )
{
	aElements[0] = Value;
}

inline void triple::SetY( real Value )
{
	aElements[1] = Value;
}

inline void triple::SetZ( real Value )
{
	aElements[2] = Value;
}

inline triple::triple( void )
{
	aElements[0] = aElements[1] = aElements[2] = 0;
}

inline triple::triple( real X, real Y, real Z )
{
	aElements[0] = X;
	aElements[1] = Y;
	aElements[2] = Z;
}

inline triple::triple( triple const &Source )
{
	aElements[0] = Source.aElements[0];
	aElements[1] = Source.aElements[1];
	aElements[2] = Source.aElements[2];
}

inline triple &triple::operator=( triple const &Source )
{
	aElements[0] = Source.aElements[0];
	aElements[1] = Source.aElements[1];
	aElements[2] = Source.aElements[2];

	return *this;
}

inline vector_3::vector_3( void ) :
	triple(0,0,0)
{
}

inline vector_3::vector_3( real X, real Y, real Z ) :
	triple(X,Y,Z)
{
}

inline vector_3::vector_3( real const *pElements )
{
	assert(pElements);
	aElements[0] = pElements[0];
	aElements[1] = pElements[1];
	aElements[2] = pElements[2];
}

inline real &matrix_3x3::operator()( int unsigned Row, int unsigned Column )
{
	assert((Row < 3) && (Column < 3));
	return aElements[Row][Column];
}

inline real const &matrix_3x3::operator()( int unsigned Row,
		int unsigned Column ) const
{
	assert((Row < 3) && (Column < 3));
	return aElements[Row][Column];
}

inline real matrix_3x3::GetElement( int Row, int Column ) const
{
	return aElements[Row][Column];
}

inline matrix_3x3 &matrix_3x3::SetElement( int Row, int Column, real Value )
{
	aElements[Row][Column] = Value;

	return *this;
}

inline matrix_3x3::matrix_3x3( do_not_initialize )
{
}

inline matrix_3x3::matrix_3x3( vector_3 const &Axis, real AngleInRadians )
{
	InitWith(Axis,AngleInRadians);
}

inline matrix_3x3::matrix_3x3( real (*pElements)[3] )
{
	assert(pElements);

	for(int i = 0;i < 3;i++)
	{
		for(int j = 0;j < 3;j++)
		{
			aElements[i][j] = pElements[i][j];
		}
	}
}
inline matrix_3x3::matrix_3x3( real const pElements[9] )
{
	assert(pElements);

	for(int i = 0;i < 9;i++)
	{
		aElements[0][i] = pElements[i];
	}
}

inline matrix_3x3::matrix_3x3( vector_3 const &Vector, mode Mode )
{
	assert(Mode == SkewSymmetric);

	matrix_3x3 &M = *this;

	M(0,0) =  0;             M(0,1) = -Vector.GetZ(); M(0,2) =  Vector.GetY();
	M(1,0) =  Vector.GetZ(); M(1,1) =  0;             M(1,2) = -Vector.GetX();
	M(2,0) = -Vector.GetY(); M(2,1) =  Vector.GetX(); M(2,2) = 0;
}

inline matrix_3x3::matrix_3x3( vector_3 const &a, vector_3 const &b, mode Mode )
{
	assert(Mode == OuterProduct);

	matrix_3x3 &M = *this;

	for(int unsigned i = 0;i < 3;i++)
	{
		M(i,0) = a(i)*b(0);
		M(i,1) = a(i)*b(1);
		M(i,2) = a(i)*b(2);
	}
}

inline matrix_3x3::matrix_3x3( mode Mode )
{
	assert(Mode == Zero);

	for(int Counter = 0;Counter < 3;Counter++)
	{
		aElements[Counter][0] = 0;
		aElements[Counter][1] = 0;
		aElements[Counter][2] = 0;
	}
}

inline matrix_3x3::matrix_3x3( real const d, mode Mode )
{
	assert(Mode == Diagonal);

	for(int Counter = 0;Counter < 3;Counter++)
	{
		aElements[Counter][0] = 0;
		aElements[Counter][1] = 0;
		aElements[Counter][2] = 0;
	}
	aElements[0][0] = aElements[1][1] = aElements[2][2] = d;
}

inline matrix_3x3::matrix_3x3(vector_3 const &XAxis, vector_3 const &YAxis,
							  vector_3 const &ZAxis )
{
	InitWith(XAxis,YAxis,ZAxis);
}

inline matrix_3x3::matrix_3x3(vector_3 const &a, vector_3 const &b,
							  vector_3 const &c, mode Mode )
{
	assert((Mode == Rows) || (Mode == Columns));

	if(Mode == Columns)
	{
		for(int i = 0;i < 3;i++)
		{
			(*this)(i,0) = a(i);
			(*this)(i,1) = b(i);
			(*this)(i,2) = c(i);
		}
	}
	else
	{
		for(int j = 0;j < 3;j++)
		{
			(*this)(0,j) = a(j);
			(*this)(1,j) = b(j);
			(*this)(2,j) = c(j);
		}
	}
}

inline matrix_3x3::matrix_3x3( matrix_3x3 const &Rotation )
{
	// put in rotation

	for(int Counter = 0;Counter < 3;Counter++)
	{
		SetElement(0,Counter,Rotation.GetElement(0,Counter));
		SetElement(1,Counter,Rotation.GetElement(1,Counter));
		SetElement(2,Counter,Rotation.GetElement(2,Counter));
	}
}

inline matrix_3x3 &matrix_3x3::ConcatenateScale( vector_3 &Scale )
{
	ConcatenateXScale(Scale.GetX());
	ConcatenateYScale(Scale.GetY());
	ConcatenateZScale(Scale.GetZ());

	return *this;
}

inline matrix_3x3 Transpose( matrix_3x3 const &Matrix )
{
	matrix_3x3 Return;

	for(int Counter = 0;Counter < 3;Counter++)
	{
		Return(0,Counter) = Matrix(Counter,0);
		Return(1,Counter) = Matrix(Counter,1);
		Return(2,Counter) = Matrix(Counter,2);
	}

	return Return;
}	

inline matrix_3x3 operator+( matrix_3x3 const &Operand1,
		matrix_3x3 const &Operand2 )
{
	matrix_3x3 Return;

	for(int Counter = 0;Counter < 3;Counter++)
	{
		Return(0,Counter) = Operand1(0,Counter) + Operand2(0,Counter);
		Return(1,Counter) = Operand1(1,Counter) + Operand2(1,Counter);
		Return(2,Counter) = Operand1(2,Counter) + Operand2(2,Counter);
	}

	return Return;
}	
	
inline matrix_3x3 operator-( matrix_3x3 const &Operand1,
		matrix_3x3 const &Operand2 )
{
	matrix_3x3 Return;

	for(int Counter = 0;Counter < 3;Counter++)
	{
		Return(0,Counter) = Operand1(0,Counter) - Operand2(0,Counter);
		Return(1,Counter) = Operand1(1,Counter) - Operand2(1,Counter);
		Return(2,Counter) = Operand1(2,Counter) - Operand2(2,Counter);
	}

	return Return;
}	
	
inline matrix_3x3 operator*( real const Value, matrix_3x3 const &Matrix )
{
	matrix_3x3 Return;

	for(int Counter = 0;Counter < 3;Counter++)
	{
		Return(0,Counter) = Value * Matrix(0,Counter);
		Return(1,Counter) = Value * Matrix(1,Counter);
		Return(2,Counter) = Value * Matrix(2,Counter);
	}

	return Return;
}	

inline matrix_3x3 &matrix_3x3::operator+=( matrix_3x3 const &Operand1 )
{
	for(int Counter = 0;Counter < 3;Counter++)
	{
		// @todo this is probably the wrong order, but I'm lazy
		aElements[0][Counter] += Operand1(0,Counter);
		aElements[1][Counter] += Operand1(1,Counter);
		aElements[2][Counter] += Operand1(2,Counter);
	}

	return *this;
}	

inline matrix_3x3 &matrix_3x3::operator-=( matrix_3x3 const &Operand1 )
{
	for(int Counter = 0;Counter < 3;Counter++)
	{
		// @todo this is probably the wrong order, but I'm lazy
		aElements[0][Counter] -= Operand1(0,Counter);
		aElements[1][Counter] -= Operand1(1,Counter);
		aElements[2][Counter] -= Operand1(2,Counter);
	}

	return *this;
}	

inline matrix_3x3 operator*( matrix_3x3 const &Matrix, real const Value )
{
	return Value * Matrix;
}

inline matrix_3x3 operator-( matrix_3x3 const &Matrix )
{
	matrix_3x3 Return;

	for(int Counter = 0;Counter < 3;Counter++)
	{
		Return(0,Counter) = -Matrix(0,Counter);
		Return(1,Counter) = -Matrix(1,Counter);
		Return(2,Counter) = -Matrix(2,Counter);
	}

	return Return;
}	

inline bool operator==( vector_3 const &A, vector_3 const &B)
{
	return((A.GetX() == B.GetX()) &&
		   (A.GetY() == B.GetY()) &&
		   (A.GetZ() == B.GetZ()));
}

inline vector_3 matrix_3x3::GetRow( int unsigned Row ) const
{
	assert(Row < 3);
	return vector_3((*this)(Row,0),(*this)(Row,1),(*this)(Row,2));
}

inline vector_3 matrix_3x3::GetColumn( int unsigned Column ) const
{
	assert(Column < 3);
	return vector_3((*this)(0,Column),(*this)(1,Column),(*this)(2,Column));
}

inline vector_3 Normalize( vector_3 const &A )
{
	vector_3 Return = A;
	return Return.Normalize();
}

inline real Length( vector_3 const &A )
{
	return A.Length();
}

inline real Length2( vector_3 const &A )
{
	return DotProduct(A,A);
}

inline vector_3 &ClipVectorMagnitude( vector_3 &A, real const MaximumLength )
{
	if(A.Length() > MaximumLength)
	{
		A.Normalize() *= MaximumLength;
	}
	return A;
}
inline real ClipRealMagnitude( real Value, real MaximumMagnitude )
{
	assert(MaximumMagnitude > 0);
	if(rabs(Value) > MaximumMagnitude)
	{
		Value = (Value > 0) ? MaximumMagnitude : -MaximumMagnitude;
	}
	return Value;
}

inline bool vector_3::IsZero( void ) const
{
	return (aElements[0] == 0) &&
			(aElements[1] == 0) &&
			(aElements[2] == 0);
}

inline void CreateOpenGLTransform( matrix_3x3 const &Orientation,
				vector_3 const &Position, float *pM)
{
	assert(pM);
	float (*pM44)[4] = (float(*)[4])pM;
	int i;
	
	// remember, OpenGL matrices are transposed!

	for(i = 0;i < 3;i++)
	{
		for(int j = 0;j < 3;j++)
		{
			pM44[j][i] = Orientation(i,j);
		}
	}

	pM44[0][3] = pM44[1][3] = pM44[2][3] = 0.0f; pM44[3][3] = 1.0f;

	for(i = 0;i < 3;i++)
	{
		pM44[3][i] = Position(i);
	}
}

inline void DecomposeOpenGLTransform( float const *pM, matrix_3x3 &Rotation,
                                      vector_3 &Translation )
{
    // @todo what about scale, projection, etc?
	assert(pM);
	float (*pM44)[4] = (float(*)[4])pM;
	int i;
	
	// remember, OpenGL matrices are transposed!

	for(i = 0;i < 3;i++)
	{
		for(int j = 0;j < 3;j++)
		{
			Rotation(i,j) = pM44[j][i];
		}
	}
	for(i = 0;i < 3;i++)
	{
		Translation(i) = pM44[3][i];
	}
}

bool matrix_3x3::IsIdentity( void ) const
{
	real const (*p)[3] = aElements;
	
	return  (p[0][0]==1) && (p[0][1]==0) && (p[0][2]==0) &&
			(p[1][0]==0) && (p[1][1]==1) && (p[1][2]==0) &&
			(p[2][0]==0) && (p[2][1]==0) && (p[2][2]==1);
}
	
bool matrix_3x3::IsZero( void ) const
{
	real const (*p)[3] = aElements;

	return  (p[0][0]==0) && (p[0][1]==0) && (p[0][2]==0) &&
			(p[1][0]==0) && (p[1][1]==0) && (p[1][2]==0) &&
			(p[2][0]==0) && (p[2][1]==0) && (p[2][2]==0);
}

real ContinuousMaxMagnitude( real v, real v_max )
{
    return v_max * (real)tanh(v/v_max);
}

vector_3 ContinuousMaxMagnitude( vector_3 const &v, real v_max )
{
    real v_l = v.Length();
    real v_m = 1.f;
    if(v_l > Epsilon)
    {
        v_m = ContinuousMaxMagnitude(v_l,v_max) / v_l;
    }
    return v_m*v;
}

#pragma warning (disable:4514)		// unreferenced inline removed

#define MATH3D_H
#endif
