/* ========================================================================
   $RCSfile: bmp.cpp,v $
   $Date: 2002/03/19 02:10:42 $
   $Revision: 1.8.12.1 $
   $Creator: Casey Muratori $
   $Notice: $
   ======================================================================== */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#if !defined(BMP_H)
#include "bmp.h"
#endif

#if !defined(LOG_H)
#include "log.h"
#endif

#if !defined(MEMORY_UTILS_H)
#include "memory_utils.h"
#endif

#if !defined(IMAGE_OPERATIONS_H)
#include "image_operations.h"
#endif

typedef unsigned char BYTE;
typedef unsigned short WORD;
typedef unsigned int DWORD;
typedef int LONG;

#pragma pack(1)
typedef struct tagRGBQUAD {
  BYTE    rgbBlue; 
  BYTE    rgbGreen; 
  BYTE    rgbRed; 
  BYTE    rgbAlpha; 
} RGBQUAD;

typedef struct tagBITMAPFILEHEADER {
        WORD    bfType;
        DWORD   bfSize;
        WORD    bfReserved1;
        WORD    bfReserved2;
        DWORD   bfOffBits;
} BITMAPFILEHEADER;

typedef struct tagBITMAPINFOHEADER{
        DWORD      biSize;
        LONG       biWidth;
        LONG       biHeight;
        WORD       biPlanes;
        WORD       biBitCount;
        DWORD      biCompression;
        DWORD      biSizeImage;
        LONG       biXPelsPerMeter;
        LONG       biYPelsPerMeter;
        DWORD      biClrUsed;
        DWORD      biClrImportant;
} BITMAPINFOHEADER;

typedef struct tagBITMAPINFO {
    BITMAPINFOHEADER    bmiHeader;
    RGBQUAD             bmiColors[1];
} BITMAPINFO;

static void
UnpackPalette(int32 Width, int32 Height, int32 TransparentIndex,
              RGBQUAD *ColorTable, void const *SourcePixelsInit,
              void *DestPixelsInit, int ColorCount)
{
    uint8 const *SourcePixels = (uint8 const *)SourcePixelsInit;
    RGBQUAD *DestPixels = (RGBQUAD *)DestPixelsInit;

    int32 Offset = ((Width + 0x3) & ~0x3) - Width;
    
    int32 HeightCounter = Height;
    while(HeightCounter--)
    {
        int32 WidthCounter = Width;
        while(WidthCounter--)
        {
            int32 ColorIndex = *SourcePixels++;
            RGBQUAD Color;
            if(ColorIndex == TransparentIndex)
            {
                Color.rgbRed = Color.rgbGreen = Color.rgbBlue = Color.rgbAlpha = 0x00;
            }
            else
            {
				assert(ColorIndex < ColorCount);	// xxxxx
				Color = ColorTable[ColorIndex];
                Color.rgbAlpha = 0xFF;
            }
            *DestPixels++ = Color;
        }

        SourcePixels += Offset;
    }
}

bool
ReadExactly(FILE *File, int32 Size, void *Buffer)
{
    return(fread(Buffer, Size, 1, File) == 1);
}

void
Handle8BitBitmap(bmp *BMP, int32 SourceStride, void *SourcePixels,
                 BITMAPINFOHEADER &InfoHeader,
                 FILE *File, int32 ColorTablePosition,
                 int32 TransparentIndex)
{
    int32 ColorCount = InfoHeader.biClrUsed;
    if(ColorCount == 0)
    {
        ColorCount = 256;
    }
    fseek(File, ColorTablePosition, SEEK_SET);

    RGBQUAD *ColorTable = (RGBQUAD *)
        malloc(ColorCount*sizeof(RGBQUAD));
    assert(ColorTable);
    if(ColorTable)
    {                            
        ReadExactly(File, ColorCount * sizeof(RGBQUAD), ColorTable);
                                
        UnpackPalette(BMP->Width, BMP->Height, TransparentIndex,
                      ColorTable, SourcePixels, BMP->Pixels, ColorCount);

        free(ColorTable);
    }
}

void
Handle24BitBitmap(bmp *BMP, int32 SourceStride, void *SourcePixels)
{
    ConvertPixelFormat(BMP->Width, BMP->Height,
                       BGR888PixelFormat,
                       SourceStride, SourcePixels,
                       BGRA8888PixelFormat,
                       BMP->Stride, BMP->Pixels);
//    assert(!"Alpha isn't set properly on 24-bit bitmaps!");
}

void
Handle32BitBitmap(bmp *BMP, int32 SourceStride, void *SourcePixels)
{
    CopyStrided(BMP->Width * 4, BMP->Height,
                SourceStride, SourcePixels,
                BMP->Stride, BMP->Pixels);
}

bmp *ReadBMP_FILE( void *stdio_FILE_pointer, int32 TransparentIndex, char const *FileNameLog )
{
    FILE *File = (FILE *)stdio_FILE_pointer;  
    assert(File);
    bmp *BMP;
    BITMAPFILEHEADER Header;
    BITMAPINFOHEADER InfoHeader;
    if(ReadExactly(File, sizeof(Header), &Header) &&
       ReadExactly(File, sizeof(InfoHeader), &InfoHeader))
    {
        int32 Width = InfoHeader.biWidth;
        int32 Height = InfoHeader.biHeight;

        assert((Width > 0) && (Height > 0));

        int32 ColorTablePosition = ftell(File);

        BMP = (bmp *)malloc(sizeof(*BMP) +
                            sizeof(uint32)*Width*Height);
        if(BMP)
        {
            BMP->Pixels = (uint32 *)(BMP + 1);
            BMP->Width = Width;
            BMP->Height = Height;
            BMP->Stride = BMP->Width * 4;

            // @todo doesn't handle top-down negative width/height?
            int32 SourceStride = Width * (InfoHeader.biBitCount / 8);
            SourceStride = (SourceStride + 0x3) & (~0x3);
                    
            int32 BMPSize = (Height * SourceStride);
            void *SourcePixels = malloc(BMPSize);
            assert(SourcePixels);
            if(SourcePixels)
            {
                fseek(File, Header.bfOffBits, SEEK_SET);
                if(ReadExactly(File, BMPSize, SourcePixels))
                {
                    switch(InfoHeader.biBitCount)
                    {
                        case 8:
                        {
                            Handle8BitBitmap(
                                BMP, SourceStride, SourcePixels,
                                InfoHeader,
                                File, ColorTablePosition,
                                TransparentIndex);
                        } break;

                        case 24:
                        {
                            Handle24BitBitmap(
                                BMP, SourceStride, SourcePixels);
                        } break;

                        case 32:
                        {
                            Handle32BitBitmap(
                                BMP, SourceStride, SourcePixels);
                        } break;
                    }
                }
                else
                {
                    Log("Unable to read BMP header from \"%s\".\n", FileNameLog);
                }

                free(SourcePixels);
            }                
            else
            {
                Log("Unable to allocate source pixel buffer "
                    "for bitmap \"%s\".\n", FileNameLog);
            }
        }
        else
        {
            Log("Unable to allocate memory for bitmap \"%s\".\n",
                FileNameLog);
        }
    }
    else
    {
        Log("Unable to read BMP header from \"%s\".\n", FileNameLog);
    }
        
    return(BMP);
}
bmp *ReadBMP(char const *FileName, int32 TransparentIndex)
{
    bmp *BMP = 0;

    FILE *File = fopen(FileName, "rb");
    if(File)
    {
        BMP = ReadBMP_FILE(File,TransparentIndex,FileName);
        fclose(File);
    }
    else
    {
        Log("Unable to open BMP file \"%s\".\n", FileName);
    }
    return BMP;
}
void
    FreeBMP(bmp *BMP)
{
    if(BMP)
    {
        free(BMP);
    }
}

void
WriteBMP(bmp &BMP, char const *FileName)
{
    FILE *File = fopen(FileName, "wb");
    if(File)
    {
        int DestStride = BMP.Height * 3;
        int SizeOfBits = BMP.Width * DestStride;
        
        BITMAPFILEHEADER Header;
        BITMAPINFO Info;
        
        Header.bfType = ('M' << 8) | 'B';
        Header.bfSize =
            sizeof(Header) + sizeof(Info) + SizeOfBits;
        Header.bfReserved1 = 0;
        Header.bfReserved2 = 0;
        Header.bfOffBits = sizeof(Header) + sizeof(Info);
        
        Info.bmiHeader.biSize = sizeof(Info.bmiHeader);
        Info.bmiHeader.biWidth = BMP.Width;
        Info.bmiHeader.biHeight = BMP.Height;
        Info.bmiHeader.biPlanes = 1;
        Info.bmiHeader.biBitCount = 24;
        Info.bmiHeader.biCompression = 0;
        Info.bmiHeader.biSizeImage = SizeOfBits;
        Info.bmiHeader.biXPelsPerMeter = 0;
        Info.bmiHeader.biYPelsPerMeter = 0;
        Info.bmiHeader.biClrUsed = 0;
        Info.bmiHeader.biClrImportant = 0;
        
        BYTE *Bits = (BYTE *)malloc(SizeOfBits);
        assert(Bits);
        ConvertPixelFormat(BMP.Width, BMP.Height,
                           BGRA8888PixelFormat,
                           BMP.Stride, BMP.Pixels,
                           BGR888PixelFormat,
                           DestStride, Bits);
        
        fwrite(&Header, sizeof(Header), 1, File);
        fwrite(&Info, sizeof(Info), 1, File);
        fwrite(Bits, SizeOfBits, 1, File);
        
        free(Bits);
        fclose(File);
    }
}
