/* ========================================================================
   $RCSfile: camera.cpp,v $
   $Date: 2002/03/17 22:53:11 $
   $Revision: 1.5.12.3 $
   $Creator: Casey Muratori $
   $Notice: $
   ======================================================================== */

#include <assert.h>
#include <math.h>

#if !defined(CAMERA_H)
#include "camera.h"
#endif

typedef real32 triple[3];
typedef real32 quad[4];
real32 const GlobalWAxis[4] = {0, 0, 0, 1};

inline void VectorZero3(real32 *Dest)
{
    Dest[0] = Dest[1] = Dest[2] = 0.0f;
}

inline void VectorEquals3(real32 *Dest, real32 const *Source)
{
    Dest[0] = Source[0];
    Dest[1] = Source[1];
    Dest[2] = Source[2];
}

inline void VectorEquals4(real32 *Dest, real32 const *Source)
{
    Dest[0] = Source[0];
    Dest[1] = Source[1];
    Dest[2] = Source[2];
    Dest[3] = Source[3];
}

inline void ScaleVectorAdd3(real32 *Dest, real32 const C,
                            real32 const *Addend)
{
    Dest[0] += C * Addend[0];
    Dest[1] += C * Addend[1];
    Dest[2] += C * Addend[2];
}

inline real32
RadiansFromDegrees(real32 Degrees)
{
    return(Degrees * Pi32 / 180.0f);
}

inline void
VectorScale3(real32 *Dest, real32 const Scalar)
{
    Dest[0] *= Scalar;
    Dest[1] *= Scalar;
    Dest[2] *= Scalar;
}

inline void
VectorAdd3(real32 *Dest, real32 const *Addend)
{
    Dest[0] += Addend[0];
    Dest[1] += Addend[1];
    Dest[2] += Addend[2];
}

inline void
VectorNegate3(real32 *Dest)
{
    Dest[0] = -Dest[0];
    Dest[1] = -Dest[1];
    Dest[2] = -Dest[2];
}

inline void
TransposeVectorTransform3(real32 *Dest, real32 const *TransformInit)
{
    triple const *Transform = (triple const *)TransformInit;
    
    real32 const D0 = Dest[0];
    real32 const D1 = Dest[1];
    real32 const D2 = Dest[2];

    Dest[0] = D0*Transform[0][0] + D1*Transform[1][0] + D2*Transform[2][0];
    Dest[1] = D0*Transform[0][1] + D1*Transform[1][1] + D2*Transform[2][1];
    Dest[2] = D0*Transform[0][2] + D1*Transform[1][2] + D2*Transform[2][2];
}

inline void
TransposeVectorTransform3(real32 *Dest,
                          real32 const *TransformInit,
                          real32 const *Source)
{
    triple const *Transform = (triple const *)TransformInit;

    Dest[0] = (Source[0]*Transform[0][0] +
               Source[1]*Transform[1][0] +
               Source[2]*Transform[2][0]);
    Dest[1] = (Source[0]*Transform[0][1] +
               Source[1]*Transform[1][1] +
               Source[2]*Transform[2][1]);
    Dest[2] = (Source[0]*Transform[0][2] +
               Source[1]*Transform[1][2] +
               Source[2]*Transform[2][2]);
}

inline void
VectorTransform3(real32 *Dest,
                 real32 const *TransformInit,
                 real32 const *Source)
{
    triple const *Transform = (triple const *)TransformInit;

    Dest[0] = (Source[0] * Transform[0][0] +
               Source[1] * Transform[0][1] +
               Source[2] * Transform[0][2]);
    Dest[1] = (Source[0] * Transform[1][0] +
               Source[1] * Transform[1][1] +
               Source[2] * Transform[1][2]);
    Dest[2] = (Source[0] * Transform[2][0] +
               Source[1] * Transform[2][1] +
               Source[2] * Transform[2][2]);
}

inline void
TransposeVectorTransform4(real32 *Dest, real32 const *Transform)
{
    real32 const D0 = Dest[0];
    real32 const D1 = Dest[1];
    real32 const D2 = Dest[2];
    real32 const D3 = Dest[3];

    Dest[0] = D0*Transform[0] + D1*Transform[4] + D2*Transform[8] + D3*Transform[12];
    Dest[1] = D0*Transform[1] + D1*Transform[5] + D2*Transform[9] + D3*Transform[13];
    Dest[2] = D0*Transform[2] + D1*Transform[6] + D2*Transform[10] + D3*Transform[14];
    Dest[3] = D0*Transform[3] + D1*Transform[7] + D2*Transform[11] + D3*Transform[15];
}

inline void
VectorSubtract3(real32 *Dest, real32 const *Addend)
{
    Dest[0] -= Addend[0];
    Dest[1] -= Addend[1];
    Dest[2] -= Addend[2];
}

inline void
VectorSubtract3(real32 *Dest, real32 const *Vector0,
                real32 const *Vector1)
{
    Dest[0] = Vector0[0] - Vector1[0];
    Dest[1] = Vector0[1] - Vector1[1];
    Dest[2] = Vector0[2] - Vector1[2];
}

inline real32
VectorLengthSquared3(real32 const *Source)
{
    return(Source[0] * Source[0] +
           Source[1] * Source[1] +
           Source[2] * Source[2]);
}

inline real32
VectorLength3(real32 const *Source)
{
    return(sqrt(VectorLengthSquared3(Source)));
}

inline void
Normalize3(real32 *Dest)
{
    VectorScale3(Dest, 1.0f / VectorLength3(Dest));
}

inline void
XRotationColumns3x3(triple *Result, real32 const AngleInRadians)
{
    real32 const AngleSin = sin(AngleInRadians);
    real32 const AngleCos = cos(AngleInRadians);

    Result[0][0] = 1.0f;
    Result[0][1] = 0.0f;
    Result[0][2] = 0.0f;
    Result[1][0] = 0.0f;
    Result[1][1] = AngleCos;
    Result[1][2] = -AngleSin;
    Result[2][0] = 0.0f;
    Result[2][1] = AngleSin;
    Result[2][2] = AngleCos;
}

inline void
YRotationColumns3x3(triple *Result, real32 const AngleInRadians)
{
    real32 const AngleSin = sin(AngleInRadians);
    real32 const AngleCos = cos(AngleInRadians);

    Result[0][0] = AngleCos;
    Result[0][1] = 0.0f;
    Result[0][2] = AngleSin;
    Result[1][0] = 0.0f;
    Result[1][1] = 1.0f;
    Result[1][2] = 0.0f;
    Result[2][0] = -AngleSin;
    Result[2][1] = 0.0f;
    Result[2][2] = AngleCos;
}

inline void
ZRotationColumns3x3(triple *Result, real32 const AngleInRadians)
{
    real32 const AngleSin = sin(AngleInRadians);
    real32 const AngleCos = cos(AngleInRadians);
    
    Result[0][0] = AngleCos;
    Result[0][1] = -AngleSin;
    Result[0][2] = 0.0f;
    Result[1][0] = AngleSin;
    Result[1][1] = AngleCos;
    Result[1][2] = 0.0f;
    Result[2][0] = 0.0f;
    Result[2][1] = 0.0f;
    Result[2][2] = 1.0f;
}

void
MatrixEqualsQuaternion3x3(real32 *Dest, real32 const *Operand)
{
    // 3 multiplies
    // Squares (for the diagonal)
    real32 const XSquared = Operand[0] * Operand[0];
    real32 const YSquared = Operand[1] * Operand[1];
    real32 const ZSquared = Operand[2] * Operand[2];

    // 6 multiplies
    // Combinations (for the non-diagonal)
    real32 const XY = Operand[0] * Operand[1];
    real32 const XZ = Operand[0] * Operand[2];
    real32 const YZ = Operand[1] * Operand[2];
    real32 const WX = Operand[3] * Operand[0];
    real32 const WY = Operand[3] * Operand[1];
    real32 const WZ = Operand[3] * Operand[2];

    // 9 multiplies, 12 adds    
    // Top row
    Dest[0] = 1 - 2*(YSquared + ZSquared);
    Dest[1] = 2*(XY - WZ);
    Dest[2] = 2*(XZ + WY);

    // Middle row
    Dest[3] = 2*(XY + WZ);
    Dest[4] = 1 - 2*(XSquared + ZSquared);
    Dest[5] = 2*(YZ - WX);

    // Bottom row
    Dest[6] = 2*(XZ - WY);
    Dest[7] = 2*(YZ + WX);
    Dest[8] = 1 - 2*(XSquared + YSquared);

    // Total: 18 multiplies, 12 adds
}

void
MatrixMultiply3x3(real32 *IntoMatrix,
                  real32 const *Matrix,
                  real32 const *ByMatrix)
{
    assert(Matrix != IntoMatrix);
    assert(ByMatrix != IntoMatrix);

    triple *Result = (triple *)IntoMatrix;
    triple const *A = (triple const *)Matrix;
    triple const *B = (triple const *)ByMatrix;

    Result[0][0] = A[0][0]*B[0][0] + A[0][1]*B[1][0] + A[0][2]*B[2][0];
    Result[0][1] = A[0][0]*B[0][1] + A[0][1]*B[1][1] + A[0][2]*B[2][1];
    Result[0][2] = A[0][0]*B[0][2] + A[0][1]*B[1][2] + A[0][2]*B[2][2];

    Result[1][0] = A[1][0]*B[0][0] + A[1][1]*B[1][0] + A[1][2]*B[2][0];
    Result[1][1] = A[1][0]*B[0][1] + A[1][1]*B[1][1] + A[1][2]*B[2][1];
    Result[1][2] = A[1][0]*B[0][2] + A[1][1]*B[1][2] + A[1][2]*B[2][2];

    Result[2][0] = A[2][0]*B[0][0] + A[2][1]*B[1][0] + A[2][2]*B[2][0];
    Result[2][1] = A[2][0]*B[0][1] + A[2][1]*B[1][1] + A[2][2]*B[2][1];
    Result[2][2] = A[2][0]*B[0][2] + A[2][1]*B[1][2] + A[2][2]*B[2][2];
}

void
InitializeDefaultCamera(camera &Camera)
{
    Camera.WpOverHp = StandardTVPhysicalAspectRatio;
    Camera.WrOverHr = StandardTVPhysicalAspectRatio;
    Camera.WwOverHw = 1.0f;
    
    Camera.FOVY = RadiansFromDegrees(60.0f);
    Camera.NearClipPlane = 1.0f;
    Camera.FarClipPlane = 1000.0f;

    VectorZero3(Camera.Position);
    VectorEquals4(Camera.Orientation, GlobalWAxis);
    VectorZero3(Camera.EAR);
    VectorZero3(Camera.Offset);

    BuildCameraMatrices(Camera);
}

void
SetCameraAspectRatios(camera &Camera,
                      real32 PhysicalAspectRatio,
                      real32 ScreenWidth, real32 ScreenHeight,
                      real32 WindowWidth, real32 WindowHeight)
{
    Camera.WpOverHp = PhysicalAspectRatio;
    Camera.WrOverHr = ScreenWidth / ScreenHeight;
    Camera.WwOverHw = WindowWidth / WindowHeight;
}

void
MoveCameraRelative(camera &Camera, real32 X, real32 Y, real32 Z)
{
    ScaleVectorAdd3(Camera.Position, X, Camera.InverseView4x4[0]);
    ScaleVectorAdd3(Camera.Position, Y, Camera.InverseView4x4[1]);
    ScaleVectorAdd3(Camera.Position, Z, Camera.InverseView4x4[2]);
}

static void
BuildProjectionMatrix(camera &Camera, real32 *Matrix, real32 *InverseMatrix)
{
    real32 TanHalfFOVY = tan(0.5f * Camera.FOVY);

    Matrix[0] = (Camera.WrOverHr /
                 (TanHalfFOVY * Camera.WwOverHw * Camera.WpOverHp));
    Matrix[1] = 0.0f;
    Matrix[2] = 0.0f;
    Matrix[3] = 0.0f;

    Matrix[4] = 0.0f;
    Matrix[5] = 1.0f / (TanHalfFOVY);
    Matrix[6] = 0.0f;
    Matrix[7] = 0.0f;

    Matrix[8] = 0.0f;
    Matrix[9] = 0.0f;
    Matrix[10] = ((Camera.NearClipPlane + Camera.FarClipPlane) /
                  (Camera.NearClipPlane - Camera.FarClipPlane));
    Matrix[11] = -1.0f;

    Matrix[12] = 0.0f;
    Matrix[13] = 0.0f;
    Matrix[14] = ((2.0f * Camera.FarClipPlane * Camera.NearClipPlane) /
                  (Camera.NearClipPlane - Camera.FarClipPlane));
    Matrix[15] = 0.0f;

    InverseMatrix[0] = 1.0f / Matrix[0];
    InverseMatrix[1] = 0.0f;
    InverseMatrix[2] = 0.0f;
    InverseMatrix[3] = 0.0f;

    InverseMatrix[4] = 0.0f;
    InverseMatrix[5] = 1.0f / Matrix[5];
    InverseMatrix[6] = 0.0f;
    InverseMatrix[7] = 0.0f;

    InverseMatrix[8] = 0.0f;
    InverseMatrix[9] = 0.0f;
    InverseMatrix[10] = 0.0f;
    InverseMatrix[11] = 1.0f / Matrix[14];

    InverseMatrix[12] = 0.0f;
    InverseMatrix[13] = 0.0f;
    InverseMatrix[14] = 1.0f / Matrix[11];
    InverseMatrix[15] = -(Matrix[10] / (Matrix[14] * Matrix[11]));
}


vec3f	viewpoint_position;
vec3f	view_direction;

static void
BuildViewMatrix(camera &Camera, real32 *Matrix, real32 *InverseMatrix)
{
    triple PostEARAxes[3];
    triple Axes[3];
    triple EARRotation[3];
    triple Temp[3];

    MatrixEqualsQuaternion3x3((real32 *)Axes, Camera.Orientation);
    
    YRotationColumns3x3(EARRotation, Camera.EAR[1]);
    MatrixMultiply3x3((real32 *)PostEARAxes,
                      (real32 *)Axes,
                      (real32 *)EARRotation);
 
    XRotationColumns3x3(EARRotation, Camera.EAR[0]);
    MatrixMultiply3x3((real32 *)Temp, (real32 *)PostEARAxes,
                      (real32 *)EARRotation);
 
    ZRotationColumns3x3(EARRotation, Camera.EAR[2]);
    MatrixMultiply3x3((real32 *)PostEARAxes, (real32 *)Temp,
                      (real32 *)EARRotation);
    
    Matrix[0] = PostEARAxes[0][0];
    Matrix[1] = PostEARAxes[0][1];
    Matrix[2] = PostEARAxes[0][2];
    Matrix[3] = 0.0f;

    Matrix[4] = PostEARAxes[1][0];
    Matrix[5] = PostEARAxes[1][1];
    Matrix[6] = PostEARAxes[1][2];
    Matrix[7] = 0.0f;

    Matrix[8] = PostEARAxes[2][0];
    Matrix[9] = PostEARAxes[2][1];
    Matrix[10] = PostEARAxes[2][2];
    Matrix[11] = 0.0f;

    real32 InversePosition[3];
    TransposeVectorTransform3(InversePosition, (real32 const *)PostEARAxes,
                              Camera.Position);
    VectorAdd3(InversePosition, Camera.Offset);
    VectorNegate3(InversePosition);
    
    Matrix[12] = InversePosition[0];
    Matrix[13] = InversePosition[1];
    Matrix[14] = InversePosition[2];
    Matrix[15] = 1.0f;

    InverseMatrix[0] = Matrix[0];
    InverseMatrix[1] = Matrix[4];
    InverseMatrix[2] = Matrix[8];
    InverseMatrix[3] = 0.0f;
    
    InverseMatrix[4] = Matrix[1];
    InverseMatrix[5] = Matrix[5];
    InverseMatrix[6] = Matrix[9];
    InverseMatrix[7] = 0.0f;

    InverseMatrix[8] = Matrix[2];
    InverseMatrix[9] = Matrix[6];
    InverseMatrix[10] = Matrix[10];
    InverseMatrix[11] = 0.0f;

    real32 Position[3];
    VectorTransform3(Position, (real32 const *)PostEARAxes, Camera.Offset);
    VectorAdd3(Position, Camera.Position);
    
    InverseMatrix[12] = Position[0];
    InverseMatrix[13] = Position[1];
    InverseMatrix[14] = Position[2];
    InverseMatrix[15] = 1.0f;

	// extern the camera pos.
	viewpoint_position.x = Position[0];
	viewpoint_position.y = Position[1];
	viewpoint_position.z = Position[2];

   	view_direction.x = -InverseMatrix[8];
    view_direction.y = -InverseMatrix[9];
    view_direction.z = -InverseMatrix[10];
}
    
void
BuildCameraMatrices(camera &Camera)
{
    BuildViewMatrix(Camera,
                    (real32 *)Camera.View4x4,
                    (real32 *)Camera.InverseView4x4);
    BuildProjectionMatrix(Camera,
                          (real32 *)Camera.Projection4x4,
                          (real32 *)Camera.InverseProjection4x4);
}

void
GetCameraLocation(camera &Camera, real32 *Result)
{
    VectorEquals3(Result, Camera.InverseView4x4[3]);
}

void
WindowSpaceToWorldSpace(camera &Camera, real32 Width, real32 Height,
                        real32 const *ScreenSpacePoint, real32 *Result)
{
    real32 const ZMin = -1.0f;
    real32 const ZMax = 1.0f;

    quad Homogenous;
    Homogenous[0] = (2.0f*ScreenSpacePoint[0] - Width) / Width;
    Homogenous[1] = (2.0f*ScreenSpacePoint[1] - Height) / Height;
    Homogenous[2] = (2.0f*ScreenSpacePoint[2] - ZMax - ZMin) / (ZMax - ZMin);
    Homogenous[3] = 1.0f;

    TransposeVectorTransform4(Homogenous,
                              (real32 const *)Camera.InverseProjection4x4);
    TransposeVectorTransform4(Homogenous,
                              (real32 const *)Camera.InverseView4x4);

    real32 InverseW = 1.0f / Homogenous[3];
    Result[0] = Homogenous[0] * InverseW;
    Result[1] = Homogenous[1] * InverseW;
    Result[2] = Homogenous[2] * InverseW;
}

void
GetPickingRay(camera &Camera, real32 Width, real32 Height,
              real32 MouseX, real32 MouseY,
              real32 *Origin, real32 *Normal)
{
    GetCameraLocation(Camera, Origin);
    
    real32 ScreenPoint[3] = {MouseX, MouseY, -1.0f};
    real32 WorldPoint[3];
    WindowSpaceToWorldSpace(Camera, Width, Height,
                            ScreenPoint, WorldPoint);

    VectorSubtract3(Normal, WorldPoint, Origin);
    Normalize3(Normal);
}

inline void
VectorCrossProduct3(real32 *Dest, real32 const *Operand1,
                    real32 const *Operand2)
{
    real32 const X = Operand1[1]*Operand2[2] - Operand1[2]*Operand2[1];
    real32 const Y = Operand1[2]*Operand2[0] - Operand1[0]*Operand2[2];
    real32 const Z = Operand1[0]*Operand2[1] - Operand1[1]*Operand2[0];
        
    Dest[0] = X;
    Dest[1] = Y;
    Dest[2] = Z;
}

inline real32
InnerProduct3(real32 const *V0, real32 const *V1)
{
    return(V0[0] * V1[0] +
           V0[1] * V1[1] +
           V0[2] * V1[2]);
}

void 
QuaternionMultiply4(real32 *Dest, real32 const *Operand1,
                    real32 const *Operand2)
{
    real32 const w1 = Operand1[3];
    real32 const w2 = Operand2[3];

    real32 Cross[3];
    VectorCrossProduct3(Cross, Operand1, Operand2);
    ScaleVectorAdd3(Cross, w1, Operand2);
    ScaleVectorAdd3(Cross, w2, Operand1);

    real32 Dot = InnerProduct3(Operand1, Operand2);

    Dest[0] = Cross[0];
    Dest[1] = Cross[1];
    Dest[2] = Cross[2];
    Dest[3] = w1*w2 - Dot;
}

void
MouseZoomCamera(camera &Camera, float dZ)
{
    Camera.Offset[2] -= dZ * 0.005f * Camera.Offset[2];
}

void
MouseOrbitCamera(camera &Camera, float dX, float dY)
{
    Camera.EAR[1] += dX * 3.14f / 180.0f;
    Camera.EAR[0] -= dY * 3.14f / 180.0f;
}

void
MouseRotateCameraRelative(camera &Camera, float dX, float dY)
{
    real32 HalfAngleX = 0.5f*(real32)dX * 3.14f / 180.0f;
    real32 HalfAngleY = 0.5f*(real32)dY * 3.14f / 180.0f;
    
    real32 XRotation[4] = {sin(HalfAngleY), 0, 0, cos(HalfAngleY)};
    QuaternionMultiply4(Camera.Orientation,
                        Camera.Orientation,
                        XRotation);

    real32 YRotation[4] = {0, sin(HalfAngleX), 0, cos(HalfAngleX)};
    QuaternionMultiply4(Camera.Orientation,
                        Camera.Orientation,
                        YRotation);
}

void
MouseMoveCameraRelative(camera &Camera,
                        float dX, float dY, float dZ)
{
    MoveCameraRelative(Camera,
                       -dX * 0.001f * Camera.Offset[2],
                       -dY * 0.001f * Camera.Offset[2],
                       0.0f);
}

void
MouseMoveCameraPlaneRelative(camera &Camera,
                             float dXPlane, float dYPlane)
{
    real32 const UpVector[3] = {0, 0, 1};

    real32 const *PlaneXVector = Camera.InverseView4x4[0];

    real32 PlaneYVector[3];
    VectorCrossProduct3(PlaneYVector, UpVector, PlaneXVector);
    Normalize3(PlaneYVector);

    ScaleVectorAdd3(Camera.Position, dXPlane, PlaneXVector);
    ScaleVectorAdd3(Camera.Position, dYPlane, PlaneYVector);
}

void
MouseMoveCameraAbsolute(camera &Camera,
                        float dX, float dY, float dZ)
{
}
