/* ========================================================================
   $RCSfile: font.cpp,v $
   $Date: 2002/03/16 21:46:22 $
   $Revision: 1.3.16.1 $
   $Creator: Casey Muratori $
   $Notice: $
   ======================================================================== */

#include <assert.h>
#include <stdio.h>
#include <windows.h>
#include <gl/gl.h>

#if !defined(FONT_H)
#include "font.h"
#endif

#if !defined(LOG_H)
#include "log.h"
#endif

struct font
{
    char const *FontFileName;
    GLuint *GLTextureHandles;

    font_header Header;
};

static bool
LoadFontImages(font &Font, FILE *File)
{
    bool Result = true;
    
    int const TextureCount = Font.Header.TextureCount;
    int const TextureWidth = Font.Header.TextureWidth;
    int const TextureHeight = Font.Header.TextureHeight;
        
    Font.GLTextureHandles =
        (GLuint *)malloc(TextureCount*sizeof(GLuint));
    assert(Font.GLTextureHandles);
    glGenTextures(TextureCount, Font.GLTextureHandles);

    int unsigned TextureSize = (sizeof(int unsigned) *
                                TextureWidth * TextureHeight);
    void *TextureBuffer = malloc(TextureSize);
    {for(int TextureIndex = 0;
         TextureIndex < TextureCount;
         ++TextureIndex)
    {
        if(fread(TextureBuffer, TextureSize, 1, File) == 1)
        {
            glBindTexture(GL_TEXTURE_2D, Font.GLTextureHandles[TextureIndex]);
            glTexImage2D(GL_TEXTURE_2D, 0, 4,
                         TextureWidth, TextureHeight, 0, GL_RGBA,
                         GL_UNSIGNED_BYTE,
                         TextureBuffer);
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
            glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
        }
        else
        {
            Result = false;
            Log("Unable to read font data");
        }
    }}
    free(TextureBuffer);

    return(Result);
}

font *
LoadFont(char const *FontFileName)
{
    font *NewFont = (font *)malloc(sizeof(font));
    assert(NewFont);
    NewFont->FontFileName = FontFileName;

    FILE *File = fopen(FontFileName, "rb");
    if(File)
    {
        if(fread(&NewFont->Header, sizeof(NewFont->Header), 1, File))
        {
            LoadFontImages(*NewFont, File);
        }
        else
        {
            Log("Unable to read font header");
        }

        fclose(File);
    }
    else
    {
        NewFont->GLTextureHandles = 0;
        Log("Unable to open font file");
    }

    return(NewFont);
}

bool
ReloadFont(font &Font)
{
    bool Result = false;
    
    FILE *File = fopen(Font.FontFileName, "rb");
    if(File)
    {
        if(fseek(File, sizeof(Font.Header), SEEK_SET))
        {
            Result = LoadFontImages(Font, File);
        }
        else
        {
            Log("Unable to seek to font image data");
        }

        fclose(File);
    }
    else
    {
        Log("Unable to open font file");
    }

    return(Result);
}

void
FreeFont(font *Font)
{
    if(Font)
    {
        glDeleteTextures(Font->Header.TextureCount,
                         Font->GLTextureHandles);
        
        free(Font->GLTextureHandles);
        free(Font);
    }
}

int
GetFontTextureCount(font &Font)
{
    return(Font.Header.TextureCount);
}

void
RenderFontTexture(font &Font, int Index)
{
    assert(Index < GetFontTextureCount(Font));

    glBindTexture(GL_TEXTURE_2D, Font.GLTextureHandles[Index]);
    glBegin(GL_QUADS);
    glTexCoord2f(0.0f, 0.0f);
    glVertex2i(0, 0);
    glTexCoord2f(1.0f, 0.0f);
    glVertex2i(Font.Header.TextureWidth, 0);
    glTexCoord2f(1.0f, 1.0f);
    glVertex2i(Font.Header.TextureWidth, Font.Header.TextureHeight);
    glTexCoord2f(0.0f, 1.0f);
    glVertex2i(0, Font.Header.TextureHeight);
    glEnd();
}

void
DrawString(font &Font, char const *String)
{
    if(String && Font.GLTextureHandles)
    {
        int CurrentTextureIndex =
            Font.Header.Characters[*String].TextureIndex;
        glBindTexture(GL_TEXTURE_2D,
                      Font.GLTextureHandles[CurrentTextureIndex]);
        glBegin(GL_QUADS);

        int XPos = 0;
		int YPos = 0;
        while(*String)
        {
			// process newline
			if (*String == '\n') {
				XPos = 0;
				YPos -= Font.Header.Height;
				String++;
				continue;
			}

            font_character const &Character =
                Font.Header.Characters[*String];
            int ThisTextureIndex = Character.TextureIndex;
            if(CurrentTextureIndex != ThisTextureIndex)
            {
                glEnd();
                CurrentTextureIndex = ThisTextureIndex;
                glBindTexture(GL_TEXTURE_2D,
                              Font.GLTextureHandles[CurrentTextureIndex]);
                glBegin(GL_QUADS);
            }
            
            glTexCoord2f(Character.XMin, Character.YMin);
            glVertex2i(XPos, YPos);
            
            glTexCoord2f(Character.XMax, Character.YMin);
            glVertex2i(XPos + Character.Width, YPos);
            
            glTexCoord2f(Character.XMax, Character.YMax);
            glVertex2i(XPos + Character.Width, YPos + Character.Height);
            
            glTexCoord2f(Character.XMin, Character.YMax);
            glVertex2i(XPos, YPos + Character.Height);

            XPos += Character.Kerning[*(String + 1)];
            ++String;
        }

        glEnd();
    }
}

void
GetDimensions(font &Font, char const *String, int &Width, int &Height)
{
    if(Font.GLTextureHandles)
    {
        Width = 0;
        while(*String)
        {
            Width += Font.Header.Characters[*String].Kerning[*(String + 1)];
            ++String;
        }
    
        Height = Font.Header.Height;
    }
}

void
GetDimensions(font &Font, int CharCount, char const *String, int &Width, int &Height)
{
    if(Font.GLTextureHandles)
    {
        Width = 0;
        while(CharCount--)
        {
            Width += Font.Header.Characters[*String].Kerning[*(String + 1)];
            ++String;
        }
        
        Height = Font.Header.Height;
    }
}
