#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <stdlib.h>
#include <math.h>
#include <float.h>
#include <assert.h>
#include <stdio.h>

#include <gl/gl.h>
#include <gl/glu.h>


#include "statistics.h"
#include "app_camera.h"
#include "globals.h"
#include "guys.h"
#include "buildings.h"
#include "hf.h"

#include "animation.h"
#include "sprite_texture.h"

#include "bmp.h"
#include "prefetch.h"

// global hooks for the app
float (*gameGuyHitTerrain)(Guy *g);
void (*gameGuyAITimer)(Guy *g);

//#ifdef _DEBUG
#define DEBUG_COLOR
//#endif

const float M_PI = 3.141592653589; // @todo

#define BIN_X     (1 << LOG_BIN_X)
#define LOG_BIN_X 6
#define BIN_Y     (1 << LOG_BIN_Y)
#define LOG_BIN_Y 6
#define BIN_SIZE  (WORLD_X_POWER_OF_2 / BIN_X)    // size of a bin in worldspace, MUST BE POWER OF TWO

#if ((BIN_SIZE-1) & BIN_SIZE) != 0
#error "bad bin size"
#endif

//////////////////////////////////////////////////////////////////////////////////////////////

#define BIN(x,y)  ((y)*BIN_X + (x))

//#define COMPUTE_BIN(g)   BIN((int) ((g)->loc.x / BIN_SIZE), (int) ((g)->loc.y / BIN_SIZE))
#define COMPUTE_BIN(g)   BIN(FLOAT2INT_BIN((g)->loc.x), FLOAT2INT2_BIN((g)->loc.y))

#define BIN_NULL  0xffff
#define DEFAULT_BIN  16    // number of guys per bin

// Global Guy array and number of guys
Guy *gGuys;
int gGuyCount=0;

#define NUM_BINS  (BIN_X*BIN_Y)

struct Bin
{
   int num_guys_in_bin;
   int max_guys;
   bool mark;
   Guy **guys;
} bin[NUM_BINS];

static uint32 bin_active[BIN_Y];

void clearFindRegions(void)
{
   memset(bin_active, 0, sizeof(bin_active));
}

#define MAX_BIN_POOL 256
static Bin bin_pool[MAX_BIN_POOL];
static int num_pool;

static void binClearAll(void)
{
}

void binAdd(int bin_id, Guy *guy)
{
   assert((uint) bin_id < BIN_X*BIN_Y);
   if ((uint) bin_id >= BIN_X*BIN_Y) {
      GUYBIN(guy) = BIN_NULL;
      return;
   }

   Bin *b = &bin[bin_id];
   if (b->num_guys_in_bin >= b->max_guys) {
      if (b->max_guys == 0) {
         if (num_pool) {
            --num_pool;
            b->guys = bin_pool[num_pool].guys;
            b->max_guys = bin_pool[num_pool].max_guys;
         } else {
            b->max_guys = DEFAULT_BIN;
            b->guys = new Guy* [b->max_guys];
         }
      } else {
         // @TODO: realloc would be nice
         Guy **old = b->guys;
         b->max_guys *= 2;
         b->guys = new Guy* [b->max_guys];
         memcpy(b->guys, old, sizeof(b->guys[0]) * b->num_guys_in_bin);
         delete [] old;
       }
   }
   int o = b->num_guys_in_bin++;
   b->guys[o] = guy;
   GUYBIN(guy) = bin_id;
   guy->bin_offset = o;
}

void binRemove(int bin_id, Guy *guy)
{
   if (bin_id == BIN_NULL) return;
   assert(bin_id == GUYBIN(guy));
   Bin *b = &bin[bin_id];
   int o = guy->bin_offset;
   assert(b->guys[o] == guy);
   int o2 = --b->num_guys_in_bin; // offset of guy we're going to move
   if (o2 != o) {
      // move o2 into o
      b->guys[o] = b->guys[o2];
      b->guys[o]->bin_offset = o;
   }
   if (o2 >= (DEFAULT_BIN-1) && o2 < (b->max_guys >> 2)) {
      Guy **old = b->guys;
      b->max_guys >>= 2;
      b->guys = new Guy * [b->max_guys];
      memcpy(b->guys, old, sizeof(b->guys[0]) * b->num_guys_in_bin);
      delete [] old;
   }
   GUYBIN(guy) = BIN_NULL;
}

void deleteGuy(int g)
{
   if(gGuys[g].ref_guy) {
       *gGuys[g].ref_guy = 0;   // kill the guy as far as the client is concerned
   }
   binRemove(gGuys[g].bin, &gGuys[g]);
   if (g != gGuyCount - 1) {
      gGuys[g] = gGuys[gGuyCount-1];
      if(gGuys[g].ref_guy) {
          *gGuys[g].ref_guy = gGuys+g;   // update the client's ref to the new location
      }
      if (gGuys[g].bin != BIN_NULL)
         bin[gGuys[g].bin].guys[gGuys[g].bin_offset] = &gGuys[g];
      gGuys[gGuyCount-1].bin = BIN_NULL;
   }
   --gGuyCount;
}

void rebinGuy(int g)
{
   binRemove(gGuys[g].bin, &gGuys[g]);
   binAdd(COMPUTE_BIN(&gGuys[g]), &gGuys[g]);
}

void binGuy(int g)
{
   binAdd(COMPUTE_BIN(&gGuys[g]), &gGuys[g]);
}

void setNumGuys(int num)
{
   if (num < gGuyCount) {
      while (--gGuyCount >= num) {
         binRemove(gGuys[gGuyCount].bin, &gGuys[gGuyCount]);
      }
   } else {
      while (gGuyCount < num) {
         gGuys[gGuyCount].bin = COMPUTE_BIN(&gGuys[gGuyCount]);
         binAdd(gGuys[gGuyCount].bin, &gGuys[gGuyCount]);
         ++gGuyCount;
      }
   }
}

#define MAX_FLOCK 512

static struct flock_info_t
{
   vec3f vel;
   vec3f loc;
   vec3f repel;
   int num_neighbor;
   Guy *g;
} flock_info[MAX_FLOCK];
static int num_flock;

void flockGuyCollisionResponse(Guy *g, Guy *h)
{
   int n = g->num_flock;
   if (n == 0) {
      if (num_flock == MAX_FLOCK-1)
         return;
      g->num_flock = n = ++num_flock;
      flock_info[n].vel.x = 0;
      flock_info[n].vel.y = 0;
      flock_info[n].vel.z = 0;
      flock_info[n].loc.x = 0;
      flock_info[n].loc.y = 0;
      flock_info[n].loc.z = 0;
      flock_info[n].repel.x = 0;
      flock_info[n].repel.y = 0;
      flock_info[n].repel.z = 0;
      flock_info[n].num_neighbor = 0;
      flock_info[n].g = g;
   }
   flock_info[n].vel.x += h->vel.x;
   flock_info[n].vel.y += h->vel.y;
   flock_info[n].vel.z += h->vel.z;
   flock_info[n].loc.x += h->loc.x;
   flock_info[n].loc.y += h->loc.y;
   flock_info[n].loc.z += h->loc.z;
   vec3f repel;
   repel.x = g->loc.x - h->loc.x;
   repel.y = g->loc.y - h->loc.y;
   repel.z = g->loc.z - g->loc.z;
   float r2_i = 1.0 / (repel.x*repel.x + repel.y*repel.y + repel.z*repel.z);
   flock_info[n].repel.x += r2_i * repel.x;
   flock_info[n].repel.y += r2_i * repel.y;
   flock_info[n].repel.z += r2_i * repel.z;
   ++flock_info[n].num_neighbor;
}

//  we want to have the following properties:
//     the "aim along other guys" has a fixed weight... or weighted by distance?
//     the "move closer" outweighs the above if far enough
//     the "repulsion" outweighs the above if close enough
//     the "aim for target goal" has a fixed weight
//     the "aim along current trajectory" has a fixed weight

//  as sampled,
//    along_other_guys is weighted by speed
//    move_closer is weighted by distance to center
//    repulsion is weighted by 1/distance
//    aim_target is weighted by speed
//    current_trajectory is weighted by speed

// so we divide the speed-weighted ones by speed
// divide move_closer by some distance-sampling-based constant
// multiply repulsion by some distance-sampling-based constant?
// at distance = 1 with no weight, repulsion weight is 1, regardless of actual distance sampled
// so, no, don't change repulsion weighting based on flocking sample distance
// so just multiply move_closer and repulsion by speed

//float fw_follow=1, fw_repel=3, fw_center = 0.5, fw_current = 3;
float fw_follow=1, fw_repel=1, fw_center = 0.5, fw_current = 1;  // @todo
//float fw_follow=0, fw_repel=5, fw_center = 0, fw_current = 3;

void aimGuyAlong(Guy *g);
void flockPostBin(void)
{
   for (int i=1; i <= num_flock; ++i) {
      Guy *g = flock_info[i].g;
      float s = g->speed * GUY_MAX_SPEED / 256.0;
      float r = 1.0 / flock_info[i].num_neighbor;
      float w_follow = fw_follow * r;
      float w_repel = fw_repel * s;

      // compute the three desired velocities
      vec3f center;
      center.x = flock_info[i].loc.x*r - g->loc.x;
      center.y = flock_info[i].loc.y*r - g->loc.y;

      float w_center  = fw_center * s / sqrt(center.x*center.x+center.y*center.y);

      g->vel.x = center.x*w_center + flock_info[i].vel.x*w_follow + flock_info[i].repel.x*w_repel + g->vel.x * fw_current;
      g->vel.y = center.y*w_center + flock_info[i].vel.y*w_follow + flock_info[i].repel.y*w_repel + g->vel.y * fw_current;

      aimGuyAlong(g);

      g->num_flock = 0;
   }
   num_flock = 0;
}

void test_guy_against_bin(Collision_Manager *cm, Bin *bin, Guy *guy, float radius_squared, uint8 guy_collide_flags) {
    int i;
    for (i = 0; i < bin->num_guys_in_bin; i++) {
        Guy *other_guy = bin->guys[i];

        if ((other_guy->be_collided_with_flags & guy_collide_flags) == 0) continue;
        if (guy == other_guy) continue;

        float dx, dy, dz;
        dx = guy->loc.x - other_guy->loc.x;
        dy = guy->loc.y - other_guy->loc.y;
        dz = guy->loc.z - other_guy->loc.z;

        float dist2 = dx * dx + dy * dy + dz * dz;
        if (dist2 <= radius_squared) {
            cm->hit_func(guy, other_guy);
        }
    }
    if (cm->end_func) cm->end_func();
}

// reshuffle the bin_id's bits to try to randomize the sampling pattern
//  

static int reverseBits(int x, int n, int m)
{
   int y=0;
   while (n--) {
      y <<= 1;
      if (n < m) y <<= 1;
      if (x & 1) y |= 1;
      x >>= 1;

   }
   return y;
}

static int shuffleX(int x)
{
   return reverseBits(x, LOG_BIN_X, LOG_BIN_Y);
}

static int shuffleY(int x)
{
   return reverseBits(x, LOG_BIN_Y, LOG_BIN_X) << 1;
}

static int shuffle_x[BIN_X], shuffle_y[BIN_Y];

static int bin_shuffle(int x)
{
   return shuffle_x[x & (BIN_X-1)] + shuffle_y[x >> LOG_BIN_X];
}

int test[BIN_Y*BIN_X];
static void shuffle_init(void)
{
   int x;
   for (x=0; x < BIN_X; ++x)
      shuffle_x[x] = shuffleX(x);
   for (x=0; x < BIN_Y; ++x)
      shuffle_y[x] = shuffleY(x);
   for (x=0; x < BIN_X*BIN_Y; ++x) {
      test[x] = -1;
   }
   for (x=0; x < BIN_X*BIN_Y; ++x) {
      int y = bin_shuffle(x);
      assert(y >= 0 && y < BIN_X*BIN_Y);
      assert(test[y] == -1);
      test[y] = x;
   }
}

static void guysFindCollisions(Collision_Manager *manager, float dt)
{
    float radius_squared = manager->collision_radius * manager->collision_radius;

    const int num_bins_total = BIN_X * BIN_Y;

    float fraction = manager->fraction_of_bins_to_process_per_second * dt;

    assert(fraction >= 0);
    if (fraction > 1) fraction = 1;

    if (fraction == 0) return;

    int num_bins_to_process = fraction * num_bins_total;
    if (num_bins_to_process == 0) num_bins_to_process = 1;

    int bin0 = manager->next_bin_index_to_process;
    int bin1 = manager->next_bin_index_to_process + num_bins_to_process;
    if (bin1 >= num_bins_total) bin1 -= num_bins_total;
    assert(bin1 < num_bins_total);

    int bin_index = bin0;
    while (1) {
        int bin_id = bin_shuffle(bin_index);
        Bin *current_bin = &bin[bin_id];
        int num_guys = current_bin->num_guys_in_bin;
        int i;
        // test whether any of the guys in the bin might collide with anyone
        // in the bin
        uint8 collide_source = 0;
        uint8 collide_dest = 0;
        for (i=0; i < num_guys; ++i) {
            collide_dest   |= current_bin->guys[i]->be_collided_with_flags;
            collide_source |= current_bin->guys[i]->collide_flags;
        }
        collide_dest &= manager->flags;
        if (collide_dest & collide_source) {
            for (i = 0; i < num_guys; i++) {
                Guy *guy = current_bin->guys[i];
                if (guy->collide_flags & collide_dest) {
                    test_guy_against_bin(manager, current_bin, guy, radius_squared, guy->collide_flags & manager->flags);
                }
            }
        }

        bin_index++;
        if (bin_index == num_bins_total) bin_index = 0;
        if (bin_index == bin1) break;
    }

    manager->next_bin_index_to_process = bin1;
}

#define MAX_CM   16
static Collision_Manager cms[MAX_CM];

Collision_Manager *createCollisionManager(float freq, float radius, int flags,
                  void (*hit_func)(Guy *a, Guy *b))
{
   for (int i=0; i < MAX_CM; ++i) {
      if (!cms[i].active) {
         cms[i].active = TRUE;
         cms[i].flags = flags;
         cms[i].collision_radius = radius;
         cms[i].fraction_of_bins_to_process_per_second = freq;
         cms[i].next_bin_index_to_process = (73 * i) & (NUM_BINS-1);
         cms[i].hit_func = hit_func;
         cms[i].end_func = NULL;
         return &cms[i];
      }
   }
   return NULL;
}

void deleteCollisionManager(Collision_Manager *c)
{
   assert(c->active);
   for (int i=0; i < MAX_CM; ++i) {
      if (c == &cms[i]) {
         c->active = FALSE;   
         return;
      }
   }
   assert(0);
}

void runCollisionManager(Collision_Manager *c, float dt)
{
   guysFindCollisions(c, dt);
}

void findCollisions(float dt)
{
   for (int i=0; i < MAX_CM; ++i) {
      if (cms[i].active)
         runCollisionManager(&cms[i], dt);
   }
}

void initFlocking(float freq, float radius, int flags)
{
   Collision_Manager *cm = createCollisionManager(freq, radius, flags, flockGuyCollisionResponse);
   cm->end_func = flockPostBin;
}

void findGuysInSquare(float x0, float y0, float x1, float y1, void (*f)(Guy *, void *p), void *p)
{
   int gx0,gx1,gy0,gy1;
   gx0 = floor(x0 / BIN_SIZE);
   gy0 = floor(y0 / BIN_SIZE);
   gx1 = floor(x1 / BIN_SIZE);
   gy1 = floor(y1 / BIN_SIZE);
   if (gx0 < 0) gx0 = 0;
   if (gx1 >= BIN_X) gx1 = BIN_X-1;
   if (gy0 < 0) gy0 = 0;
   if (gy1 >= BIN_Y) gy1 = BIN_Y-1;
   for (int y = gy0; y <= gy1; ++y) {
      for (int x = gx0; x <= gx1; ++x) {
         Bin *b = &bin[BIN(x,y)];
         int n = b->num_guys_in_bin;
         for (int i=0; i < n; ++i) {
            Guy *g = b->guys[i];
            if (g->loc.x >= x0 && g->loc.y >= y0 && g->loc.x <= x1 && g->loc.y <= y1) {
               f(g, p);
            }
         }
      }
   }
}

void findGuysInCircle(float cx, float cy, float r, void (*f)(Guy *, float dsquared, void *p), void *p)
{
   float r2 = r*r;
   float x0 = cx-r, y0 = cy-r, x1 = cx+r, y1 = cy+r;
   int gx0,gx1,gy0,gy1;
   gx0 = floor(x0 / BIN_SIZE);
   gy0 = floor(y0 / BIN_SIZE);
   gx1 = floor(x1 / BIN_SIZE);
   gy1 = floor(y1 / BIN_SIZE);
   if (gx0 < 0) gx0 = 0;
   if (gx1 >= BIN_X) gx1 = BIN_X-1;
   if (gy0 < 0) gy0 = 0;
   if (gy1 >= BIN_Y) gy1 = BIN_Y-1;
   for (int y = gy0; y <= gy1; ++y) {
      for (int x = gx0; x <= gx1; ++x) {
         Bin *b = &bin[BIN(x,y)];
         int n = b->num_guys_in_bin;
         for (int i=0; i < n; ++i) {
            Guy *g = b->guys[i];
            float d2 = (g->loc.x - cx)*(g->loc.x - cx) + (g->loc.y - cy)*(g->loc.y - cy);
            if (d2 <= r2) {
               f(g, d2, p);
            }
         }
      }
   }
}

void findGuysInSphere(float cx, float cy, float cz, float r, void (*f)(Guy *, float dsquared, void *p), void *p)
{
   float r2 = r*r;
   float x0 = cx-r, y0 = cy-r, x1 = cx+r, y1 = cy+r;
   int gx0,gx1,gy0,gy1;
   gx0 = floor(x0 / BIN_SIZE);
   gy0 = floor(y0 / BIN_SIZE);
   gx1 = floor(x1 / BIN_SIZE);
   gy1 = floor(y1 / BIN_SIZE);
   if (gx0 < 0) gx0 = 0;
   if (gx1 >= BIN_X) gx1 = BIN_X-1;
   if (gy0 < 0) gy0 = 0;
   if (gy1 >= BIN_Y) gy1 = BIN_Y-1;
   for (int y = gy0; y <= gy1; ++y) {
      for (int x = gx0; x <= gx1; ++x) {
         Bin *b = &bin[BIN(x,y)];
         int n = b->num_guys_in_bin;
         for (int i=0; i < n; ++i) {
            Guy *g = b->guys[i];
            float d2 = (g->loc.x - cx)*(g->loc.x - cx) + (g->loc.y - cy)*(g->loc.y - cy) + (g->loc.z - cz)*(g->loc.z - cz);
            if (d2 <= r2) {
               f(g, d2, p);
            }
         }
      }
   }
}

void setSpeed(Guy *g, float speed)
{
   speed = 256.0 * speed / GUY_MAX_SPEED;
   if (speed > 255) speed = 255;
   g->speed = speed;
}

int find_animation_direction(float x, float y) {
   double theta = atan2(y, x);
   if (theta < 0) theta += 2 * M_PI;

   float fraction = theta * (1.0 / (2 * M_PI));
   if (fraction > 1) fraction = 1;
   float scaled = fraction * 255.0f;

   int direction = (int)(scaled + 0.5f);
   return direction;
}


void setDirection(Guy *g, float vx, float vy) {
    g->vel.x = vx;
    g->vel.y = vy;
    g->animation_direction = find_animation_direction(vx, vy);
}

void setVelocity(Guy *g, float vx, float vy, float vz) {
    setDirection(g, vx, vy);
    g->vel.z = vz;
}

void setDirectionAndSpeed(Guy *g, float vx, float vy, float speed)
{
   setDirection(g, vx, vy);
   setSpeed(g, speed);
}

float gravity[256] = { 0,0,0, DEFAULT_GRAVITY };
int z;

// set this guy to run along his current velocity vector
void aimGuyAlong(Guy *g)
{
   float d = sqrt(g->vel.x*g->vel.x + g->vel.y*g->vel.y);
   float speed = (g->speed * GUY_MAX_SPEED) / 256.0;
   if (d != 0) {
      float mul = speed/d;
      setVelocity(g, g->vel.x*mul, g->vel.y*mul, g->vel.z);
   } else {
      g->vel.x = g->vel.y = g->vel.z = 0;
   }
}

int guyAimFor(Guy *g, float x, float y, float speed)
{
   float d = (x - g->loc.x)*(x - g->loc.x) + (y - g->loc.y)*(y - g->loc.y);
   int dt;
   setDirectionAndSpeed(g, x - g->loc.x, y - g->loc.y, speed);
   g->physics_mode = PM_FLOORED;
   aimGuyAlong(g);
   dt = FLOAT_TIME_TO_INT_TIME(sqrt(d) / speed);
   if (dt > INT_MAX_TIMER) dt = INT_MAX_TIMER;
   if (dt < 5) dt = 5;
   return dt;
}

int guyAimNear(Guy *g, float x, float y, float r, float speed)
{
   return guyAimFor(g, x+(GenerateUnitRandomReal()*2-1)*r, y+(GenerateUnitRandomReal()*2-1)*r, speed);
}

static Guy **aitimer;
static int max_aitimer, queue_aitimer;

static void queueAppGuyAITimer(Guy *g)
{
   if (queue_aitimer >= max_aitimer) {
      if (max_aitimer == 0) {
         max_aitimer = 512;
         aitimer = (Guy **) malloc(sizeof(aitimer[0]) * max_aitimer);
      } else {
         max_aitimer *= 2;
         aitimer = (Guy **) realloc(aitimer, sizeof(aitimer[0]) * max_aitimer);
      }
   }
   aitimer[queue_aitimer++] = g;
}

static void dequeueAITimer(void)
{
   int i;
   for (i=0; i < queue_aitimer; ++i) {
       if(gameGuyAITimer) {
           gameGuyAITimer(aitimer[i]);
       }
   }
   queue_aitimer = 0;
}

inline int convert(int theta_256) {
    return ((272 - theta_256) & 255) >> 5;
}

struct TexVertex
{
   float x,y,z;
   uint16 s,t;
};

static void renderDudesBegin(void)
{
   glEnable(GL_BLEND);
   glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA);

   glEnable(GL_ALPHA_TEST);
   glAlphaFunc(GL_GREATER, 0.6f);

   glPolygonOffset(-4,-4);
   glEnable(GL_POLYGON_OFFSET_FILL);
}

static void renderDudesEnd(void)
{
   glDisable(GL_BLEND);
   glDisable(GL_ALPHA_TEST);
   glDisable(GL_POLYGON_OFFSET_FILL);
}

void binMove(Guy *g, int b)
{
   binRemove(GUYBIN(g), g);
   binAdd(b, g);
}

#define MEM_QUADS   128
static uint16 indices[65536];

static inline void simulate_guy(Guy *g, float dt_float)
{
   float old_x = g->loc.x;
   float old_y = g->loc.y;

   g->loc.x += g->vel.x * dt_float;
   g->loc.y += g->vel.y * dt_float;
   if (g->physics_mode != PM_FLY_THROUGH_BUILDINGS) {
      int zx = FLOAT2INT_BUILDING_X(g->loc.x);
      int zy = FLOAT2INT_BUILDING_Y(g->loc.y);

      // test for 0 <= fx < max, 0 <= fy < max
      if (BUILDING_TEST(zx,zy)) {
         // infrequent
         g->loc.x = old_x;
         g->loc.y = old_y;

         int ox = FLOAT2INT_BUILDING_X(old_x);
         int oy = FLOAT2INT_BUILDING_Y(old_y);

         // decide which axes to bounce
         if (ox == zx) {
            g->vel.y = -g->vel.y;
         } else if (oy == zy) {
            g->vel.x = -g->vel.x;
         } else {
            // both of them changed. so we might bounce on a diagonal...
            // so if the north is blocked we bounce down; if west is blocked
            // we bounce east; if both, bounce both; if neither, but diagonal,
            // then bounce randomly (based on whichever corner is closer)
            if (BUILDING_TEST(zx, oy)) {
               g->vel.x = -g->vel.x;
               if (BUILDING_TEST(ox,zy))
                  g->vel.y = -g->vel.y;
            } else if (BUILDING_TEST(ox,zy)) {
               g->vel.y = -g->vel.y;
            } else {
               g->vel.x = -g->vel.x; // not really perfect
            }
         }
         setVelocity(g, g->vel.x, g->vel.y, g->vel.z); // update animation facing
      }
   }
   int fx = FLOAT2FIX_HF(g->loc.x);
   int fy = FLOAT2FIX2_HF(g->loc.y);

   int gx = fx >> 16;
   int gy = fy >> 16;

   int rx = fx & 65535;
   int ry = fy & 65535;

   float h;

   #define HF_ height_field

   if (rx > ry) {  // random
      h = HF_[gy][gx] + (HF_[gy][gx+1]-HF_[gy][gx])*rx/65536.0 + (HF_[gy+1][gx+1]-HF_[gy][gx+1])*ry/65536.0;
   } else {
      h = HF_[gy][gx] + (HF_[gy+1][gx+1]-HF_[gy+1][gx])*rx/65536.0 + (HF_[gy+1][gx]-HF_[gy][gx])*ry/65536.0;
   }

   if (g->physics_mode == PM_FLOORED)
      g->loc.z = h;
   else {
      float z0 = g->vel.z;
      g->vel.z -= gravity[g->physics_mode] * dt_float;
      g->loc.z += (g->vel.z + z0)/2*dt_float;
      if (g->loc.z < h) {
         g->loc.z = h;
         if (g->vel.z < 0) {
             float r = 1.0;
             if(gameGuyHitTerrain) {
                 r = gameGuyHitTerrain(g);
             }
             g->vel.z = -g->vel.z*r;
         }
      }
   }
   appGuyMove(g, g->loc);
}


void appGuyMove(Guy* g, const vec3f& new_loc)
{
	g->loc = new_loc;
	int b = COMPUTE_BIN(g);
	if (b != GUYBIN(g)) binMove(g, b);
}

static uint32 color_data_source[MEM_QUADS*4];
static uint32 color_table[256] = {
   0xffffffff,
   0xff808080,
   0xff4040ff,
   0xff40ff40,
   0xffff4040,
   0xffffff40,
   0xff40ffff,
   0xffff40ff,
};


static struct init {
	init() {
		for (int i = 1; i < 256; i++) {
			float	f = genrand_real1();
			uint8	r = 190 + 30 * f + genrand_real1() * 31;
			uint8	g = 190 + 30 * f + genrand_real1() * 15;
			uint8	b = 190 + 30 * f + genrand_real1() * 20;
			color_table[i] = 0xFF000000 | (r << 16) | (g << 8) | b;
		}
	}
} init;


void drawAndRun(Guy *g, int n, int dt)
{
   assert(sizeof(Guy) == 64);
   if (sizeof(Guy) != 64) return;
   binClearAll();

   vec3f *CH = (vec3f*)CameraHorizontal();
   vec3f *CV = (vec3f*)CameraVertical();

   vec3f CV_fixed;
   CV_fixed.x = CV->x;
   CV_fixed.y = CV->y;
   CV_fixed.z = CV->z + 0.5;
   vec3f_normeq(&CV_fixed);
   CV = &CV_fixed;

   TexVertex *vertex_data_orig = NULL;
   vertex_data_orig = new TexVertex[MEM_QUADS*4];

   glEnableClientState(GL_VERTEX_ARRAY);
   glEnableClientState(GL_TEXTURE_COORD_ARRAY);

   glTexCoordPointer(2, GL_SHORT, sizeof(vertex_data_orig[0]), &vertex_data_orig[0].s);
   glVertexPointer(3, GL_FLOAT, sizeof(vertex_data_orig[0]), &vertex_data_orig[0].x);

#ifdef DEBUG_COLOR
   glColorPointer(4, GL_UNSIGNED_BYTE, 0, &color_data_source);
   glEnableClientState(GL_COLOR_ARRAY);
#endif

   glMatrixMode(GL_TEXTURE);
   glPushMatrix();
   glScalef(1/32768.0, 1/32768.0, 1);

   float dt_float = INT_TIME_TO_FLOAT_TIME(dt);

   g_animation_manager->ensure_texture_is_ready();
   glBindTexture(GL_TEXTURE_2D, g_animation_manager->packed_sprite_texture_handle);
   glColor3f(1,1,1);

   Animation_Texture_Set *texture_sets = g_animation_manager->texture_sets;
   Animation_Mode *animation_modes = g_animation_manager->animation_modes;

   sprite_image *image = &g_animation_manager->my_sprite_texture->Animations[0].Frames[0].Directions[0];

   renderDudesBegin();

   const float *camera_x_vec = CameraHorizontal();
   const float *camera_y_vec = CameraBackward();

   float camera_x = camera_y_vec[0];
   float camera_y = camera_y_vec[1];

   int camera_direction_256 = find_animation_direction(camera_x, camera_y) + 272;

   for (int i=0; i < n; i += MEM_QUADS) {
      int k = MEM_QUADS;
      if (i + k > n) k = n - i;
      int vertex_count = k*4;
      Guy *end_guy = g + k;

      TexVertex *vertex_data = vertex_data_orig;
#ifdef DEBUG_COLOR
      uint32 *color_data = color_data_source;
#endif
      do {
         do_prefetch(g+8);  // may be a no-op depending on #defines; prefetch 512 bytes ahead

         simulate_guy(g, dt_float);

         if (dt > g->ai_timer) {
            g->ai_timer = 0xffff;
            if (queue_aitimer < max_aitimer)
               aitimer[queue_aitimer++] = g;
            else
               queueAppGuyAITimer(g);
         } else {
            g->ai_timer -= dt;
         }

         #ifndef QUICK_IMAGE
         {  // brace this code to make it obvious to compiler that only 'image' escapes
            uint8 animation_mode_index = g->animation_mode_index;
            Animation_Mode *mode = &animation_modes[animation_mode_index];
            int packing_index = mode->texture_set->sprite_texture_index;
            int frame_index = ((mode->frame_index_times_256 + (g->animation_timing_offset << 3)) >> 8);

            // the following sequence must be computed AFTER sim, which can change direction
            int direction_index = 0; // can be either 0 or 8 directions
            assert(frame_index < g_animation_manager->my_sprite_texture->Animations[packing_index].FrameCount*2);
            if(g_animation_manager->my_sprite_texture->Animations[packing_index].Frames[0].DirectionCount != 1)
            {
                direction_index = ((camera_direction_256 - g->animation_direction) & 255) >> 5;
            }
            image = &g_animation_manager->my_sprite_texture->Animations[packing_index].Frames[frame_index].Directions[direction_index];
         }
         #endif

         float chx = CH->x * image->HalfWidth;
         float chy = CH->y * image->HalfWidth;
         float chz = CH->z * image->HalfWidth;
         float cvx = CV->x * image->FullHeight;
         float cvy = CV->y * image->FullHeight;
         float cvz = CV->z * image->FullHeight;

         *(int *)&vertex_data[0].s = image->packed_u0v0;
         vertex_data[0].x = g->loc.x - chx;
         vertex_data[0].y = g->loc.y - chy;
         vertex_data[0].z = g->loc.z - chz;
         *(int *)&vertex_data[1].s = image->packed_u1v0;
         vertex_data[1].x = g->loc.x + chx;
         vertex_data[1].y = g->loc.y + chy;
         vertex_data[1].z = g->loc.z + chz;
         *(int *)&vertex_data[2].s = image->packed_u1v1;
         vertex_data[2].x = vertex_data[1].x + cvx;
         vertex_data[2].y = vertex_data[1].y + cvy;
         vertex_data[2].z = vertex_data[1].z + cvz;
         *(int *)&vertex_data[3].s = image->packed_u0v1;
         vertex_data[3].x = vertex_data[0].x + cvx;
         vertex_data[3].y = vertex_data[0].y + cvy;
         vertex_data[3].z = vertex_data[0].z + cvz;
         vertex_data += 4;
#ifdef DEBUG_COLOR
         uint32 color = color_table[g->debug_color];
         color_data[0] = color;
         color_data[1] = color;
         color_data[2] = color;
         color_data[3] = color;
         color_data += 4;
#endif

         ++g;
      } while (--k);
      glDrawArrays(GL_QUADS, 0, vertex_count);
      //glDrawElements(GL_QUADS, vertex_count, GL_UNSIGNED_SHORT, indices);
   }

   glDisableClientState(GL_VERTEX_ARRAY);
   glDisableClientState(GL_TEXTURE_COORD_ARRAY);
   glDisableClientState(GL_COLOR_ARRAY);

   glMatrixMode(GL_TEXTURE);
   glPopMatrix();

   renderDudesEnd();

   delete [] vertex_data_orig;
   dequeueAITimer();
}


//////////////////////////////////////////////////////////////////////////////////////

void drawAndRunGuys(int dt)
{
   drawAndRun(gGuys, gGuyCount, dt);
}

void guysDraw(float dt)
{
   glEnable(GL_TEXTURE_2D);
   drawHF();
   glDisable(GL_TEXTURE_2D);
   drawBuildings();
   glEnable(GL_TEXTURE_2D);
   int idt = FLOAT_TIME_TO_INT_TIME(dt);
   if (idt == 0) idt = 1;
   drawAndRunGuys(idt);
}

static void *alloc_guys;

// @todo get this outta here!
float floatrand(void)
{
   return (rand() & 32767) / 32768.0;
}

float floatrandneg(void)
{
   return floatrand()*2-1;
}
void guysClearGuy( Guy *guy )
{
    memset(guy,0,sizeof(Guy));
    setVelocity(guy, 0,0,0);
    guy->physics_mode = PM_FLOORED;
    guy->ai_timer  = INT_MAX_TIMER;
    guy->debug_color = DEBUG_COLOR_NORMAL;
}

void guysInit(void)
{
   int i;

   assert(BIN_X * BIN_SIZE >= WORLD_X_POWER_OF_2);
   assert(BIN_Y * BIN_SIZE >= WORLD_Y_POWER_OF_2);

   hfInit();
   texInit();
   buildingsInit();

   alloc_guys = malloc(MAX_GUYS * sizeof(gGuys[0]) + 4096); // extra for prefetching

   // cache align
   gGuys = (Guy *) ((((uint32) alloc_guys) + 63) & ~63);

   for (i=0; i < MAX_GUYS; ++i) {
      guysClearGuy(gGuys+i);
   }
}
void guysRegisterRef( Guy *guy, Guy **ref_guy )
{
    assert(guy && !guy->ref_guy && ref_guy);
    if(guy && !guy->ref_guy && ref_guy) {
        *ref_guy = guy;
        guy->ref_guy = ref_guy;
    }
}
void guysUnregisterRef( Guy **ref_guy )
{
    assert(ref_guy);
    if(ref_guy && *ref_guy) {
        (*ref_guy)->ref_guy = 0;
        *ref_guy = 0;
    }
}

void guysFinishInit(int guyCnt)
{
   int i;
   gGuyCount=guyCnt;

   for (i=0; i < MAX_GUYS; ++i)
      aimGuyAlong(&gGuys[i]);

   for (i=0; i < gGuyCount; ++i) {
      binGuy(i);
   }
}

// @todo!!!
void guysKey(int ch)
{
   int n;
   switch (ch) {
      case '-': n = gGuyCount >> 1; if (n < 1) n = 1; setNumGuys(n); break;
      case '+': case '=': n = gGuyCount << 1; if (n > MAX_GUYS) n = MAX_GUYS; setNumGuys(n); break;
   }
}

void collision_init(void) {
    shuffle_init();
}
