/* ========================================================================
   $RCSfile: image_operations.cpp,v $
   $Date: 2002/03/14 01:08:43 $
   $Revision: 1.5 $
   $Creator: Jeff Roberts $
   $Notice: $
   ======================================================================== */

#include <stdlib.h>
#include <math.h>

#if !defined(IMAGE_OPERATIONS_H)
#include "image_operations.h"
#endif

#if !defined(LOG_H)
#include "log.h"
#endif

typedef real32 SCALEFLOAT;
typedef uint32 u32;
typedef int32 s32;
typedef uint16 u16;
typedef int16 s16;
typedef uint8 u8;
typedef int8 s8;

#define radmalloc malloc
#define radfree free

typedef struct
{
  SCALEFLOAT    weight;
  int32   pixel;
} CONTRIB;

typedef struct RADSCALE
{
  uint32 destwidth;
  uint32 destheight;

  uint32 srcwidth;
  uint32 srcheight;

  int32* w_contrib_counts;
  CONTRIB* w_contribs;

  int32 w_contrib_max;
  int32* h_contrib_counts;

  CONTRIB* h_contribs;
  int32 h_contrib_max;

  SCALEFLOAT tmp[1];
} RADSCALE;

RADSCALE* RADScaleOpen(int32 filter_type, uint32 destwidth,uint32 destheight,
                       uint32 srcwidth, uint32 srcheight);
void RADScale32(RADSCALE* rs, void* dest, u32 destpitch, void const * src, u32 srcpitch);
void RADScaleClose(RADSCALE* rs);

void
ScaleImage(pixel_filter_type FilterType,
           int32 SourceWidth, int32 SourceHeight, int32 SourceStride,
           uint8 const *SourcePixels,
           int32 DestWidth, int32 DestHeight, int32 DestStride,
           uint8 *DestPixels)
{
    if(FilterType < OnePastLastPixelFilterType)
    {
        RADSCALE *Scaler = RADScaleOpen(FilterType, DestWidth, DestHeight,
                                        SourceWidth, SourceHeight);
        RADScale32(Scaler, DestPixels, DestStride, SourcePixels, SourceStride);
        RADScaleClose(Scaler);
    }
    else
    {
        Log("Invalid pixel filter specified (%d)", FilterType);
    }
}

// adapted from Graphics Gems III

typedef SCALEFLOAT (*filterft)(SCALEFLOAT t);

#define box_support     (0.5)

SCALEFLOAT box_filter(SCALEFLOAT t)
{
  if ((t > -0.5) && (t <= 0.5))
    return(1.0);
  return(0.0);
}

#define linear_support  (1.0)

SCALEFLOAT linear_filter( SCALEFLOAT t)
{
  if (t < 0.0)
    t = -t;
  if(t < 1.0)
    return((SCALEFLOAT)(1.0 - t));
  return(0.0);
}

#define mcubic_support  (2.0)

SCALEFLOAT mcubic_filter(SCALEFLOAT t)
{
  SCALEFLOAT tt;

  tt = t * t;

  if (t < 0)
    t = -t;

  if(t < 1.0)
  {
    return((SCALEFLOAT)(  ((7.0/6.0)*t*tt)+
                          (-2.0*tt)+
                          (8.0/9.0)));
  }
  else
  {
    if (t<2.0)
    {
      return((SCALEFLOAT)(  ((-7.0/18.0)*t*tt)+
                            (2.0*tt)+
                            ((-10.0/3.0)*t)+
                            (16.0/9.0)));
    }
  }
  return(0);
}

static void calc_contribs(u32 dest,u32 src,s32** ret_contrib_counts,CONTRIB** ret_contribs,s32* ret_contrib_max,filterft filterf,SCALEFLOAT fwidth)
{
  SCALEFLOAT scale;
  s32* contrib_counts=0;
  CONTRIB* contribs=0;
  s32 contrib_max=0;
  s32 i,j;

  if ( ( src != 1 ) && ( dest != 1 ) )
    scale = ( SCALEFLOAT ) ( dest - 1 ) / ( SCALEFLOAT ) ( src - 1 );
  else
    scale = ( SCALEFLOAT ) dest / ( SCALEFLOAT ) src;

  // pre-calculate filter contributions

  if (scale < 1.0)
  {
    // shrinking the area
    SCALEFLOAT width = fwidth / scale;

    contrib_max=((s32)ceil(width * 2 + 1));
    contrib_counts=(s32 *)radmalloc( (sizeof(s32)+
                                      sizeof(CONTRIB)*contrib_max)*dest );
    if (contrib_counts==0)
      goto dump;

    contribs=(CONTRIB*)(contrib_counts+dest);

    for (i = 0; i < (s32)dest; ++i)
    {
        SCALEFLOAT center;
       s32 right;
       CONTRIB* cur=contribs+i*contrib_max;
       SCALEFLOAT total_weight = 0.0F;
       s32 total_count = 0;

       center = (SCALEFLOAT) i / scale;
       right = (s32)floor(center + width);

       for (j =(s32)ceil(center - width); j <= right; ++j)
       {
         if (j<0)
         {
           cur->pixel = ( -j ) % src;
         }
         else if (j>=(s32)src)
         {
           cur->pixel = (src - 1 ) - ( ( j - ( src - 1 ) ) % src );
         }
         else
         {
           cur->pixel = j;
         }
         cur->weight=(*filterf)((center - (SCALEFLOAT) j) * scale) * scale;
         total_weight += cur->weight;
         ++total_count;
         cur++;
       }

       //
       // Normalize the weights...
       //

       total_weight = ( 1.0f / total_weight );
       cur=contribs+i*contrib_max;
       while( total_count-- )
       {
         cur->weight *= total_weight;
         cur++;
       }

       contrib_counts[i]=cur-(contribs+i*contrib_max);
    }
  }
  else
  {
    // zooming the area
    contrib_max=(int32)(fwidth * 2 + 1);
    contrib_counts=(s32 *)radmalloc( (sizeof(s32)+
                                      sizeof(CONTRIB)*contrib_max)*dest );
    if (contrib_counts==0)
      goto dump;

    contribs=(CONTRIB*)(contrib_counts+dest);

    for(i = 0; i < (s32)dest; ++i)
    {
        SCALEFLOAT center;
       s32 right;
       CONTRIB* cur=contribs+i*contrib_max;
       SCALEFLOAT total_weight = 0.0F;
       s32 total_count = 0;

       center = (SCALEFLOAT) i / scale;
       right = (s32)floor(center + fwidth);
       for (j =(s32)ceil(center - fwidth); j <= right; ++j)
       {
         if (j<0)
         {
           cur->pixel = ( -j ) % src;
         }
         else if (j>=(s32)src)
         {
           cur->pixel = (src - 1 ) - ( ( j - ( src - 1 ) ) % src );
         }
         else
         {
           cur->pixel = j;
         }
         cur->weight=(*filterf)(center - (SCALEFLOAT) j);
         total_weight += cur->weight;
         ++total_count;
         cur++;
       }

       //
       // Normalize the weights...
       //

       total_weight = ( 1.0f / total_weight );
       cur=contribs+i*contrib_max;
       while( total_count-- )
       {
         cur->weight *= total_weight;
         cur++;
       }

       contrib_counts[i]=cur-(contribs+i*contrib_max);
    }
  }
 dump:
  *ret_contrib_counts=contrib_counts;
  *ret_contribs=contribs;
  *ret_contrib_max=contrib_max;
}


static SCALEFLOAT filter_support[]={mcubic_support,linear_support,box_support};
static filterft filter_function[]={mcubic_filter,linear_filter,box_filter};

RADSCALE* RADScaleOpen(s32 filter_type, u32 destwidth,u32 destheight,u32 srcwidth, u32 srcheight)
{
  RADSCALE* ret;
  SCALEFLOAT fwidth;
  filterft filterf;

  ret=(RADSCALE*)radmalloc(sizeof(RADSCALE)+(srcheight * sizeof(SCALEFLOAT)*4));
  if (ret==0)
    return(0);

  fwidth=filter_support[filter_type];
  filterf=filter_function[filter_type];

  ret->destwidth=destwidth;
  ret->destheight=destheight;
  ret->srcwidth=srcwidth;
  ret->srcheight=srcheight;

  calc_contribs(destwidth,srcwidth,&ret->w_contrib_counts,&ret->w_contribs,&ret->w_contrib_max,filterf,fwidth);
  if (ret->w_contrib_counts==0)
  {
   err:
    radfree(ret);
    return(0);
  }

  calc_contribs(destheight,srcheight,&ret->h_contrib_counts,&ret->h_contribs,&ret->h_contrib_max,filterf,fwidth);
  if (ret->h_contrib_counts==0)
  {
    radfree(ret->w_contrib_counts);
    goto err;
  }

  return(ret);
}

void RADScaleClose(RADSCALE* rs)
{
  radfree(rs->w_contrib_counts);
  radfree(rs->h_contrib_counts);
  radfree(rs);
}

void RADScale32(RADSCALE* rs, void* dest, u32 destpitch, void const * src, u32 srcpitch)
{
  s32 xx,i, j;          /* loop variables */
  SCALEFLOAT weight;
  SCALEFLOAT weighta,weightr,weightg,weightb;   /* filter calculation variables */
  CONTRIB* cur;

  for(xx = 0; xx < (s32)rs->destwidth; xx++)
  {

    /* Apply horz filter to make dst column in tmp. */
    SCALEFLOAT* colu=rs->tmp;
    u8* scanline=((u8*)src);

    for(i = 0; i < (s32)rs->srcheight; ++i)
    {
      weighta = 0.0;
      weightr = 0.0;
      weightg = 0.0;
      weightb = 0.0;

      cur=rs->w_contribs+xx*rs->w_contrib_max;
      for(j = 0; j < rs->w_contrib_counts[xx]; ++j)
      {
          s32 pel = *((uint32 *)scanline + cur[j].pixel);

          weight = cur[j].weight;
//#define TEST
#ifdef TEST
        weighta += 0 * weight;
        weightr += 255 * weight;
        weightg += 255 * weight;
        weightb += 255 * weight;
#else
        weighta += ((u32)pel>>24) * weight;
        weightr += ((pel>>16)&255) * weight;
        weightg += ((pel>>8)&255) * weight;
        weightb += (pel&255) * weight;
#endif
      }

      colu[0] = weighta;
      colu[1] = weightr;
      colu[2] = weightg;
      colu[3] = weightb;

      colu+=4;
      scanline+=srcpitch;
    }


    /* The temp column has been built. Now stretch it
         vertically into dst column. */
    scanline=((u8*)dest)+(xx*4);
    for(i = 0; i < (s32)rs->destheight; ++i)
    {
      weighta = 0.0;
      weightr = 0.0;
      weightg = 0.0;
      weightb = 0.0;

      cur=rs->h_contribs+i*rs->h_contrib_max;
      for(j = 0; j < rs->h_contrib_counts[i]; ++j)
      {
        colu=rs->tmp+(cur[j].pixel*4);
        weight = cur[j].weight;

        weighta += colu[0] * weight;
        weightr += colu[1] * weight;
        weightg += colu[2] * weight;
        weightb += colu[3] * weight;
      }

      if (weighta>0.0)
        j=(weighta>=254.5)?(255<<24):((int32(weighta + 0.5f))<<24);
      else
        j=0;
      if (weightr>0.0)
        j|=(weightr>=254.5)?(255<<16):((int32(weightr + 0.5f))<<16);
      if (weightg>0.0)
        j|=(weightg>=254.5)?(255<<8):((int32(weightg + 0.5f))<<8);
      if (weightb>0.0)
        j|=(weightb>=254.5)?255:((int32(weightb + 0.5f)));

      *((u32*)scanline)=j;

      scanline+=destpitch;
    }
  }

}

//
// Casey code here:
//


pixel_layout const RGB555PixelFormat =
{
    2, {0, 5, 10, 0}, {5, 5, 5, 0},
};

pixel_layout const RGB565PixelFormat =
{
    2, {0, 5, 11, 0}, {5, 6, 5, 0},
};

pixel_layout const RGBA5551PixelFormat =
{
    2, {0, 5, 10, 15}, {5, 5, 5, 1},
};

pixel_layout const RGBA4444PixelFormat =
{
    2, {0, 4, 8, 12}, {4, 4, 4, 4},
};

pixel_layout const RGB888PixelFormat =
{
    3, {0, 8, 16, 0}, {8, 8, 8, 0},
};

pixel_layout const RGBA8888PixelFormat =
{
    4, {0, 8, 16, 24}, {8, 8, 8, 8},
};

pixel_layout const BGR555PixelFormat =
{
    2, {10, 5, 0, 0}, {5, 5, 5, 0},
};

pixel_layout const BGR565PixelFormat =
{
    2, {11, 5, 0, 0}, {5, 6, 5, 0},
};

pixel_layout const BGRA5551PixelFormat =
{
    2, {10, 5, 0, 15}, {5, 5, 5, 1},
};

pixel_layout const BGRA4444PixelFormat =
{
    2, {8, 4, 0, 12}, {4, 4, 4, 4},
};

pixel_layout const BGR888PixelFormat =
{
    3, {16, 8, 0, 0}, {8, 8, 8, 0},
};

pixel_layout const BGRA8888PixelFormat =
{
    4, {16, 8, 0, 24}, {8, 8, 8, 8},
};

static uint32
GenerateMask(int const FirstBit,
             int const OnePastLastBit)
{
    // Build a mask that has a consecutive batch of 1's in some part
    // of the word.
    uint32 Mask = 0;
    {for(int BitIndex = FirstBit;
         BitIndex < OnePastLastBit;
         ++BitIndex)
    {
        Mask |= 1 << BitIndex;
    }}

    return(Mask);
}

static uint32
ShiftMore(uint32 const Value, int32 const Shift)
{
    return((Shift > 0) ? (Value<<Shift) : (Value>>(-Shift)));
}

inline uint32 Reverse32(uint32 Value)
    {return((Value << 24) | ((Value & 0x0000FF00) << 8) |
            ((Value & 0x00FF0000) >> 8) | (Value >> 24));}

static uint32
ConvertPixel(int32 *Shifts, uint32 *Masks, uint32 Pixel)
{
    if(PROCESSOR_BIG_ENDIAN)
    {
        Pixel = Reverse32(Pixel);
    }
    
    Pixel = ((ShiftMore(Pixel, Shifts[0]) & Masks[0]) +
             (ShiftMore(Pixel, Shifts[1]) & Masks[1]) +
             (ShiftMore(Pixel, Shifts[2]) & Masks[2]) +
             (ShiftMore(Pixel, Shifts[3]) & Masks[3]));

    if(PROCESSOR_BIG_ENDIAN)
    {
        Pixel = Reverse32(Pixel);
    }

    return(Pixel);
}

void 
ConvertPixelFormat(int32 Width, int32 Height,
                   pixel_layout const &SourceLayout,
                   int32 SourceStride, void const *SourceBitsInit,
                   pixel_layout const &DestLayout,
                   int32 DestStride, void *DestBitsInit)
{    
    uint8 const *SourceBits = (uint8 const *)SourceBitsInit;
    uint8 *DestBits = (uint8 *)DestBitsInit;
    
    // Determine the shifts and masks
    int32 Shifts[4];
    uint32 Masks[4];
    {for(int ComponentIndex = 0;
         ComponentIndex < 4;
         ++ComponentIndex)
    {
        int32 const FromShift = SourceLayout.ShiftForComponent[ComponentIndex];
        int32 const FromCount = SourceLayout.BitsForComponent[ComponentIndex];
            
        int32 const ToShift = DestLayout.ShiftForComponent[ComponentIndex];
        int32 const ToCount = DestLayout.BitsForComponent[ComponentIndex];

        // Determine the shift
        int32 const RelativeShift = (ToShift + ToCount) - (FromShift + FromCount);

        // Determine the mask
        int32 const RelativeFromShift = FromShift + RelativeShift;
        uint32 Mask = 0;
        if(ToCount > 0)
        {
            Mask = GenerateMask((RelativeFromShift > ToShift) ?
                                RelativeFromShift : ToShift,
                                RelativeFromShift + FromCount);
        }
        
        // Store the shift and mask
        Shifts[ComponentIndex] = RelativeShift;
        Masks[ComponentIndex] = Mask;
    }}

    // Prepare the iterators
    int32 const SourceGap = SourceStride - (Width * SourceLayout.BytesPerPixel);
    int32 const DestGap = DestStride - (Width * DestLayout.BytesPerPixel);

    // First, copy everything but the last scanline
    {for(int32 Y = (Height - 1);
         Y;
         --Y)
    {
        {for(int32 X = Width;
             X;
             --X)
        {
            // Get the source pixel and advance the source pointer
            uint32 const Pixel = *((uint32*)SourceBits);
            SourceBits += SourceLayout.BytesPerPixel;

            // Write the destination pixel
            *((uint32*)DestBits) = ConvertPixel(Shifts, Masks, Pixel);
            
            // Advance the destination pointer
            DestBits += DestLayout.BytesPerPixel;
        }}

        // Account for the difference in the strides and the widths
        SourceBits += SourceGap;
        DestBits += DestGap;
    }}

    // Next, copy everything in the last scan line but the last pixel
    {for(int32 X = (Width - 1);
         X;
         --X)
    {
        // Get the source pixel and advance the source pointer
        uint32 const Pixel = *((uint32*)SourceBits);
        SourceBits += SourceLayout.BytesPerPixel;

        // Write the destination pixel
        *((uint32*)DestBits) = ConvertPixel(Shifts, Masks, Pixel);

        // Advance the destination pointer
        DestBits += DestLayout.BytesPerPixel;
    }}
    
    // Finally, copy the last pixel a byte at a time.  This is necessary
    // because otherwise we might over-read or over-write by one or two
    // bytes, depending on the bytes-per-pixel of the source and dest.
    uint32 Pixel = 0;
    uint8 *BytePointer = (uint8 *)&Pixel;
    int32 BytesPerPixel = SourceLayout.BytesPerPixel;
    while(BytesPerPixel--)
    {
        *BytePointer++ = *SourceBits++;
    }

    uint32 DestTemp = ConvertPixel(Shifts, Masks, Pixel);
    
    BytePointer = (uint8 *)&DestTemp;
    BytesPerPixel = DestLayout.BytesPerPixel;
    while(BytesPerPixel--)
    {
        *DestBits++ = *BytePointer++;
    }
}
