#include "sound_internal.h"
#include "mprintf.h"

typedef void **pointer;

void runtime_error(char *s) {
    fprintf(stdout, "FATAL RUNTIME ERROR: %s\n", s);
    assert(0);
}


struct stringbucket {
    stringbucket *next;
    char *point;
    char *start;
    char *end;
    char data[sizeof(pointer)];
};

#define SBMALLOCSIZ(c)  (c->block_size - 16)
#define SBUCKETSIZ (sizeof(struct stringbucket) - sizeof(pointer))
#define STRINGBUCKET_LEN(c) (SBMALLOCSIZ(c) - SBUCKETSIZ)

static void reset_stringbucket(stringbucket *b) {
    b->next = NULL;
    b->start = b->data;
    b->point = b->start;
}

static stringbucket *make_stringbucket(Concatenator *c) {
    stringbucket *b;

    b = (stringbucket *)malloc(SBMALLOCSIZ(c));
    b->end = b->data + STRINGBUCKET_LEN(c);

    reset_stringbucket(b);

    return b;
}

Concatenator::Concatenator(int _block_size) {
    block_size = _block_size;
    nbuckets = 1;
    first = last = make_stringbucket(this);
}

Concatenator::~Concatenator(void) {
    stringbucket *b, *c;
    
    for (b = first; b != NULL; b = c) {
        c = b->next;
        free((char *)b);
    }
}

#define IsFull(b) (b->point == b->end)
#define SpaceLeft(b) (b->end - b->point)

void Concatenator::add(char c) {
    if (IsFull(last)) expand();

    *(last->point++) = c;
}

void Concatenator::add_nozeroterm(char *s, unsigned int len) {
    unsigned int left;
    unsigned int amount;

    left = SpaceLeft(last);
    assert (left >= 0);

    if (left == 0) {
        expand();
        left = SpaceLeft(last);
    }

    amount = (left > len) ? len : left;

    memcpy(last->point, s, amount);
    last->point += amount;

    if (amount < len) add_nozeroterm(s + amount, len - amount);
}

void Concatenator::add(char *s) {
    add_nozeroterm(s, strlen(s));
}

void Concatenator::printf(char *fmt, ...) {
	va_list ap;
	va_start(ap, fmt);

	char *result = mprintf(fmt, ap);

	va_end(ap);

	if (result) {
		add(result);
		free(result);
	}
}

void Concatenator::add_asciiz(char *s) {
    add_nozeroterm(s, strlen(s) + 1);
}

unsigned int Concatenator::length(void) {
    unsigned int tally; 
    stringbucket *b;

    tally = 0;
    for (b = first; b != NULL; b = b->next) {
        tally += b->point - b->start;
    }

    return tally;
}

static void result_into_mem(stringbucket *b, char *s) {
    unsigned int len;

    if (b == NULL) return;
    len = b->point - b->data;
    memcpy(s, b->data, len);
    result_into_mem(b->next, s + len);
}

char *Concatenator::get_result(void) {
    unsigned int len;
    char *result;

    len = length();
    result = (char *)malloc(len + 1);

    result_into_mem(first, result);
    result[len] = '\0';

    return result;
}

char *Concatenator::get_nozeroterm_result(unsigned int *len_return) {
    unsigned int len;
    char *result;

    len = length();
    result = (char *)malloc(len);

    result_into_mem(first, result);
    *len_return = len;
    return result;
}

/*
 * In the functions below we assume that BUFSIZ is big enough to hold
 * the representation of integers or doubles.  This is hopefully a safe
 * assumption.
 *
 */
void Concatenator::add(int i) {
    char buf[BUFSIZ];

    sprintf(buf, "%d", i);
    add(buf);
}

void Concatenator::add(long l) {
    char buf[BUFSIZ];

    sprintf(buf, "%ld", l);
    add(buf);
}

void Concatenator::add(short i) {
    char buf[BUFSIZ];
    sprintf(buf, "%d", i);
    add(buf);
}

void Concatenator::add(unsigned int i) {
    char buf[BUFSIZ];

    sprintf(buf, "%u", i);
    add(buf);
}

void Concatenator::add(double d) {
    char buf[BUFSIZ];

    sprintf(buf, "%f", d);
    add(buf);
}

void Concatenator::add_u4b(int value) {
    unsigned char c0, c1, c2, c3;

    c0 = (value & 0x000000ff) >>  0;
    c1 = (value & 0x0000ff00) >>  8;
    c2 = (value & 0x00ff0000) >> 16;
    c3 = (value & 0xff000000) >> 24;

    int len = length();

    add((char)c0);
    add((char)c1);
    add((char)c2);
    add((char)c3);

    int len2 = length();
    assert(len2 == len + 4);
}

void Concatenator::add_u2b(int value) {
    unsigned char c0, c1;

    c0 = (value & 0x00ff) >>  0;
    c1 = (value & 0xff00) >>  8;

    add((char)c0);
    add((char)c1);
}

void Concatenator::add_u1b(int value) {
    unsigned char c0;
    c0 = (value & 0x00ff) >>  0;
    add((char)c0);
}

void Concatenator::expand(void) {
    stringbucket *b;

    b = make_stringbucket(this);
    assert(b != NULL);

    last->next = b;
    last = b;
    nbuckets++;
}

void Concatenator::reset(void) {
    stringbucket *buck, *buck2;

    buck = first->next;

    while (buck != NULL) {
        buck2 = buck->next;

        free((char *)buck);
        buck = buck2;
    }

    reset_stringbucket(first);
    last = first;
    nbuckets = 1;
}



List::List(void) {
    head = NULL;
    tail = NULL;
    items = 0;
}

void List::init(){
    head = NULL;
    tail = NULL;
    items = 0;    
}

List::~List(void) {
    Listnode *n, *m;

    for (n = head; n != NULL; n = m) {
        m = n->next;
        delete n;
    }
}

void *List::nth(unsigned int i) {
    assert ((int) i < items);

    Listnode *n = head;
    while (i--) {
        n = n->next;
    }

    assert(n != NULL);

    return n->data;
}


void List::add(void *item) {
    Listnode *n;

    n = new Listnode();
    n->next = NULL;
    n->data = item;
    n->prev = tail;

    if (tail == NULL) {
        head = tail = n;
    } else {
        tail->next = n;
        tail = n;
    }

    items++;
}

void List::add_at_head(void *item) {
    Listnode *n;

    n = new Listnode();
    n->prev = NULL;
    n->data = item;
    n->next = head;

    if (head == NULL) {
        head = tail = n;
    } else {
        head->prev = n;
        head = n;
    }

    items++;
}

// List::insert_behind()
//
// Inserts item BEHIND the the insert point (ipoint).  Returns true 
// if it could find the ipoint, and insert item behind it.  Otherwise
// it returns false.
//
// If ipoint == NULL, then item will be inserted as the head.
//
// If there are multiple items with the address ipoint, then we insert this
// behind the first one we encounter.
//
// returns:  [true/false] [item was inserted/not]

bool List::insert_behind(void *ipoint, void *item){
    Listnode *ln,*n;
    bool found=false;

    assert(item!=NULL);

    if(ipoint==NULL){
        add_at_head(item);
        return true;
    }

    for(ln=head; ln!=NULL; ln=ln->next){
        if(ln->data==ipoint){
            found = true;
            break;
        }
    }

    if(!found) return false;

    if(ln==tail){
        add(item);
        return true;
    }
  
    n = new Listnode();
    n->next = ln->next;
    n->data = item;
    n->prev = ln;
    ln->next->prev = n;
    ln->next = n;

    items++;

    return true;
}

// List::add_unique(void *item)
//
//  Adds the item to the list if it is not allready in there.
//  If it is, then it will not add it to the list.
//
//  returns:  [true/false] [item was added to list/not]

bool List::add_unique(void *item){
    Listnode *ln;

    for(ln=head; ln!=NULL; ln=ln->next){
        if(ln->data==item) return false;
    }
  
    add(item);
    return(true);
}



//
// List::remove
//
// Remove only the first occurrence of item in self.
//

bool List::remove(void *item) {
    Listnode *n;

    for (n = head; n != NULL; n = n->next) {
        if (n->data == item) {
            if (n->prev == NULL) {
                head = n->next;
            } else {
                n->prev->next = n->next;
            }

            if (n->next == NULL) {
                tail = n->prev;
            } else {
                n->next->prev = n->prev;
            }

            delete n;
            items--;
            return true;
        }
    }

    return false;
}


//
// List::remove_nth
//
// Remove the nth item of list (return it)
//

void *List::remove_nth(unsigned int k, bool *ok) {
    void *data;
    Listnode *n;
    unsigned int cur = 0;

    for (n = head; n != NULL; n = n->next) {
        if (cur == k){

            if (n->prev == NULL) {
                head = n->next;
            } else {
                n->prev->next = n->next;
            }

            if (n->next == NULL) {
                tail = n->prev;
            } else {
                n->next->prev = n->prev;
            }

            data = n->data;
            delete n;
            items--;

            if(ok) *ok = true;
            return data;
        }
        cur++;
    }

    if(ok) *ok = false;
    return NULL;
}


bool List::find(void *item, int *pos_ret) {
    Listnode *n;

    int i = 0;
    for (n = head; n != NULL; n = n->next) {
        if (n->data == item) {
            if (pos_ret) *pos_ret = i;
            return true;
        }

        i++;
    }	

    return false;
}

void *List::remove_head() {
    assert(head != NULL);
    assert(items > 0);

    Listnode *n = head;
    void *retval = n->data;

    if (head->next == NULL) {
        head = tail = NULL;
    } else {
        head = n->next;
        head->prev = NULL;
    }

    delete n;
    items--;
    return retval;
}


void *List::remove_tail() {
    assert(tail != NULL);
    assert(items > 0);

    Listnode *n = tail;
    void *retval = n->data;

    if (tail->prev == NULL) {
        head = tail = NULL;
    } else {
        tail = n->prev;
        tail->next = NULL;
    }

    delete n;
    items--;
    return retval;
}

void *List::peek_at_head() {
    assert(head != NULL);
    assert(items > 0);

    return head->data;
}

void *List::peek_at_tail() {
    assert(tail != NULL);
    assert(items > 0);

    return tail->data;
}


void List::clean() {
    Listnode *m, *n = head;

    m = n;
    while (n != NULL) {
        n = n->next;
        delete m;
        m = n;
    }

    head = NULL;
    tail = NULL;
    items = 0;
}


unsigned int List::total_items(void) {
    return items;
}


//void List::foreach(void (*proc)(void *data)) {
//    Listnode *n;
//    for (n = head; n != NULL; n = n->next) {
//        (*proc)(n->data);
//    }
//}

//void List::foreach(void (*proc)(void *data, void *client_data),
//		   void *client_data) {
//    Listnode *n;
//    for (n = head; n != NULL; n = n->next) {
//        (*proc)(n->data, client_data);
//    }
//}

void foreach_remove(List *l, Listnode *n) {
    if (n->prev) {
        n->prev->next = n->next;
    } else {
        l->head = n->next;
    }

    if (n->next) {
        n->next->prev = n->prev;
    } else {
        l->tail = n->prev;
    }

    l->items--;
    delete n;
}




void _auto_array_allocate(void **data_ret, int *items_ret,
                          int start_items, int item_size) {
    int bytes = start_items * item_size;
    void *data = malloc(bytes);
    *data_ret = data;
    *items_ret = start_items;
}

void _auto_array_deallocate(void **data_ret) {
    free(*data_ret);
}

void _auto_array_expand(void **data_ret, int *items_ret,
                        int item_size) {
  
    int old_items = *items_ret;
    int new_items = 2 * old_items;
    int old_bytes = old_items * item_size;
    int new_bytes = new_items * item_size;

    void *old_data = *data_ret;
    void *new_data = malloc(new_bytes);

    memcpy(new_data, old_data, old_bytes);
    free(old_data);

    *data_ret = new_data;
    *items_ret = new_items;
}




Hash_Table::Hash_Table(int _table_size) {
    table_size = _table_size;
    lists = new List[table_size];
    items = 0;
}

Hash_Table::~Hash_Table() {
    delete [] lists;
}

inline int Hash_Table::get_hash_code(int key) {
    if (key < 0) key = -key;
    return key % table_size;
}

struct Hash_Info {
    int key;
    void *ptr;
};

void Hash_Table::add(void *ptr, int key) {
    int code = get_hash_code(key);
    Hash_Info *info = new Hash_Info;
    info->key = key;
    info->ptr = ptr;

    lists[code].add(info);
    items++;
}

void *Hash_Table::remove(int key) {
    int code = get_hash_code(key);
    List *l = &lists[code];

    Hash_Info *info;
    Foreach(l, info) {
        if (info->key == key) {
            void *result = info->ptr;
            Foreach_Remove_Current_Element(l);
            delete info;
            items--;
            return result;
        }
    } Endeach;

    return NULL;
}

void *Hash_Table::find(int key) {
    int code = get_hash_code(key);
    List *l = &lists[code];

    Hash_Info *info;
    Foreach(l, info) {
        if (info->key == key) {
            info->ptr;
        }
    } Endeach;

    return NULL;
}

bool Hash_Table::remove(void *ptr, int key) {
    int code = get_hash_code(key);
    List *l = &lists[code];

    Hash_Info *info;
    Foreach(l, info) {
        if (info->ptr == ptr) {
            Foreach_Remove_Current_Element(l);
            delete info;
            items--;
            return true;
        }
    } Endeach;

    return false;
}


