#include "sound_internal.h"
#include "vorbis_decoder.h"

const int DO_MUSIC = 1;

struct Channel_Streamer {
    Channel_Streamer(Sound_Manager *manager);
    ~Channel_Streamer();

    Sound_Channel *channel;
    Vorbis_Decoder *decoder;
    Sound_Manager *my_manager;

    float decibel_change_per_second;

    bool started_streaming;
    bool done;
    bool first_append;
};


extern void os_set_high_priority();

unsigned long sound_manager_pump(void *arg) {
    Sound_Manager *manager = (Sound_Manager *)arg;
    int turn_index = 0;

    os_set_high_priority();

    while (1) {
        if (manager->should_exit) return 0;
        manager->update();

        os_sleep(10);
    }
}

Sound_Manager::Sound_Manager() {
    player = new Sound_Player_Win32();
    player->init();
    catalog = new Sound_Catalog();
    my_thread = new Thread(sound_manager_pump);
    my_lock = new Mutex();
    my_lock->init(false);
    should_exit = false;
    dummy_manager = false;
}

Sound_Manager::~Sound_Manager() {
    delete player;
    delete my_thread;
    delete my_lock;
}

void Sound_Manager::init() {
    if (DISABLE_SOUND) {
        dummy_manager = true;
    } else {
        bool success = player->driver_init();
        if (!success) dummy_manager = true;
    }

    my_thread->start(this);
}

void Sound_Manager::shutdown() {
    should_exit = true;
    my_thread->wait_for_exit();
    player->shutdown();
}

static Vorbis_Decoder *make_vorbis_decoder(Sound_Data *data) {
    Vorbis_Decoder *decoder = new Vorbis_Decoder();
    decoder->set_decode_source(data);

    return decoder;
}

void Sound_Manager::append(Sound_Channel *channel, Sound_Data *sound) {
    my_lock->lock();
    if (sound->type == Sound_Data::VORBIS_COMPRESSED) {
        Vorbis_Decoder *decoder = NULL;
        if (!DO_MUSIC) {
            my_lock->unlock();
            return;
        }

        decoder = make_vorbis_decoder(sound);
        if (decoder == NULL) {
            my_lock->unlock();
            return;
        }

        channel->flags |= Sound_Channel::DO_NOT_STOP_WHEN_STARVED;
        channel->nchannels = 2;
        Channel_Streamer *streamer = new Channel_Streamer(this);
        streamer->decoder = decoder;
        streamer->channel = channel;
        channel_streamers.add(streamer);
        update_channel_streamer(streamer);
    } else {
        channel->append_samples(sound->nsamples, sound->samples);
    }

    my_lock->unlock();
}

Sound_Channel *Sound_Manager::play_sound(Sound_Data *sound, float volume, float pan) {
    if (sound == NULL) return NULL;
    if (dummy_manager) return NULL;

	if (pan < -1) pan = -1;
	if (pan > 1) pan = 1;

    my_lock->lock();

    Sound_Channel *channel = player->create_channel();
    channel->base_volume = volume;
    channel->set_nchannels(sound->nchannels);
	channel->pan = pan;

    append(channel, sound);
    channel->play();

    my_lock->unlock();

    return channel;
}

Sound_Channel *Sound_Manager::create_persistent_channel() {
    if (dummy_manager) return NULL;

    my_lock->lock();
    Sound_Channel *channel = player->create_channel();
    my_lock->unlock();

    return channel;
}

void Sound_Manager::release_persistent_channel(Sound_Channel *channel) {
    if (dummy_manager) return;

    my_lock->lock();
    player->destroy_channel(channel);
    my_lock->unlock();
}

void Sound_Manager::update_channel_streamer(Channel_Streamer *streamer) {
    const int SLACK_TIME = 250;
    const int SAMPLES_TO_KEEP_BUFFERED = 44100 * 2 * 2;

    Vorbis_Decoder *decoder = streamer->decoder;
    Sound_Channel *channel = streamer->channel;

//    double dt = client_globals.time_info->current_dt;
//    double decibel_change = streamer->decibel_change_per_second * dt;
    
/*
    if (channel) {
        double new_volume = channel->volume + decibel_change;
        if (new_volume < 0) {
            streamer->done = true;
            return;
        }

        channel->volume_in_decibels = new_volume;
    }
*/

    if (streamer->started_streaming) {
        while (1) {
            int samples_present = channel->count_samples();
            if (samples_present && (channel->driver_channel_index == -1)) channel->play();
            if (samples_present >= SAMPLES_TO_KEEP_BUFFERED) return;

            int bytes_read = decoder->read(spurt_buf, SPURT_SIZE);

            if (bytes_read <= 0) {
                streamer->done = true;
                return;
            }

            if (bytes_read == 0) break;

            int samples_read = bytes_read / 2;
            short *samples = (short *)spurt_buf;

            channel->append_samples(samples_read, samples);
        }
    } else {
        streamer->started_streaming = true;
        decoder->begin_streaming();

        if (decoder->channels < 1) {
            // Empty file... ah well.
            streamer->done = true;
            assert(0);
            return;
        }

        update_channel_streamer(streamer);
    }
}

void Sound_Manager::play_sound(char *name, float volume, float pan) {
    if (dummy_manager) return;
    my_lock->lock();

    Sound_Data *sound = catalog->get_sound(name);
    
    if (sound == NULL) return;

    Sound_Channel *channel = play_sound(sound, volume, pan);
    channel->self_destruct_when_done();

    my_lock->unlock();
}

void Sound_Manager::acquire_lock() {
    my_lock->lock();
}

void Sound_Manager::release_lock() {
    my_lock->unlock();
}

#ifdef NOT
void Sound_Manager::fade_streaming_tracks(Sound_Category category,
                                          float db_per_second) {
    if (dummy_manager) return;

    my_lock->lock();

    Channel_Streamer *streamer;
    Foreach(&channel_streamers, streamer) {
        if (categories_match(streamer->category, category)) {
            streamer->decibel_change_per_second = -db_per_second;
        }
    } Endeach;

    my_lock->unlock();
}
#endif NOT

void Sound_Manager::get_debug_strings(List *l) {
    my_lock->lock();
    char buf[BUFSIZ];

    Sound_Channel *channel;
    Foreach(&player->channels, channel) {
        sprintf(buf, "chan %d pos %d vol %.3f",
                channel->driver_channel_index, channel->hardware_play_position,
                channel->base_volume);
		
        char *s = strdup(buf);
        l->add(s);
    } Endeach;

    my_lock->unlock();
}


void Sound_Manager::update() {
    if (dummy_manager) return;

    my_lock->lock();

    Channel_Streamer *streamer;
    Foreach(&channel_streamers, streamer) {
        update_channel_streamer(streamer);
        if (streamer->done) {
            Foreach_Remove_Current_Element(&channel_streamers);
            delete streamer;  // This will stop the sound channel.
        }
    } Endeach;

    player->update();

    my_lock->unlock();
}

Channel_Streamer::Channel_Streamer(Sound_Manager *manager) {
//    decoder = NULL;
    channel = NULL;
    my_manager = manager;
    done = false;
    started_streaming = false;
    first_append = true;
}

Channel_Streamer::~Channel_Streamer() {
    channel->stop();
    my_manager->player->destroy_channel(channel);
//    delete decoder;
}
