#include "sound_internal.h"

double SPEAKER_ATTENUATION_MAX = 16;

Sound_Channel::Sound_Channel(Sound_Player *player) {
    my_player = player;
    flags = 0;
    driver_channel_index = -1;
    nchannels = 1;
    total_samples_appended = 0;
    hardware_samples_spooled = 0;
    hardware_buffer_rollovers = 0;

    base_volume = 1;
	pan = 0;
    position_x = 0;
    position_y = 0;
    position_z = 0;
    distance_min = 130.0f;
    distance_max = 3000.0f;

    spatialized = false;

    hardware_play_position = 0;
    hardware_write_position = 0;
}

Sound_Channel::~Sound_Channel() {
}

void Sound_Channel::set_nchannels(int n) {
    nchannels = n;
}

void Sound_Channel::pause(bool should_pause) {
    assert(0);
    if (should_pause) {
        flags |= PAUSED;
    } else {
        flags &= ~PAUSED;
    }
}

void Sound_Channel::play() {
    my_player->driver_play(this);
    flags |= PLAYING;
}

void Sound_Channel::stop() {
    if (driver_channel_index != -1) {
        my_player->driver_stop(this);
    }

    flags &= ~PLAYING;
}

int Sound_Channel::append_to_last_buffer(int nsamples, short *samples) {
    if (sample_buffers.items == 0) return 0;
    Sound_Sample_Buffer *buffer = (Sound_Sample_Buffer *)(sample_buffers.tail->data);
    int room_left = SOUND_SAMPLE_BUFFER_SIZE - buffer->index_of_free_area;

    int nsamples_to_copy = nsamples;
    
    if (nsamples_to_copy > room_left) nsamples_to_copy = room_left;

#ifdef DDD
    int to_copy = nsamples_to_copy;
    short *dest_pos = buffer->samples + buffer->index_of_free_area;

    static int sample_index = 0;
    const float FREQUENCY = 513.5f;

    int i;
    for (i = 0; i < to_copy; i++) {
        float value = 32000 * cos(2 * M_PI * FREQUENCY * ((sample_index + i) / 44100.0));
        dest_pos[i] = (int)value;
    }
    sample_index += to_copy;
#endif DDD

#ifndef DDD
    memcpy(buffer->samples + buffer->index_of_free_area, samples, 
           nsamples_to_copy * sizeof(short));
#endif !DDD

    buffer->index_of_free_area += nsamples_to_copy;
    buffer->nsamples_to_consume += nsamples_to_copy;
    assert(buffer->index_of_free_area <= SOUND_SAMPLE_BUFFER_SIZE);

    return nsamples_to_copy;
}

void Sound_Channel::append_samples(int nsamples, short *samples) {

    int nshorts = nsamples;
    int orig_nsamples = nsamples;

    while (nsamples > 0) {
        int copied = append_to_last_buffer(nsamples, samples);
        if (copied == 0) {
            Sound_Sample_Buffer *buffer = new Sound_Sample_Buffer;
            buffer->nsamples_to_consume = 0;
            buffer->index_of_free_area = 0;
            buffer->consumer_cursor = buffer->samples;
            sample_buffers.add(buffer);
        }

        nsamples -= copied;
        samples += copied;
    }

    total_samples_appended += orig_nsamples;
}

void Sound_Channel::eliminate_samples() {
    Sound_Sample_Buffer *buffer;
    Foreach(&sample_buffers, buffer) {
        delete buffer;
        Foreach_Remove_Current_Element(&sample_buffers);
    } Endeach;

    flags |= Sound_Channel::PLEASE_SHAVE_SAMPLES;
}

void Sound_Channel::self_destruct_when_done() {
    flags |= SELF_DESTRUCT_WHEN_DONE;
}

int Sound_Channel::count_samples() {
    int sum = 0;
    Sound_Sample_Buffer *buffer;
    Foreach(&sample_buffers, buffer) {
        sum += buffer->nsamples_to_consume;
    } Endeach;

    return sum;
}


void Sound_Player::init() {
    listener_x = 0;
    listener_y = 0;
    listener_z = 0;
}

void Sound_Player::shutdown() {
}


double dot_to_speaker_attenuation(double dot) {
    if (dot >= 0.0) return 0;
    return -dot * SPEAKER_ATTENUATION_MAX;
}

void Sound_Player::update() {
    driver_update();

    List to_destroy;
    Sound_Channel *channel;
    Foreach(&channels, channel) {
        bool stop_sound = false;

        if (channel->flags & Sound_Channel::DRIVER_LEVEL_IS_DONE) stop_sound = true;
        if ((channel->driver_channel_index == -1) && (channel->flags & Sound_Channel::PLAYING)) stop_sound = true;

        if (stop_sound) {
            channel->flags &= ~Sound_Channel::DRIVER_LEVEL_IS_DONE;

            if (channel->flags & Sound_Channel::SELF_DESTRUCT_WHEN_DONE) {
                to_destroy.add(channel);
            } else {
                channel->flags &= ~Sound_Channel::PLAYING;
            }
        }
    } Endeach;

    Foreach(&to_destroy, channel) {
        destroy_channel(channel);
    } Endeach;
}

Sound_Channel *Sound_Player::create_channel() {
    Sound_Channel *channel = new Sound_Channel(this);
    channels.add(channel);
    return channel;
}

void Sound_Player::destroy_channel(Sound_Channel *channel) {
    channel->stop();
    channels.remove(channel);
    delete channel;
}

int hash_string(char *s) {
    int value = 0;
    while (*s) {
        value += *s;
        value *= 127;

        s++;
    }

    return value;
}
