#include "sound_internal.h"

#define STRICT
#include <windows.h>
#include <dsound.h>
#include <objbase.h>
#include <commdlg.h>
#include <mmreg.h>
#include <math.h> // XXX for debugging (calling 'cos')

#pragma warning(disable: 4244)

//const int CHANNEL_BUFFER_SIZE = 16384 * 4;
const int CHANNEL_BUFFER_SIZE = 8192 * 2;
const double LATE_STAGE_DECIBEL_DISPLACEMENT_MAX = 50;

#define SAFE_DELETE(p)  { if(p) { delete (p);     (p)=NULL; } }
#define SAFE_RELEASE(p) { if(p) { (p)->Release(); (p)=NULL; } }

struct Win32_Channel {
    void reset();

    IDirectSoundBuffer *dsound_buffer;
    int length_in_bytes;
    int driver_channel_index;
    int nchannels;

    unsigned long flags;
    int byte_position_for_writing;
    int silent_samples_written;
    short last_written_sample;

    enum {
        ALLOCATED = 0x1,
        SLOW_STOPPING = 0x2,
        PLAYING_SILENCE = 0x4
    };
};


void os_set_high_priority() {
    BOOL res;
    HANDLE my_proc;
    HANDLE my_thread;

    my_proc = GetCurrentProcess();
//    res = SetPriorityClass(my_proc, HIGH_PRIORITY_CLASS);
//    if(res!=TRUE) fprintf(stderr, "Couldn't get HIGH_PRIORITY_CLASS...\n");

    my_thread = GetCurrentThread();
    res = SetThreadPriority(my_thread,THREAD_PRIORITY_HIGHEST);
    if(res!=TRUE) fprintf(stderr,"Couldn't get THREAD_PRIORITY_HIGHEST\n");
}


bool Sound_Player_Win32::driver_init() {
    driver_initialized = false;

    HRESULT hr;

    // Create IDirectSound using the primary sound device
    if (FAILED(hr = DirectSoundCreate(NULL, &dsound, NULL)))
        return false;

    HWND hwnd = GetForegroundWindow();
    if (hwnd == NULL) hwnd = GetDesktopWindow();
    if (hwnd == NULL) return false;

    //    if (hr = CoInitialize(NULL)) return hr;

    // Set coop level to DSSCL_PRIORITY
    if (FAILED(hr = dsound->SetCooperativeLevel(hwnd, DSSCL_PRIORITY)))
        return false;

    LPDIRECTSOUNDBUFFER primary_buffer = NULL;

    // Get the primary buffer 
    DSBUFFERDESC dsbd;
    ZeroMemory( &dsbd, sizeof(DSBUFFERDESC) );
    dsbd.dwSize        = sizeof(DSBUFFERDESC);
    dsbd.dwFlags       = DSBCAPS_PRIMARYBUFFER;
    dsbd.dwBufferBytes = 0;
    dsbd.lpwfxFormat   = NULL;
       
    if (FAILED (hr = dsound->CreateSoundBuffer(&dsbd, &primary_buffer, NULL)))
        return false;

    WAVEFORMATEX wfx;
    ZeroMemory(&wfx, sizeof(WAVEFORMATEX));
    wfx.wFormatTag      = WAVE_FORMAT_PCM; 
    wfx.nChannels       = 2; 
    wfx.nSamplesPerSec  = 44100; 
    wfx.wBitsPerSample  = 16;
    wfx.nBlockAlign     = (wfx.wBitsPerSample / 8) * wfx.nChannels;
    wfx.nAvgBytesPerSec = wfx.nSamplesPerSec * wfx.nBlockAlign;

    if (FAILED (hr = primary_buffer->SetFormat(&wfx)))
        return false;

    SAFE_RELEASE(primary_buffer);

    bool success = create_static_buffers();
    if (!success) return false;

    driver_initialized = true;
    return true;
}

bool Sound_Player_Win32::driver_shutdown() {
    destroy_static_buffers();

    SAFE_RELEASE(dsound); 

    driver_initialized = false;
    return true;
}

bool Sound_Player_Win32::create_static_buffers() {
    int NUM_MUSIC_CHANNELS = 4;

    int i;
    for (i = 0; i < MAX_SOUND_CHANNELS; i++) {
        WAVEFORMATEX wfx;
        ZeroMemory( &wfx, sizeof(WAVEFORMATEX) ); 
        wfx.wFormatTag      = WAVE_FORMAT_PCM; 
        wfx.nSamplesPerSec  = 44100; 
        wfx.wBitsPerSample  = 16;

        int nchannels;
        if (i < MAX_SOUND_CHANNELS - NUM_MUSIC_CHANNELS) {
            nchannels = 1;
        } else {
            nchannels = 2;
        }

        wfx.nChannels = nchannels;
        wfx.nBlockAlign     = (wfx.wBitsPerSample / 8) * wfx.nChannels;
        wfx.nAvgBytesPerSec = wfx.nSamplesPerSec * wfx.nBlockAlign;

        DSBUFFERDESC dsbd;
        ZeroMemory(&dsbd, sizeof(DSBUFFERDESC));
        dsbd.dwSize        = sizeof(DSBUFFERDESC);
        dsbd.dwFlags       = DSBCAPS_GETCURRENTPOSITION2
            | DSBCAPS_CTRLPAN
            | DSBCAPS_CTRLVOLUME
            | DSBCAPS_GLOBALFOCUS
            | DSBCAPS_STATIC;
//            | DSBCAPS_LOCSOFTWARE;
        dsbd.dwBufferBytes = CHANNEL_BUFFER_SIZE;
        dsbd.lpwfxFormat   = &wfx;

        IDirectSoundBuffer *dsound_buffer;

        HRESULT hr;
        hr = dsound->CreateSoundBuffer(&dsbd, &dsound_buffer, NULL);
        if (FAILED(hr)) break;

        Win32_Channel *channel = new Win32_Channel();
        win32_channels[i] = channel;
        channel->nchannels = nchannels;
        channel->driver_channel_index = i;
        channel->length_in_bytes = dsbd.dwBufferBytes;
        channel->dsound_buffer = dsound_buffer;
        channel->reset();
    }

    num_driver_channels = i;

    if (i == 0) return false;
    return true;
}

void Sound_Player_Win32::destroy_static_buffers() {
    int i;
    for (i = 0; i < num_driver_channels; i++) {
        Win32_Channel *channel = win32_channels[i];
        SAFE_RELEASE(channel->dsound_buffer);
        delete channel;
    }
}

inline Win32_Channel *Sound_Player_Win32::get_win32_channel(Sound_Channel
                                                            *channel) {
    int index = channel->driver_channel_index;
    if (index == -1) return NULL;

    assert(index >= 0);
    assert(index < num_driver_channels);

    Win32_Channel *win32_channel = win32_channels[index];
    return win32_channel;
}

int Sound_Player_Win32::do_normal_copy(short *dest_pos, int dest_samples,
                                       Sound_Channel *channel,
                                       Win32_Channel *win32_channel) {

    if (channel->sample_buffers.items == 0) {
        Win32_Channel *win32_channel = get_win32_channel(channel);
        assert(win32_channel);
        if ((channel->total_samples_appended > 0) && !(channel->flags & Sound_Channel::DO_NOT_STOP_WHEN_STARVED)) {
            initiate_slow_stop(win32_channel);
            return 0;
        } else {
            int i;
            for (i = 0; i < dest_samples; i++) {
                dest_pos[i] = 0;
            }

            return dest_samples;
        }
    }

    Sound_Sample_Buffer *buffer = (Sound_Sample_Buffer *)channel->sample_buffers.nth(0);
	
    int src_samples = buffer->nsamples_to_consume;
    short *src_pos = buffer->consumer_cursor;

    int to_copy = src_samples;
    if (to_copy > dest_samples) to_copy = dest_samples;

    memcpy(dest_pos, src_pos, sizeof(short) * to_copy);
    win32_channel->last_written_sample = src_pos[to_copy - 1];

    src_pos += to_copy;

    buffer->consumer_cursor = src_pos;
    buffer->nsamples_to_consume -= to_copy;

    if (buffer->nsamples_to_consume == 0) {
        channel->sample_buffers.remove_head();
        delete buffer;
    }
    
    return to_copy;
}

int Sound_Player_Win32::do_silence_fill(short *dest_pos, int dest_samples,
                                        Win32_Channel *win32_channel) {
    int i;
    for (i = 0; i < dest_samples; i++) {
        dest_pos[i] = 0;
    }

    win32_channel->silent_samples_written += dest_samples;
    return dest_samples;
}

int Sound_Player_Win32::do_slow_stop(short *dest_pos, int dest_samples,
                                     Win32_Channel *channel) {
    int i;
    for (i = 0; i < dest_samples; i++) {
        dest_pos[i] = 0;
    }

    channel->flags &= ~Win32_Channel::SLOW_STOPPING;
    channel->flags |= Win32_Channel::PLAYING_SILENCE;

    channel->last_written_sample = 0;

    Sound_Channel *sound_channel = get_sound_channel(channel);
    if (sound_channel) sound_channel->hardware_buffer_rollovers = 0;

    return dest_samples;
}

int round_to_int(double f) {
    if (f >= 0) {
        return (int)floor(f + 0.5);
    } else {
        return -(int)floor(-f + 0.5);
    }
}

void Sound_Player_Win32::adjust_spatialization(Sound_Channel *channel,
                                               Win32_Channel *win32_channel) {
    if (channel == NULL) return;

    float volume = channel->base_volume;

    if (channel->spatialized) {
        float dx = channel->position_x - listener_x;
        float dy = channel->position_y - listener_y;
        float dz = channel->position_z - listener_z;

        float dist = sqrt(dx * dx + dy * dy + dz * dz);
        if (dist < channel->distance_min) dist = channel->distance_min;
        if (dist > channel->distance_max) dist = channel->distance_max;

        float perc;
        float denominator = channel->distance_min - channel->distance_max;
        if (denominator == 0) {
            perc = 1;
        } else {
            perc = (dist - channel->distance_max) / denominator;
        }

        if (perc < 0) perc = 0;
        if (perc > 1) perc = 1;

        volume *= perc;
    }

    // DirectSound wants volume in hundredths of decibels.
    // It sucks.

	// compensate for their stupid log scale
	const float	LN_BASE = 2.718281828f;
    LONG cdb;
	if (volume < 0.0001) {
		cdb = DSBVOLUME_MIN;
	} else {
		// -50db is about as quiet as we care about.
		cdb = -5000 + 5000 * log(1 + (LN_BASE - 1) * volume);
	}

#if 0
    LONG max = DSBVOLUME_MAX;
    LONG min = DSBVOLUME_MIN;

    float fvol = volume * (max - min) + min;
    LONG cdb = round_to_int(fvol);

    if (cdb > max) cdb = max;
    if (cdb < min) cdb = min;
#endif // 0
    
    HRESULT hr;
    hr = win32_channel->dsound_buffer->SetVolume(cdb);

	// Set pan.
	// -5000 to 5000
	long	pan = 0;
	if (channel->pan < 0) {
		pan = -5000 * log(1 + (LN_BASE - 1) * -channel->pan);
	} else if (channel->pan > 0) {
		pan = 5000 * log(1 + (LN_BASE - 1) * channel->pan);
	}
	win32_channel->dsound_buffer->SetPan(pan);
}

int get_samples_between(int front, int rear, int buffer_len) {
    if (front > rear) return (front - rear) / 2;
    return (front + (buffer_len - rear)) / 2;
}

void Sound_Player_Win32::cycle_samples(Win32_Channel *win32_channel,
                                       bool first) {
    HRESULT hr;
    void *pointer1;
    DWORD bytes1;
    void *pointer2;
    DWORD bytes2;
    
    Sound_Channel *channel = get_sound_channel(win32_channel);
    adjust_spatialization(channel, win32_channel);

    IDirectSoundBuffer *dsound_buffer = win32_channel->dsound_buffer;
    
    DWORD play_position;
    DWORD safe_write_position;
    hr = dsound_buffer->GetCurrentPosition(&play_position, &safe_write_position);

    if (play_position < (DWORD) channel->hardware_play_position) {
        channel->hardware_buffer_rollovers++;
    }

    channel->hardware_play_position = play_position;
    channel->hardware_write_position = safe_write_position;

    if (channel->flags & Sound_Channel::PLEASE_SHAVE_SAMPLES) {
        dsound_buffer->SetCurrentPosition(safe_write_position);
        win32_channel->byte_position_for_writing = safe_write_position;
    }

    DWORD write_position = win32_channel->byte_position_for_writing;
    int samples_ahead = get_samples_between(write_position, safe_write_position,
                                            win32_channel->length_in_bytes);

    int max_fill_ahead = channel->nchannels * 44100 * 0.1f;
    
    int samples_desired = max_fill_ahead - samples_ahead;
    if (samples_desired <= 0) return;

    hr = dsound_buffer->Lock(write_position, samples_desired * 2,
                             &pointer1, &bytes1, &pointer2, &bytes2, 0);
    if (FAILED(hr)) return;

//    bytes1 &= ~1; // Wtf does it return an odd pointer?? Jeez

    int limit;
    int pre_play = play_position - sizeof(short) * win32_channel->nchannels;
    if (write_position >= safe_write_position) {
        limit = win32_channel->length_in_bytes;
        if ((pre_play >= (int)safe_write_position) && (pre_play < limit)) limit = pre_play;
    } else {
        limit = pre_play;
    }

    // For now we only fill 'bytes1'.  In the future we may try
    // to hit both buffers.

    int nbytes = limit - write_position;
    if (nbytes > (int) bytes1) nbytes = bytes1;
    if (nbytes > samples_desired * 2) nbytes = samples_desired * 2;

    assert(!(nbytes & 1));
    assert(nbytes >= 0);

    int dest_samples = nbytes / sizeof(short);
    short *dest_pos = (short *)pointer1;

    bool stop_sound = false;

    while (dest_samples) {
        int samples_filled;

        if (win32_channel->flags & Win32_Channel::SLOW_STOPPING) {
            samples_filled = do_slow_stop(dest_pos, dest_samples, win32_channel);
        } else if (win32_channel->flags & Win32_Channel::PLAYING_SILENCE) {
            samples_filled = do_silence_fill(dest_pos, dest_samples, 
                                             win32_channel);
            if (channel->hardware_buffer_rollovers > 0) {
                stop_sound = true;
            }
        } else {
            samples_filled = do_normal_copy(dest_pos, dest_samples, 
                                            channel, win32_channel);
        }

        dest_samples -= samples_filled;
        dest_pos += samples_filled;
        win32_channel->byte_position_for_writing += samples_filled * 2;
        channel->hardware_samples_spooled += samples_filled;
    }

    dsound_buffer->Unlock(pointer1, bytes1, pointer2, bytes2);


    if (win32_channel->byte_position_for_writing >= win32_channel->length_in_bytes) {
        assert(win32_channel->byte_position_for_writing == win32_channel->length_in_bytes);
        win32_channel->byte_position_for_writing = 0;
    }

    unsigned int pos = win32_channel->byte_position_for_writing;
    if (safe_write_position >= play_position) {
        assert((pos <= play_position) || (pos >= safe_write_position));
    } else {
        assert((pos >= safe_write_position) && (pos <= play_position));
    }

    if (stop_sound) {
        Sound_Channel *channel = get_sound_channel(win32_channel);
        channel->driver_channel_index = -1;
        channel->flags |= Sound_Channel::DRIVER_LEVEL_IS_DONE;

        driver_stop(win32_channel);
    }
}

void Sound_Player_Win32::driver_update() {
    bool success = restore_buffers();
    if (!success) return;

    int i;
    for (i = 0; i < num_driver_channels; i++) {
        Win32_Channel *win32_channel = win32_channels[i];
        if (win32_channel->flags & Win32_Channel::ALLOCATED) {
            cycle_samples(win32_channel, false);
        }
    }
}

bool Sound_Player_Win32::driver_play(Sound_Channel *channel) {

    if (channel->driver_channel_index != -1) return true;

    if (channel->driver_channel_index == -1) {
        channel->driver_channel_index = 
            driver_get_buffer_id(channel->nchannels);
        if (channel->driver_channel_index == -1) return false;
    }

    Win32_Channel *win32_channel = get_win32_channel(channel);
    if (win32_channel == NULL) return false;

    IDirectSoundBuffer *dsound_buffer = win32_channel->dsound_buffer;
    HRESULT hr;
    hr = dsound_buffer->Stop();
//    hr = dsound_buffer->SetCurrentPosition(0);

// START LAMENESS
    void *pointer1;
    DWORD bytes1;
    void *pointer2;
    DWORD bytes2;
    dsound_buffer->Restore();
    hr = dsound_buffer->Lock(0, win32_channel->length_in_bytes,
                             &pointer1, &bytes1, &pointer2, &bytes2,
                             DSBLOCK_ENTIREBUFFER);
    if (FAILED(hr)) return false;

    if (pointer1) memset(pointer1, 0, bytes1);
    if (pointer2) memset(pointer2, 0, bytes2);

    dsound_buffer->Unlock(pointer1, bytes1, pointer2, bytes2);
// END LAMENESS

    hr = dsound_buffer->SetCurrentPosition(0);

    cycle_samples(win32_channel, true);

    adjust_spatialization(channel, win32_channel);
    hr = dsound_buffer->Play(0, 0, DSBPLAY_LOOPING);
    adjust_spatialization(channel, win32_channel);

    if (FAILED(hr)) return false;
    return true;
}

bool Sound_Player_Win32::driver_stop(Sound_Channel *channel) {
    Win32_Channel *win32_channel = get_win32_channel(channel);
    if (win32_channel == NULL) return true;

    bool result = driver_stop(win32_channel);
    if (result) {
        channel->flags |= Sound_Channel::DRIVER_LEVEL_IS_DONE;
        channel->driver_channel_index = -1;
        return true;
    }

    return false;
}

bool Sound_Player_Win32::driver_stop(Win32_Channel *channel) {
    IDirectSoundBuffer *dsound_buffer = channel->dsound_buffer;
    HRESULT hr = dsound_buffer->Stop();
    if (FAILED(hr)) return false;

    driver_release_buffer_id(channel->driver_channel_index);

    return true;
}

void Sound_Player_Win32::initiate_slow_stop(Win32_Channel *channel) {
    if (!(channel->flags & (Win32_Channel::SLOW_STOPPING | Win32_Channel::PLAYING_SILENCE))) {
        channel->flags |= Win32_Channel::SLOW_STOPPING;
    }
}

bool Sound_Player_Win32::driver_slow_stop(Sound_Channel *channel) {
    Win32_Channel *win32_channel = get_win32_channel(channel);
    if (win32_channel == NULL) return true;

    initiate_slow_stop(win32_channel);
    return true;
}

Sound_Channel *Sound_Player_Win32::get_sound_channel(Win32_Channel *win32_channel) {
    Sound_Channel *channel;
    Foreach(&channels, channel) {
        if (channel->driver_channel_index == win32_channel->driver_channel_index) {
            return channel;
        }
    } Endeach;

    return NULL;
}

int Sound_Player_Win32::driver_get_buffer_id(int nchannels) {
/*
    int start = rand() % num_driver_channels;
    Win32_Channel *channel = win32_channels[start];
    if (channel->nchannels == nchannels) {
        if (!(channel->flags & Win32_Channel::ALLOCATED)) {
            channel->reset();
            channel->flags |= Win32_Channel::ALLOCATED;
            return start;
        }
    }
*/

    int i;
    for (i = 0; i < num_driver_channels; i++) {
        Win32_Channel *channel = win32_channels[i];
        if (channel->nchannels != nchannels) continue;
        if (!(channel->flags & Win32_Channel::ALLOCATED)) {
            channel->reset();
            channel->flags |= Win32_Channel::ALLOCATED;
            break;
        }
    }

    if (i < num_driver_channels) return i;

    return -1;
}

bool Sound_Player_Win32::driver_release_buffer_id(int id) {
    if (id == -1) return true;

    Win32_Channel *channel = win32_channels[id];
    assert(channel->flags & Win32_Channel::ALLOCATED);
    channel->flags &= ~Win32_Channel::ALLOCATED;
    return true;
}

void Sound_Player_Win32::init() {
    driver_initialized = false;
    dsound = NULL;

    Sound_Player::init();
}

void Sound_Player_Win32::shutdown() {
    if (driver_initialized) driver_shutdown();
    Sound_Player::shutdown();
}

bool Sound_Player_Win32::restore_buffers() {
    int channels_okay = 0;
    bool try_again = false;

    int i;
    for (i = 0; i < num_driver_channels; i++) {
        HRESULT hr;
        DWORD dwStatus;

        Win32_Channel *win32_channel = win32_channels[i];
        IDirectSoundBuffer *dsound_buffer = win32_channel->dsound_buffer;
	
        hr = dsound_buffer->GetStatus(&dwStatus);
        if (FAILED(hr)) continue;

        if (dwStatus & DSBSTATUS_BUFFERLOST) {
            // Since the app could have just been activated, then
            // DirectSound may not be giving us control yet, so 
            // the restoring the buffer may fail.  
            // If it does, sleep until DirectSound gives us control.

            hr = dsound_buffer->Restore();
            if (hr == DSERR_BUFFERLOST) try_again = true;
        }

        channels_okay++;
    }

    if (try_again) {
        os_sleep(50);
		return restore_buffers();
    }

    if (channels_okay == num_driver_channels) return true;
    return false;
}

void Win32_Channel::reset() {
    flags = 0;
    byte_position_for_writing = 0;
    silent_samples_written = 0;
    last_written_sample = 0;
}

void os_sleep(int time) {
    Sleep(time);
}
