#include "sound_internal.h"
#include "tagmap.h"

struct oneway_mapping {
    char *name;
    int value;
    int environment_index;
    oneway_mapping *next;

    oneway_mapping(char *name, int value, int environment,
                   bool alloc);
    ~oneway_mapping();
};


oneway_mapping::oneway_mapping(char *_name, int _value, int env,
                               bool alloc) {
    if (alloc) {
        name = strdup(_name);
    } else {
        name = _name;
    }

    value = _value;
    environment_index = env;
}

oneway_mapping::~oneway_mapping() {
}


Oneway_Tagmap::Oneway_Tagmap(int size, bool allocate) {
    next_environment_handle = 0;
    table_size = size;
    allocate_names = allocate;

    mapping_table = new oneway_mapping *[table_size];

    int i;
    for (i = 0; i < table_size; i++) {
        mapping_table[i] = NULL;
    }
}

Oneway_Tagmap::~Oneway_Tagmap() {
    // XXX do this 
}

inline int hash_string(char *s, int table_size) {
    int code = 0;

    while (*s) {
        code += (*s) * 17;
        s++;
    }

    code %= table_size;
    return code;
}

int Oneway_Tagmap::get_environment_handle() {
    return next_environment_handle++;
}

void Oneway_Tagmap::retract_environment(int handle) {
    int i;

    for (i = 0; i < table_size; i++) {
        oneway_mapping *cursor = mapping_table[i];
        oneway_mapping **ptr = &mapping_table[i];

        while (cursor) {
            oneway_mapping *next = cursor->next;

            if (cursor->environment_index == handle) {
                *ptr = cursor->next;
                if (allocate_names) free(cursor->name);
                delete cursor;
            } else {
                ptr = &(cursor->next);
            }

            cursor = next;
        }
    }
}

void Oneway_Tagmap::add(char *s, int value, int env) {
    oneway_mapping *mapping = new oneway_mapping(s, value, env,
                                                 allocate_names);

    int hash = hash_string(s, table_size);
    oneway_mapping *cursor = mapping_table[hash];
    oneway_mapping **ptr = &mapping_table[hash];

    while (1) {
        if (cursor != NULL) {
            if (env < cursor->environment_index) {
                ptr = &(cursor->next);
                cursor = cursor->next;
            }
        }

        *ptr = mapping;
        mapping->next = cursor;
        return;
    }
}

bool Oneway_Tagmap::remove(char *s) {
    int removed = 0;

    int hash = hash_string(s, table_size);
    oneway_mapping *cursor = mapping_table[hash];
    oneway_mapping **last_ptr = &mapping_table[hash];

    while (cursor) {
        if (strcmp(cursor->name, s) == 0) {
            removed++;
            *last_ptr = cursor->next;

            oneway_mapping *tmp = cursor;
            cursor = cursor->next;

            // XXX This will leak if we allocated copies of the
            // entry names for this tagmap.

            delete tmp;
        } else {
            last_ptr = &cursor->next;
            cursor = cursor->next;
        }
    }
    
    return removed ? true : false;
}

bool Oneway_Tagmap::lookup(char *s, int *value_ret) {
    int hash = hash_string(s, table_size);
    oneway_mapping *cursor = mapping_table[hash];

    while (cursor) {
        if (strcmp(cursor->name, s) == 0) {
            *value_ret = cursor->value;
            return true;
        }

        cursor = cursor->next;
    }

    return false;
}

/*
bool Oneway_Tagmap::lookup(int value, char **string_ret) {
    oneway_mapping *mapping;

    Foreach(&mappings, mapping, {
	if (mapping->value == value) {
	    *string_ret = mapping->name;
	    return true;
	}
    });

    return false;
}
*/


List *load_string_table(char *filename) {
    FILE *f = fopen(filename, "r");
    if (f == NULL) return NULL;

    List *result = new List();

    char buf[BUFSIZ];
    char *s;

    while (1) {
        s = fgets(buf, BUFSIZ, f);
        if (s == NULL) break;

        int value, chars_eaten;
        int success = sscanf(buf, "%d:%n", &value, &chars_eaten);

        char *name = (buf + chars_eaten);

        int len = strlen(name);
        if (len == 0) continue;
        name[len - 1] = '\0';

        result->add(strdup(name));
    }

    fclose(f);
    return result;
}

bool save_string_table(char *filename, List *table) {
    FILE *f = fopen(filename, "w");
    if (f == NULL) return false;

    int k = 0;
    char *name;
    Foreach(table, name) {
        fprintf(f, "%d:%s\n", k, name);
        k++;
    } Endeach;

    fclose(f);
    return true;
}

Crossmap::Crossmap(List *inputs, Oneway_Tagmap *outputs) {
    int n = inputs->items;
    xref_table = new int[n];
    valid_table = new bool[n];

    int i;
    for (i = 0; i < n; i++) {
        valid_table[i] = false;
    }

    char *s;
    int m = 0;

    Foreach(inputs, s) {
        int value;
        bool success = outputs->lookup(s, &value);

        if (success) {
            valid_table[m] = true;
            xref_table[m] = value;
        }

        m++;
    } Endeach;

    table_size = n;
}

Crossmap::~Crossmap() {
    delete [] xref_table;
    delete [] valid_table;
}

bool Crossmap::xref(int input, int *output_ret) {
    if (input < 0) return false;
    if (input >= table_size) return false;

    if (valid_table[input] == false) return false;

    *output_ret = xref_table[input];
    return true;
}


String_Table_Builder::String_Table_Builder() {
    entries = new List();
}

String_Table_Builder::~String_Table_Builder() {
    delete entries;
}

int String_Table_Builder::add(int n) {
    int m;
    int index;

    index = 0;
    Foreach(entries, m) {
        if (m == n) return index;
        index++;
    } Endeach;

    index = entries->items;  // Actually it's already entries->items.
    entries->add((void *)n);
    return index;
}

bool String_Table_Builder::lookup(int n, int *result) {
    int m;
    int index;

    index = 0;
    Foreach(entries, m) {
        if (m == n) {
            *result = index;
            return true;
        }

        index++;
    } Endeach;

    return false;
}

List *String_Table_Builder::build_string_table(char *(*proc)(int)) {
    int n;
    List *result = new List();

    Foreach(entries, n) {
        char *name = proc(n);
        result->add(strdup(name));
    } Endeach;

    return result;
}

void String_Table_Builder::serialize(Concatenator *c,
                                     char *(*proc)(int)) {

    c->add_u2b(entries->items);

    int n;
    Foreach(entries, n) {
        char *name = proc(n);
        c->add_asciiz(name);
    } Endeach;
}

void free_string_table(List *table) {
    char *s;
    Foreach(table, s) { 
        free(s); 
    } Endeach;
    delete table;
}
