/* ========================================================================
   $RCSfile: sprite_texture.cpp,v $
   $Date: 2002/03/14 01:08:43 $
   $Revision: 1.15 $
   $Creator: Casey Muratori $
   $Notice: $
   ======================================================================== */

#define WIN32_LEAN_AND_MEAN
#define WIN32_EXTRA_LEAN
#include <windows.h>
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#if !defined(SPRITE_TEXTURE_H)
#include "sprite_texture.h"
#endif

#if !defined(MEMORY_UTILS_H)
#include "memory_utils.h"
#endif

#if !defined(BMP_H)
#include "bmp.h"
#endif

#if !defined(LOG_H)
#include "log.h"
#endif

#if !defined(MEMORY_ARENA_H)
#include "memory_arena.h"
#endif

#if !defined(IMAGE_OPERATIONS_H)
#include "image_operations.h"
#endif

int32 const GlobalTransparentIndex = 0xff;      // because I modified deutex.exe to make this so

static void AddPixels(sprite_texture &State,
                      int32 ImageWidth, int32 ImageHeight, void *Pixels,
                      int &packed_u0_v0,
                      int &packed_u1_v0,
                      int &packed_u1_v1,
                      int &packed_u0_v1,
                      uint16 &XMin, uint16 &YMin,
                      uint16 &XMax, uint16 &YMax,
                      real32 &HalfWidth, real32 &FullHeight,
                      float x_scale, float y_scale)
{
    assert(Pixels);
    
    // hit 'return' if necessary
    if((State.CurrentX + ImageWidth) >= State.TextureDimension)
    {
        State.CurrentX = 0;
        State.CurrentY += State.CurrentRowHeight;
        // start a new row of the current image height
        State.CurrentRowHeight = ImageHeight;
    }
    else
    {
        // make sure we're tracking the tallest sprite in this row
        State.CurrentRowHeight = max(State.CurrentRowHeight,ImageHeight);
    }

    assert((State.CurrentX + ImageWidth) < State.TextureDimension);
    assert((State.CurrentY + ImageHeight) < State.TextureDimension);

    // compute texture coordinates
    XMin = State.CurrentX;
    YMin = State.CurrentY;
    XMax = (State.CurrentX + ImageWidth);
    YMax = (State.CurrentY + ImageHeight);

#define toShort(f)   ((int16) ((f) * 32768))
    
    int16 UMin = toShort((real32)XMin / (real32)State.TextureDimension);
    int16 VMin = toShort((real32)YMin / (real32)State.TextureDimension);
    int16 UMax = toShort((real32)XMax / (real32)State.TextureDimension);
    int16 VMax = toShort((real32)YMax / (real32)State.TextureDimension);

    ((int16 *)(&packed_u0_v0))[0] = UMin;
    ((int16 *)(&packed_u0_v0))[1] = VMin;
    ((int16 *)(&packed_u1_v0))[0] = UMax;
    ((int16 *)(&packed_u1_v0))[1] = VMin;
    ((int16 *)(&packed_u0_v1))[0] = UMin;
    ((int16 *)(&packed_u0_v1))[1] = VMax;
    ((int16 *)(&packed_u1_v1))[0] = UMax;
    ((int16 *)(&packed_u1_v1))[1] = VMax;

    HalfWidth = (XMax - XMin) / 24.0 * x_scale;
    FullHeight = (YMax - YMin) / 12.0 * y_scale;

    // do the copy
    CopyStrided(ImageWidth * 4, ImageHeight,
                ImageWidth * 4, Pixels,
                State.TextureDimension * 4,
                &State.Pixels[State.CurrentX +
                             State.CurrentY * State.TextureDimension]);
    
    // update cursor
    State.CurrentX += ImageWidth;
}

void MirrorImage(int32 Width, int32 Height, uint32 *Pixels)
{
    while(Height--)
    {
        int32 WidthCount = Width/2;
        uint32 *Source = Pixels;
        uint32 *Dest = Pixels + (Width - 1);
        while(WidthCount--)
        {
            uint32 Temp = *Source;
            *Source = *Dest;
            *Dest = Temp;
            
            ++Source;
            --Dest;
        }

        Pixels += Width;
    }
}

FILE *FindAndOpenUniqueRegex( char const *FileNameRegex )
{
    FILE *File = 0;

    // figure out the path and file part of the string because WIN32_FIND_DATA sucks
    char tmp[MAX_PATH];
    sprintf(tmp,FileNameRegex);
    char Path[MAX_PATH];
    char *FileNamePart = 0;
    GetFullPathName(tmp,sizeof(Path),Path,&FileNamePart);
    assert(FileNamePart);

    WIN32_FIND_DATA FindData;
    HANDLE Handle = FindFirstFile(Path,&FindData);
    if(Handle != INVALID_HANDLE_VALUE)
    {
        if(!(FindData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) &&
           !FindNextFile(Handle,&FindData))
        {
            // okay, we found a file and it's unique
            // luckily, the FindData makes it hacky to do this
            // because the cFileName is just the file, and there's
            // no way to get the directory part from it!!!!
            strcpy(FileNamePart,FindData.cFileName);
            File = fopen(Path,"rb");
        }
        FindClose(Handle);
    }
    return File;    
}
bmp *FindAndReadDoomFrame( char const *DoomPrefix, char FrameLetter, int32 ImageNumber )
{
    struct doom_sprite_name_possibilities
    {
        char *fmt;
        bool Mirror;
    } DoomNames[] = {
        "%s%c%d.bmp",   false,
        "%s%c%d??.bmp", false,
        "%s??%c%d.bmp", true,
    };
    int const DoomNamesCount = (sizeof(DoomNames)/sizeof(DoomNames[0]));

    FILE *File = 0;
    bool Mirror = false;
    for(int i = 0;i < DoomNamesCount; ++i)
    {
        char Path[MAX_PATH];
        sprintf(Path,DoomNames[i].fmt,DoomPrefix,FrameLetter,ImageNumber);
        File = FindAndOpenUniqueRegex(Path);
        if(File)
        {
            Mirror = DoomNames[i].Mirror;
            break;
        }
    }
    bmp *Bitmap = 0;
    if(File)
    {
        Bitmap = ReadBMP_FILE(File,GlobalTransparentIndex,DoomPrefix);
        if(Bitmap && Mirror)
        {
            MirrorImage(Bitmap->Width,Bitmap->Height,Bitmap->Pixels);
        }
        fclose(File);
    }
    return Bitmap;
}

static void
MakeSpriteImageFromDoom(sprite_texture &State, sprite_image &Image,
                        char const *DoomPrefix, char FrameLetter,
                        int32 ImageNumber, float x_scale, float y_scale)
{
    bmp *Bitmap = FindAndReadDoomFrame(DoomPrefix,FrameLetter,ImageNumber);
    if(Bitmap)
    {
        // @todo need some slop for mipmapping? (it doesn't seem so visually, alphatest saving us)?
        AddPixels(State, Bitmap->Width, Bitmap->Height, Bitmap->Pixels,
                  Image.packed_u0v0, Image.packed_u1v0, Image.packed_u1v1, Image.packed_u0v1,
                  Image.XMin, Image.YMin, Image.XMax, Image.YMax,
                  Image.HalfWidth, Image.FullHeight, x_scale, y_scale);
        FreeBMP(Bitmap);
    }
    else
    {
        // we failed, so put some noticable garbage in the guy
        int const w = 100; uint32 b[w*w]; for(int i=0;i<w*w;++i) { b[i] = 0xFF00FFFF; }
        AddPixels(State,w,w,b,
                  Image.packed_u0v0, Image.packed_u1v0, Image.packed_u1v1, Image.packed_u0v1,
                  Image.XMin, Image.YMin, Image.XMax, Image.YMax,
                  Image.HalfWidth, Image.FullHeight, x_scale, y_scale);
        Log("Unable to open %s%c%d (or its mirror equivalents)\n",
            DoomPrefix,FrameLetter,ImageNumber);
    }
}

static void
MakeSpriteFrameFromDoom(sprite_texture &State, sprite_frame &Frame,
                        int32 DirectionCount,
                        char const *DoomPrefix, char FrameLetter,
                        float x_scale, float y_scale)
{
    Frame.DirectionCount = DirectionCount;
    Frame.Directions = ArenaPushArray(*State.Arena, DirectionCount, sprite_image);
    assert(Frame.Directions);

    {for(int32 DirectionIndex = 0;
         DirectionIndex < DirectionCount;
         ++DirectionIndex)
    {
        MakeSpriteImageFromDoom(State, Frame.Directions[DirectionIndex],
                                DoomPrefix, FrameLetter,
                                (DirectionCount == 1) ? 0 : DirectionIndex + 1,
                                x_scale, y_scale);
    }}    
}

static void
MakeSpriteAnimationFromDoom(sprite_texture &State, sprite_animation &Animation,
                            int32 FrameCount, int32 DirectionCount,
                            char const *DoomPrefix, char FirstFrameLetter,
                            float x_scale, float y_scale, bool loop)
{
    Animation.FrameCount = FrameCount;
    Animation.Frames = ArenaPushArray(*State.Arena, (FrameCount*2+(loop ? 0 : 2)), sprite_frame);
    assert(Animation.Frames);

    {for(int32 FrameIndex = 0;
         FrameIndex < FrameCount;
         ++FrameIndex)
    {
        MakeSpriteFrameFromDoom(State, Animation.Frames[FrameIndex],
                                DirectionCount, DoomPrefix,
                                FirstFrameLetter + FrameIndex,
                                x_scale, y_scale);
    }}
    if (!loop) {
       Animation.Frames[FrameCount].DirectionCount = Animation.Frames[FrameCount-1].DirectionCount;
       Animation.Frames[FrameCount].Directions = Animation.Frames[FrameCount-1].Directions;
       ++Animation.FrameCount;
       FrameCount = Animation.FrameCount;
    }
    for (int i=FrameCount; i < FrameCount*2; ++i) {
       Animation.Frames[i].Directions = Animation.Frames[i-FrameCount].Directions;
       Animation.Frames[i].DirectionCount = Animation.Frames[i-FrameCount].DirectionCount;
    }
}

sprite_texture *
CreateSpriteTexture(int32 TextureDimension,
                    int32 AnimationCount)
{
    sprite_texture *SpriteTexture = 0;

    memory_arena *Arena = NewMemoryArena();
    if(Arena)
    {
        SpriteTexture = ArenaPush(*Arena, sprite_texture);
        assert(SpriteTexture);
        if(SpriteTexture)
        {
            SpriteTexture->MaxAnimationCount = AnimationCount;
            SpriteTexture->Arena = Arena;
            SpriteTexture->TextureDimension = TextureDimension;
            SpriteTexture->Pixels = ArenaPushArray(
                *Arena,
                SpriteTexture->TextureDimension *
                SpriteTexture->TextureDimension,
                uint32);
            assert(SpriteTexture->Pixels);
            memset(SpriteTexture->Pixels, 0,
                   SpriteTexture->TextureDimension *
                   SpriteTexture->TextureDimension * 4);
            SpriteTexture->CurrentX = 0;
            SpriteTexture->CurrentY = 0;
            SpriteTexture->CurrentRowHeight = 0;
            
            SpriteTexture->AnimationCount = 0;
            SpriteTexture->Animations = ArenaPushArray(
                *Arena, SpriteTexture->MaxAnimationCount,
                sprite_animation);
            assert(SpriteTexture->Animations);

            SpriteTexture->TransparentIndex = GlobalTransparentIndex;
        }
        else
        {
            FreeMemoryArena(Arena);
        }
    }

    return(SpriteTexture);
}

int32
AddAnimationToSpriteTexture(sprite_texture &State,
                            char const *Prefix, char FirstFrameLetter,
                            int32 FrameCount, int32 DirectionCount,
                            float x_scale, float y_scale, bool loop)
{
    assert(State.AnimationCount < State.MaxAnimationCount);

    int32 AnimationIndex = State.AnimationCount++;
    sprite_animation &Animation =
        State.Animations[AnimationIndex];
    MakeSpriteAnimationFromDoom(
        State, Animation,
        FrameCount, DirectionCount, Prefix, FirstFrameLetter, x_scale, y_scale, loop);

    return(AnimationIndex);
}

void
FreeSpriteTexture(sprite_texture *SpriteTexture)
{
    if(SpriteTexture)
    {
        FreeMemoryArena(SpriteTexture->Arena);
    }
}

void
DebugDumpSpriteTexture(sprite_texture &SpriteTexture,
                       char const *FileName)
{
    bmp BMP;
    BMP.Width = SpriteTexture.TextureDimension;
    BMP.Height = SpriteTexture.TextureDimension;
    BMP.Stride = SpriteTexture.TextureDimension * 4;
    BMP.Pixels = (uint32 *)SpriteTexture.Pixels;
    WriteBMP(BMP, FileName);
}

