#if !defined(SPRITE_TEXTURE_H)
/* ========================================================================
   $RCSfile: sprite_texture.h,v $
   $Date: 2002/03/14 01:08:43 $
   $Revision: 1.14 $
   $Creator: Casey Muratori $
   $Notice: $
   ======================================================================== */

#if !defined(GLOBALS_H)
#include "globals.h"
#endif

/*
  OK.  Basically what you do is this:

  sprite_texture *SpriteTexture = CreateSpriteTexture(1024, 4);

  That makes a 1024x1024 texture and room for storing info about
  4 animations.  You then add the animations like this:
  
  AddAnimationToSpriteTexture(*SpriteTexture, "c:/wads/sarg", 'A', 4, 8);
  AddAnimationToSpriteTexture(*SpriteTexture, "c:/wads/sarg", 'E', 4, 8);
  ...
  AddAnimationToSpriteTexture(*SpriteTexture, "c:/wads/icey", 'A', 4, 8);

  You pass the directory and prefix as a string, then the "first
  frame letter" which is Doom's way of identifying what animation
  you are, the number of frames (ie., letter range), and then the
  number of facing directions you want loaded.  It will look
  for all that crap, pack it into the texture, and then return.

  By default, it will use DefaultImageWidth and DefaultImageHeight
  for the size to pack, but you can set this (per animation call,
  actually - so it doesn't need to be constant):
  
  SetSpriteTexturePackingSize(*SpriteTexture, 128, 128);

  This would pack at 128x128.

  When you're done with the sprite_texture, make sure you
  free it all with

  FreeSpriteTexture(SpriteTexture);
  
  If you want to test the texture that's generated, you can dump
  it to a 24 bit bitmap:
  
  DebugDumpSpriteTexture(*ST, "c:/wads/PackedSarge.bmp");
*/

// SCALE_IMAGES controls whether or not all bitmaps are scaled
// to fit in the allotted image space.  If it's 1, then images
// are scaled to the specified image size always, and their
// original width and height is recorded.  If it's 0, then
// the images are padded or cropped to fit the image size.
#define SCALE_IMAGES 0

struct sprite_image
{
    // These are the UV bounds for this image in the
    // texture stored in sprite_texture (all images
    // share one texture).
    int packed_u0v0;
    int packed_u1v0;
    int packed_u1v1;
    int packed_u0v1;

    // These are the UV bounds, but expressed in terms
    // of actual integer X/Y.
    uint16 XMin, YMin;
    uint16 XMax, YMax;

    real32 HalfWidth, FullHeight;

#if SCALE_IMAGES
    // These are the original dimensions of the bitmap
    // that is now packed into the specified UV range.
    int32 OriginalWidth;
    int32 OriginalHeight;
#endif
};

struct sprite_frame
{
    // These are the various facing directions for this
    // frame of the animation.
    // Ripped from the doom specs:
    //         3
    //       4 | 2
    //        \|/
    //      5--*----> 1   Thing is facing this direction
    //        /|\
    //       6 | 8
    //         7
         
    int32 DirectionCount;
    sprite_image *Directions;
};

struct sprite_animation
{
    // These are the frames of the animation.
    int32 FrameCount;
    sprite_frame *Frames;
};

struct sprite_texture
{    
    // These are the animations for this sprite.
    // Each one is named.
    int32 AnimationCount;
    sprite_animation *Animations;

    // This is the actual texture data.  It is
    // raw RGBA pixels, Dimension x Dimension.
    int32 TextureDimension;
    uint32 *Pixels;
    
    // This is the state of the sprite_texture
    // since it can be incrementally built.  You
    // generally shouldn't touch these unless you
    // really know what you're doing.
    struct memory_arena *Arena;
    int32 MaxAnimationCount;
    int32 CurrentX;
    int32 CurrentY;
    int32 CurrentRowHeight;

    int32 TransparentIndex;
};

sprite_texture *CreateSpriteTexture(int32 TextureDimension,
                                    int32 MaxAnimationCount);

int32 AddAnimationToSpriteTexture(
    sprite_texture &Texture, char const *Prefix,
    char FirstFrameLetter, int32 FrameCount, int32 DirectionCount,
    float x_scale, float y_scale, bool loop);

void FreeSpriteTexture(sprite_texture *SpriteTexture);

void DebugDumpSpriteTexture(sprite_texture &SpriteTexture,
                            char const *FileName);

#define SPRITE_TEXTURE_H
#endif
