/*----------------------------------------------------------------------------
  value_map.cpp
*/

#include <assert.h>
#include <math.h>

#define WIN32_LEAN_AND_MEAN
#define WIN32_EXTRA_LEAN
#include <windows.h>
#include <gl/gl.h>
#include <gl/glu.h>
#include <stdio.h>
#include <stdlib.h>

#include "value_map.h"
#include "guys.h"
#include "hf.h"
#include "buildings.h"

/*----------------------------------------------------------------------------

Load from either a bmp, a psd, or @todo maybe a gif later?

*/

// BMP file header
// bits are dword aligned scanlines
#pragma pack(1)
struct BMPFile {
    uint16   bfType;
    uint32   bfSize;
    uint16   bfReserved1;
    uint16   bfReserved2;
    uint32   bfOffBits;
    uint32   biSize;
    int32    biWidth;
    int32    biHeight;
    uint16   biPlanes;
    uint16   biBitCount;
    uint32   biCompression;
    uint32   biSizeImage;
    int32    biXPelsPerMeter;
    int32    biYPelsPerMeter;
    uint32   biClrUsed;
    uint32   biClrImportant;
};

// PSD file header, everything is bigendian
struct PSDFile {
    uint32 Signature;
    uint16 Version;
    uint8  Reserved[6];
    uint16 Channels;
    uint32 Rows;
    uint32 Columns;
    uint16 Depth;
    uint16 Mode;
};
#define SWAP16(w) (uint16((((w)>>8)&0xFF)|(((w)<<8)&0xFF00)))
#define SWAP32(w) (uint32((((w)>>24)&0xFF)|(((w)<<24)&0xFF000000)|(((w)<<8)&0xFF0000)|(((w)>>8)&0xFF00)))
// followed by chunks (uint32 Length as first field):
// Color mode data section
// Image resources section
// Layer and Mask information section
// uint16 Compression = 0 for raw, = 1 for RLE
// Image data section, one channel at a time
// then bits, scanline order, no pad

value_map *MapLoad( char const *FileName, int unsigned Channel, bool Invert )
{
    value_map *Map = 0;
    uint8 *Data = 0;

    assert(FileName);
    HANDLE File = CreateFile(FileName,GENERIC_READ,0,0,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,0);
    assert(File != INVALID_HANDLE_VALUE);
    HANDLE FileMap = CreateFileMapping(File,0,PAGE_READONLY,0,0,0);
    assert(FileMap != INVALID_HANDLE_VALUE);
    void *FileData = MapViewOfFile(FileMap,FILE_MAP_READ,0,0,0);
    assert(FileData);
    uint8 *FileBytes = (uint8 *)FileData;
    if(FileBytes)
    {
        if(FileBytes[0] == 'B' && FileBytes[1] == 'M')
        {
            // it's a Bowel Movement Phile, BMP
            BMPFile *Bmp = (BMPFile *)FileBytes;
            int32 Width = Bmp->biWidth;
            int32 Height = Bmp->biHeight;
            uint32 BmpChannels = uint32(Bmp->biBitCount/8);
            assert(Width > 0 && Height > 0);
            assert(Bmp->biCompression == 0);
            assert(Channel < BmpChannels);

            Map = new value_map;  assert(Map);
            Data = new uint8[Width*Height];  assert(Data);
            if(Map && Data &&
               (Width > 0) && (Height > 0) &&
               (Bmp->biCompression == 0) && (Channel < BmpChannels))
            {
                Map->Min = 0.f;
                Map->Max = 255.f;
                Map->Width = Width;
                Map->Height = Height;
                Map->Data = Data;

                // rip out the specified channel
                // this code works for 8, 24, 32
                // (and 16 if you define that as two 8-bit channels :)
                // we want the result to be bottom-up because it matches the terrain
                // @todo handle top-down bmps?
                int32 LineStride = ((Width * (Bmp->biBitCount / 8) + 0x3) & (~0x3));

                // we reverse the BGRA to RGBA, so 012 -> 210
                if(BmpChannels > 2) {
                    switch(Channel) {
                        case 0: Channel = 2; break;
                        case 1: Channel = 1; break;
                        case 2: Channel = 0; break;
                    }
                }

                // pixel computations
                uint32 PixelShift = Channel * 8;
                uint32 PixelAdd = BmpChannels;
                // this will read off the end, but hey, BMP scanlines are
                // 32 bit aligned, so it works!
                uint8 *l = FileBytes + Bmp->bfOffBits;
                for(uint i = 0;i < Map->Height;++i)
                {
                    uint8 *p = l;
                    for(uint j = 0;j < Map->Width;++j)
                    {
                        uint32 d = *(uint32*)p;
                        uint8 b = uint8(((d)>>PixelShift)&0xFF);
                        if(Invert) {
                            b = ~b;
                        }
                        Map->Data[i*Map->Height+j] = b;
                        p += PixelAdd;
                    }
                    l += LineStride;
                }
            }
            else
            {
                delete Map;  Map = 0;
                delete [] Data;  Data = 0;
            }
        }
        else
        if(FileBytes[0] == '8' && FileBytes[1] == 'B' &&
           FileBytes[2] == 'P' && FileBytes[3] == 'S')
        {
            // it's a Photoshop PSD file
            PSDFile *Psd = (PSDFile *)FileBytes;
            uint32 Width = SWAP32(Psd->Columns);
            uint32 Height = SWAP32(Psd->Rows);
            assert(SWAP16(Psd->Version) == 1);
            assert(SWAP16(Psd->Depth) == 8);
            assert(Channel < SWAP16(Psd->Channels));

            Map = new value_map;  assert(Map);
            Data = new uint8[Width*Height];  assert(Data);
            if(Map && Data &&
               (SWAP16(Psd->Version) == 1) && 
               (SWAP16(Psd->Depth) == 8) && (Channel < SWAP16(Psd->Channels)))
            {
                Map->Min = 0.f;
                Map->Max = 255.f;
                Map->Width = Width;
                Map->Height = Height;
                Map->Data = Data;

                // we want the result to be bottom-up because it matches the terrain
                Data += (Height-1)*Width;  // point to last scanline
                int LineStride = -int(Width);   // walk up the bitmap

                FileBytes += sizeof(PSDFile);
                // skip Color mode
                FileBytes += SWAP32(*(uint32 *)(FileBytes)) + 4;
                // skip Image resources
                FileBytes += SWAP32(*(uint32 *)(FileBytes)) + 4;
                // skip Layers
                FileBytes += SWAP32(*(uint32 *)(FileBytes)) + 4;
                uint16 Compression = SWAP16(*(uint16*)FileBytes);
                FileBytes += 2;
                if(Compression == 0)
                {
                    // uncompressed data

                    // skip to the specified channel
                    uint32 ChannelSize = Width * Height;
                    FileBytes += ChannelSize * Channel;

                    // rip out the specified channel
                    for(uint i = 0;i < Map->Height;++i)
                    {
                        for(uint j = 0;j < Map->Width;++j)
                        {
                            if(Invert) {
                                Data[j] = ~(*FileBytes++);
                            } else {
                                Data[j] = *FileBytes++;
                            }
                        }
                        Data += LineStride;
                    }
                }
                else
                {
                    // RLE compressed data
                    // header of uint16 line lengths
                    // n b+
                    // n=[00,7F]: copy n+1 b's
                    // n=[81,FF]: dupe b -n+1 times
                    // n=80: finished/skip

                    // figure out where this channel's scanlines start
                    uint BytesToChannel = 2*Height*Channel;
                    uint LineOffset = 2*Height*SWAP16(Psd->Channels);
                    uint i = 0;
                    while(i < BytesToChannel) {
                        LineOffset += SWAP16(*(uint16*)(FileBytes+i));
                        i += 2;
                    }
                    FileBytes += LineOffset;

                    // now we're at Channel
                    int BytesToWrite = Width*Height;
                    int LineCount = 0;
                    while(BytesToWrite > 0) {
                        // decompress run
                        int Count = *(int8*)FileBytes;
                        ++FileBytes;
                        if(Count >= 0) {
                            // copy
                            Count += 1;
                            BytesToWrite -= Count;
                            LineCount += Count;
                            while(Count--) {
                                if(Invert) {
                                    *Data++ = ~(*FileBytes++);
                                } else {
                                    *Data++ = *FileBytes++;
                                }
                            }
                        } else
                        if(Count > -128) {
                            // run
                            Count = -Count + 1;
                            BytesToWrite -= Count;
                            memset(Data,
                                   Invert ? ~(*FileBytes++) : *FileBytes++,
                                   Count);
                            Data += Count;
                            LineCount += Count;
                        }
                        // line end?
                        assert(uint32(LineCount) <= Width);
                        if(uint32(LineCount) >= Width) {
                            // go back to start of this line and then to next
                            LineCount = 0;
                            Data += LineStride - Width;
                        }
                        assert(BytesToWrite >= 0);
                    }
                }

            }
            else
            {
                delete Map;  Map = 0;
                delete [] Data;  Data = 0;
            }
        }
    }

    UnmapViewOfFile(FileData);
    CloseHandle(FileMap);
    CloseHandle(File);

    return Map;
} 
value_map *MapCreate( uint Width, uint Height, uint8 Init )
{
    value_map *m = new value_map; assert(m);
    assert(Width*Height);
    if(m && Width*Height) {
        m->Width = Width;
        m->Height = Height;
        m->Min = 0.f;
        m->Max = 255.f;
        m->Data = new uint8[Width*Height];  assert(m->Data);
        if(m->Data) {
            memset(m->Data,Init,Width*Height);
        } else {
            delete m;
            m = 0;
        }
    }
    return m;
}
void MapWritePSD( char const *FileName, value_map *Map )
{
    assert(FileName && Map && Map->Data);
    if(FileName && Map && Map->Data) {
        FILE *file = fopen(FileName,"wb");
        PSDFile header;
        ((char*)&header)[0] = '8';
        ((char*)&header)[1] = 'B';
        ((char*)&header)[2] = 'P';
        ((char*)&header)[3] = 'S';
        header.Version = SWAP16(1);
        memset(header.Reserved,0,6);
        header.Channels = SWAP16(1);
        header.Rows = SWAP32(Map->Height);
        header.Columns = SWAP32(Map->Width);
        header.Depth = SWAP16(8);
        header.Mode = SWAP16(1);
        fwrite(&header,sizeof(header),1,file);
        uint32 z = 0;
        fwrite(&z,4,1,file);  // Color mode
        fwrite(&z,4,1,file);  // Image resources
        fwrite(&z,4,1,file);  // Layer & Mask
        fwrite(&z,2,1,file);  // compression = 0
        // write out data top-down
        for(uint i = 0;i < Map->Height;++i) {
            fwrite(Map->Data + Map->Width*(Map->Height-1) - i*Map->Width,Map->Width,1,file);
        }
        fclose(file);
    }
}

void MapFree( value_map *Map )
{
    assert(Map);
    if(Map)
    {
        delete [] Map->Data;
        delete Map;
    }
}

void MapSetRange( value_map *Map, float Min, float Max )
{
    assert(Map);
    if(Map)
    {
        Map->Min = Min;
        Map->Max = Max;
    }
}

inline float GetRangeValue( value_map *Map, float v )
{
    assert(Map && Map->Data && v >= 0.f && v <= 255.f);
    return Map->Min + (v/255.f)*(Map->Max - Map->Min);
}
inline float GetRangeValue( value_map *Map, uint8 v )
{
    assert(Map && Map->Data);    
    return GetRangeValue(Map,float(v));
}
inline float GetRangeValue( value_map *Map, uint j, uint i )
{
    assert(Map && Map->Data && i < Map->Height && j < Map->Width);
    return GetRangeValue(Map,Map->Data[i*Map->Width+j]);
}
inline float GetRangeValueClosed( value_map *Map, uint j, uint i )
{
    assert(Map);
    return GetRangeValue(Map,min(j,Map->Width-1),min(i,Map->Height-1));
}
uint8 MapUnmapValue( value_map *Map, float v )
{
    assert(Map);
    if(Map) {
        float vv = 255.f * (v - Map->Min)/(Map->Max - Map->Min);
        vv = min(255,max(0,vv));
        uint8 cv = uint8(vv);
        return cv;
    }
    return 0;
}
float MapMapValue( value_map *Map, uint8 c )
{
    return GetRangeValue(Map,c);
}
uint MapGetX( value_map *Map, float x )
{
    assert(Map);
    if(Map) {
        // rescale onto the map
        float w = float(Map->Width);
        float mx = (x/float(WORLD_X)) * w;

        // GL_CLAMP_TO_EDGE
        float cx = min(max(0.5,mx),w-0.5);
        float ix = floor(mx);

        uint j = uint(ix);
        return j;
    }
    return 0;
}
uint MapGetY( value_map *Map, float y )
{
    assert(Map);
    if(Map) {
        // rescale onto the map
        float h = float(Map->Height);
        float my = (y/float(WORLD_Y)) * h;

        // GL_CLAMP_TO_EDGE
        float cy = min(max(0.5,my),h-0.5);
        float iy = floor(my);

        uint i = uint(iy);
        return i;
    }
    return 0;
}
float MapGetPointSample( value_map *Map, float x, float y )
{
    assert(Map && Map->Data);
    if(Map && Map->Data)
    {
        // rescale onto the map
        float w = float(Map->Width);
        float h = float(Map->Height);
        float mx = (x/float(WORLD_X)) * w;
        float my = (y/float(WORLD_Y)) * h;

        // GL_CLAMP_TO_EDGE
        float cx = min(max(0.5,mx),w-0.5);
        float cy = min(max(0.5,my),h-0.5);
        float ix = floor(cx);
        float iy = floor(cy);

        uint i = uint(iy);
        uint j = uint(ix);

        assert(i < Map->Height && j < Map->Width);
        return GetRangeValue(Map,j,i); 
    }
    return 0.f; // @todo ???
}
float MapSetPointSample( value_map *Map, float x, float y, float v )
{
    assert(Map && Map->Data);
    if(Map && Map->Data)
    {
        // hf 0,1 space is x/WORLD_X
        // so put us in our map space
        // then use the correct point sample rules, [0,w]
        // we include the far edge
        float mx = (x/WORLD_X) * Map->Width;
        float my = (y/WORLD_Y) * Map->Height;
        float ix = floor(mx);
        float iy = floor(my);
        uint i = uint(iy);
        uint j = uint(ix);

        // handle far edges so we're closed on both sides
        if((j == Map->Width) && ((ix - mx) == 0.f)) {
            j = Map->Width - 1;
        }
        if((i == Map->Height) && ((iy - my) == 0.f)) {
            i = Map->Height - 1;
        }

        float vv = 255.f * (v - Map->Min)/(Map->Max - Map->Min);
        vv = min(255,max(0,vv));
        uint8 cv = uint8(vv);

        // set the value, or not
        if(i < Map->Height && j < Map->Width)
        {
            Map->Data[i*Map->Width+j] = cv;
        }
        return GetRangeValue(Map,cv);
    }

    return 0.f; // @todo ???
}

float MapGetBilerpSample( value_map *Map, float x, float y )
{
    assert(Map && Map->Data);
    if(Map && Map->Data)
    {
        // rescale onto the map
        float w = float(Map->Width);
        float h = float(Map->Height);
        float mx = (x/float(WORLD_X)) * w;
        float my = (y/float(WORLD_Y)) * h;

        // GL_CLAMP_TO_EDGE
        float cx = min(max(0.5,mx),w-0.5);
        float cy = min(max(0.5,my),h-0.5);

        // pick bilerp coordinates
        float ix = floor(cx-0.5);
        float iy = floor(cy-0.5);
        float alpha = cx - (ix + 0.5);
        float beta = cy - (iy + 0.5);
        uint i0 = uint(iy);
        uint i1 = i0 + 1;
        uint j0 = uint(ix);
        uint j1 = j0 + 1;
        assert(i0 <= Map->Height && i1 <= Map->Height && j0 <= Map->Width && j1 <= Map->Width);

        // get pixel values, vxy
        float v00 = GetRangeValueClosed(Map,j0,i0);
        float v10 = GetRangeValueClosed(Map,j1,i0);
        float v01 = GetRangeValueClosed(Map,j0,i1);
        float v11 = GetRangeValueClosed(Map,j1,i1);

        // do the bilerp
        return (1-alpha)*(1-beta)*v00 + alpha*(1-beta)*v10 + (1-alpha)*beta*v01 + alpha*beta*v11;
    }
    return 0.0; // @todo ???
}

// map uses

void MapAllocateGuyDensity( value_map *Map, Guy *Template, uint StartIndex, uint TotalGuys,
                            bool RandomizeFrameOffsets, bool SlowButAwesome )
{
    // @todo handle guys inside buildings and whatnot!

    assert(Map && Map->Data);
    assert(StartIndex + TotalGuys <= MAX_GUYS);
    if(Map && Map->Data && (StartIndex + TotalGuys <= MAX_GUYS))
    {
        if(!SlowButAwesome) {
            // fairly random but not great in low density areas
            
            // first, find total mass of map
            float mass = 0.;
            for(uint i = 0;i < Map->Height;++i)
            {
                for(uint j = 0;j < Map->Width;++j)
                {
                    mass += Map->Data[i*Map->Width+j];
                }
            }
            assert(mass > 0);

            // now allocate guys onto terrain
            // @todo need some way to finish init partial guys?  later:  what does this mean???  just StartIndex?
            // @todo should allocate rest of guys (TotalGuys = MAX_GUYS - StartIndex) if not fit?
            if(mass > 0)
            {
                // find how many guys are allocated per unit density in map
                float density = TotalGuys/mass;

                // figure out map sample coverage area on terrain
                float tw = float(WORLD_X);
                float th = float(WORLD_Y);
                float sw = tw/float(Map->Width);
                float sh = th/float(Map->Height);
            

                // iterate over map, distributing guys
                float frac = 0.f;
                uint gi = StartIndex;
                float lastx, lasty;
                uint GuysCount = 0;
                for(uint i = 0;i < Map->Height;++i)
                {
                    for(uint j = 0;j < Map->Width;++j)
                    {
                        // figure out how many guys to allocate, including
                        // fractional guys from previous samples
                        // @todo dither this better? use Siggraph 2001 dither?
                        float gcf = density * Map->Data[i*Map->Width+j];
                        uint gci = floor(gcf);
                        frac += gcf - gci;
                        gci += floor(frac);
                        frac -= floor(frac);

                        // allocated gci guys randomly in sample box
                        for(uint g = 0;g < gci;++g)
                        {
                            Guy *pg = gGuys + gi + g;
                            *pg = *Template;
                            if(RandomizeFrameOffsets) {
                                pg->animation_timing_offset = rand() % ((4*256) >> 3);
                            }
                            float x = (j+genrand_real2())*sw;
                            float y = (i+genrand_real2())*sh;
                            assert(x < tw && y < th);
                            pg->loc.x = x;
                            pg->loc.y = y;
                            pg->loc.z = ground(pg->loc.x, pg->loc.y);
                            ++GuysCount;
                            // save the last spot
                            lastx = j*sw;
                            lasty = i*sh;
                        }
                        gi += gci;
                    }
                }
                // clean up the last guy (should only be one) by dropping him at
                // the last drop spot
                assert((GuysCount == TotalGuys - 1) || (GuysCount == TotalGuys));
                while(GuysCount < TotalGuys)
                {
                    Guy *pg = gGuys + StartIndex + GuysCount;
                    *pg = *Template;
                    if(RandomizeFrameOffsets) {
                        pg->animation_timing_offset = rand() % ((4*256) >> 3);
                    }
                    pg->loc.x = lastx;
                    pg->loc.y = lasty;
                    pg->loc.z = ground(pg->loc.x, pg->loc.y);
                    ++GuysCount;
                }
                assert(GuysCount == TotalGuys);
            }
        } else {
            // totally slow but really random density algorithm

            // find all the pixels with a non-0 density
            struct dense_pixel {
                uint i,j;
                uint8 v;
            };
            dense_pixel *Pixels = new dense_pixel[Map->Width*Map->Height]; assert(Pixels);
            uint PixelCount = 0;
            for(uint i = 0;i < Map->Height;++i)
            {
                for(uint j = 0;j < Map->Width;++j)
                {
                    uint8 v = Map->Data[i*Map->Width+j];
                    if(v) {
                        Pixels[PixelCount].i = i;
                        Pixels[PixelCount].j = j;
                        Pixels[PixelCount].v = v;
                        ++PixelCount;
                    }
                }
            }
            // figure out map sample coverage area on terrain
            float tw = float(WORLD_X);
            float th = float(WORLD_Y);
            float sw = tw/float(Map->Width);
            float sh = th/float(Map->Height);

            // distribute the guys over the terrain
            // @todo really want a Poisson distribution here, not white noise
            Guy *pg = gGuys + StartIndex;
            while(TotalGuys) {
                uint p = PixelCount*genrand_real2();
                assert(p < PixelCount);
                uint8 r = genrand_int32() >> 24;
                if(r <= Pixels[p].v) {
                    *pg = *Template;
                    if(RandomizeFrameOffsets) {
                        pg->animation_timing_offset = rand() % ((4*256) >> 3);
                    }
                    float x = (Pixels[p].j + genrand_real2())*sw;
                    float y = (Pixels[p].i + genrand_real2())*sh;
                    assert(x < tw && y < th);
                    pg->loc.x = x;
                    pg->loc.y = y;
                    pg->loc.z = ground(pg->loc.x, pg->loc.y);
                    --TotalGuys;
                    ++pg;
                }
            }
            delete [] Pixels;
        }
    }
}

struct building_template
{
    float x0, y0, x1, y1;
    float h;
    float r, g, b;
    uint8 Code;
} BuildingTemplates[MAX_BUILDINGS];
uint BuildingCount = 0;
void AllocateBuilding( uint8 Code, float x0, float y0, float x1, float y1, float h, float r, float g, float b, bool CodeBased )
{
    // @todo cache last building
    // returns the template if it's new
    building_template *bt = 0;
    for(uint i = 0;i < BuildingCount;++i) {
        building_template &b = BuildingTemplates[i];
        if(((CodeBased && (b.Code == Code)) || (!CodeBased)) &&
           b.x0 == x0 && b.y0 == y0 && b.x1 == x1 && b.y1 == y1)
        {
            bt = &b;
            break;
        }
    }
    assert(bt || BuildingCount < MAX_BUILDINGS);
    if(!bt && BuildingCount < MAX_BUILDINGS) {
        bt = BuildingTemplates + BuildingCount;
        bt->x0 = x0; bt->y0 = y0;
        bt->x1 = x1; bt->y1 = y1;
        bt->h = h;
        bt->r = r; bt->g = g; bt->b = b;
        bt->Code = Code;
        BuildingCount++;
    } else
    if(bt) {
        // update the height
        bt->h = max(bt->h,h);
    }
}
void FindExtents( value_map *Map, uint &ii, uint &jj, int di, int dj, bool CodeBased )
{
    int i = int(ii), j = int(jj);
    assert(Map && Map->Data && (ii < Map->Height) && (jj < Map->Width));
    uint8 Code = Map->Data[i*Map->Width+j];
    while((i >= 0) && (i < int(Map->Height)) && (j >= 0) && (j < int(Map->Width)))
    {
        if((CodeBased && (Map->Data[(i+di)*Map->Width+(j+dj)] == Code)) ||
           (!CodeBased && Map->Data[(i+di)*Map->Width+(j+dj)] != 0)) {
            // same code, keep going
            i += di;
            j += dj;
        } else {
            // we've found the extents
            ii = uint(i);
            jj = uint(j);
            return;
        }
    }
    // we went off the edge of the map
    ii = uint(min(max(i,0),int(Map->Height-1)));
    jj = uint(min(max(j,0),int(Map->Width-1)));
}
void MapAllocateBuildings( value_map *Positions, value_map *Heights,
                           value_map *Red, value_map *Green, value_map *Blue,
                           bool CodeBased )
{
    // This is crazy, I tell you!  We iterate over Positions,
    // and find the non-0 color axis-aligned boxes, getting their
    // extents from Positions, and their heights from Heights.
    // The Positions values can also be the building "code", so
    // the value in Positions could be used later, presumably.

    BuildingCount = 0;
    assert(Positions && Positions->Data);
    assert(Heights && Heights->Data);
    if(Positions && Positions->Data && Heights && Heights->Data)
    {
        // okay, the algorithm:
        // assume no overlapping buildings
        // iterate over all the pixels, if non-0 then find extents and
        // possibly alloc a building
        
        // figure out map sample coverage area on terrain
        float sw = WORLD_X/float(Positions->Width);
        float sh = WORLD_Y/float(Positions->Height);
            
        for(uint i = 0;i < Positions->Height;++i)
        {
            for(uint j = 0;j < Positions->Width;++j)
            {
                // @todo cache this shit
                uint8 Code = Positions->Data[i*Positions->Width+j];
                if(Code) {
                    // we're in a building, find its extents
                    uint i0=i, i1=i, j0=j, j1=j;
                    FindExtents(Positions,i0,j,-1, 0,CodeBased);
                    FindExtents(Positions,i1,j, 1, 0,CodeBased);
                    FindExtents(Positions,i,j0, 0,-1,CodeBased);
                    FindExtents(Positions,i,j1, 0, 1,CodeBased);
                    float x0 = j0*sw;
                    float y0 = i0*sh;
                    float x1 = (j1+1)*sw;
                    float y1 = (i1+1)*sh;
                    float h = MapGetBilerpSample(Heights,(x1+x0)/2.f,(y1+y0)/2.f);
                    float r, g, b;
                    // center pixel of building
                    uint ii = (i1 + i0)/2; uint jj = (j1 + j0)/2;
                    // colors?
                    if(Red && Green && Blue) {
                        r = GetRangeValue(Red,jj,ii);
                        g = GetRangeValue(Green,jj,ii);
                        b = GetRangeValue(Blue,jj,ii);
                    } else {
                        r = GenerateUnitRandomReal();
                        g = GenerateUnitRandomReal();
                        b = GenerateUnitRandomReal();
                    }
                    // which code mode?
                    if(!CodeBased) {
                        // find the Code of the center pixel of the building
                        Code = Positions->Data[ii*Positions->Width+jj];
                    }
                    AllocateBuilding(Code,x0,y0,x1,y1,h,r,g,b,CodeBased);
                }
            }
        }

        // okay, we've allocated a bunch of buildings, so make them for real
        while(BuildingCount) {
            --BuildingCount;
            building_template &b = BuildingTemplates[BuildingCount];
            buildingCreate(b.x0,b.y0,b.x1,b.y1,b.h,b.r,b.g,b.b);
        }
    }
}

void MapSetHeightField( value_map *Map )
{
    assert(Map && Map->Data);
    if(Map && Map->Data) {

        // map from height field into map, ensuring corner pixel centers line up
        // hfx -> mapx:  0 -> 0.5, hfw-1 -> mapw-0.5
        // x (mapw - 1)/(hfw - 1) + 0.5
        // mapx -> worldx: 0 -> 0, mapw -> worldw
        float hwmw = float(Map->Width - 1)/float(HF_X - 1);
        float hhmh = float(Map->Height - 1)/float(HF_Y - 1);
        float mwww = float(WORLD_X)/float(Map->Width);
        float mhwh = float(WORLD_Y)/float(Map->Height);

        for(int i = 0;i < HF_Y;++i) {
            for(int j = 0;j < HF_X;++j) {
                height_field[i][j] = MapGetBilerpSample(Map,mwww*(j*hwmw+0.5),mhwh*(i*hhmh+0.5));
            }
        }
    }
}

extern int ground_tex, ground_tex_w, ground_tex_h;
extern uint8 *ground_tex_data;
void texMake32(int w, int h, uint8 *data, int id, uint WrapMode);
inline void CopyChannel( uint8 (*dst)[4], uint8 *src, uint32 idx, uint32 Count )
{
    assert(dst && src);
    while(Count--) {
        (*dst++)[idx] = *src++;
    }
}
void MapSetTerrainTextureChannels( value_map *Red, value_map *Green, value_map *Blue, bool ResizeTexture, bool MipMap )
{
    assert(Red || Green || Blue);
    assert(!Red || Red->Data);
    assert(!Green || Green->Data);
    assert(!Blue || Blue->Data);

    // find at least one non-0 map
    value_map *m = Red ? Red : (Green ? Green : (Blue ? Blue : 0));
    assert(m);
    if(m) {
        // we throw away the old texture on ResizeTexture

        if(!ground_tex) {
            ResizeTexture = true;
            glGenTextures(1,(GLuint*)&ground_tex);
        }

        assert(ground_tex);
        glBindTexture(GL_TEXTURE_2D,ground_tex);

        if(ResizeTexture) {
            ground_tex_w = int(m->Width); ground_tex_h = int(m->Height);
        }
        int w = ground_tex_w, h = ground_tex_h;
        assert(w && h);

        if(ResizeTexture) {
            delete [] ground_tex_data;
            ground_tex_data = new uint8[w*h*4];  assert(ground_tex_data);
        }
        uint8 (*p)[4] = (uint8 (*)[4])ground_tex_data;

        if(p) {
            if(ResizeTexture) {
                memset(p,0,w*h*4);
            }
            // do fast copy if we're compatible
            if(w == int(m->Width) && h == int(m->Height) &&
               (!Red || (Red->Width == m->Width && Red->Height == m->Height)) &&
               (!Green || (Green->Width == m->Width && Green->Height == m->Height)) &&
               (!Blue || (Blue->Width == m->Width && Blue->Height == m->Height)))
            {
                if(Red) { CopyChannel(p,Red->Data,2,w*h); }
                if(Green) { CopyChannel(p,Green->Data,1,w*h); }
                if(Blue) { CopyChannel(p,Blue->Data,0,w*h); }
            } else {
                float ws = WORLD_X/float(w), hs = WORLD_Y/float(h);
                // sample from the pixel centers of the texture into the bilerped map
                for(int i = 0;i < h;++i) {
                    for(int j = 0;j < w;++j) {
                        // texMake32 takes textures in GL_BGRA_EXT format
                        uint8 *bgra = p[i*w+j];
                        #define DOMAP(m,idx) if(m) { \
                            float v = MapGetBilerpSample(m,(j+0.5f)*ws,(i+0.5f)*hs); \
                            bgra[idx] = MapUnmapValue(m,v); }
                        DOMAP(Red,2);
                        DOMAP(Green,1);
                        DOMAP(Blue,0);
                        #undef DOMAP
                    }
                }
            }
            if(ResizeTexture || MipMap) {
                texMake32(w,h,p[0],ground_tex,GL_CLAMP_TO_EDGE);
            } else {
                glTexSubImage2D(GL_TEXTURE_2D,0,0,0,w,h,GL_BGRA_EXT,GL_UNSIGNED_BYTE,p[0]);
            }
        }
    }
}
void MapSetTerrainTexture( char const *FileName, bool ResizeTexture )
{
    assert(FileName);
    if(FileName) {
        value_map *Red = MapLoad(FileName,0);
        value_map *Green = MapLoad(FileName,1);
        value_map *Blue = MapLoad(FileName,2);
        assert(Red && Green && Blue);
        if(Red && Green && Blue) {
            MapSetTerrainTextureChannels(Red,Green,Blue,ResizeTexture);
        }
        MapFree(Blue);
        MapFree(Green);
        MapFree(Red);
    }
}


#if 0
// test code
    {
        value_map *m = MapLoad("data/angry.psd",0);
        for(float x = 0;x < WORLD_X;x += 1) {
            for(float y = 0;y < WORLD_Y;y += 1) {
                float vv = MapSetPointSample(m,x,y,255*floatrand());
                assert(fabs(vv - MapGetPointSample(m,x,y)) < 1e-5);
                assert(m->Data[MapGetY(m,y)*m->Width+MapGetX(m,x)] == MapUnmapValue(m,MapGetPointSample(m,x,y)));
            }
        }
    }

#endif
