/*----------------------------------------------------------------------------
  value_map.h - functions for augmented data maps for doing queries on
  terrain type, etc.
*/

#if !defined(VALUE_MAP_H)

#if !defined(GLOBALS_H)
#include "globals.h"
#endif

/*----------------------------------------------------------------------------

  value_map
  - struct holding 8 bit arrays of data, Width*Height
  - packed/no pad bytes, stride is Width
  - bottom-up to match world, so 0,0 is lower left corner
  - feel free to iterate over the data and munge it yourself
  - allocated with new and new []
*/
struct value_map
{
    float Min, Max;
    uint Width, Height;
    uint8 *Data;
};    

/*----------------------------------------------------------------------------

  MapLoad loads 8 bit channels for miscellaneous maps.  It will load from:
  - BMP
    8, 24, or 32 bpp BMP (or 16 but it's treated as 2 8 bit channels)
    BMPs are RGBA in channels 0123 (this means we internally reverse BGR->RGB)
    8bpp is obviously just channel 0
  - PSD (Photoshop)
    channels are RGB3456...->0123456... (3456 are arbitrary extra alpha channels)
    grayscale PSDs are channel 0, CMYK are 0123, etc.
  MapCreate
  - makes a blank value_map with the parms
  MapWritePSD
  - writes out a greyscale photoshop file
  - primarily for debugging
  
    @todo want a MapLoadAll that loads all the channels?
*/
value_map *MapLoad( char const *FileName, int unsigned Channel = 0, bool Invert = false );
value_map *MapCreate( uint Width, uint Height, uint8 Init = 0 );
void MapFree( value_map *Map );
void MapWritePSD( char const *FileName, value_map *Map );
    
/*----------------------------------------------------------------------------
  
  MapSetRange
  - sets the mapping from [0,255] to [Min,Max]
  MapGetBilerpSample
  - does what it says, uses GL_CLAMP_TO_EDGE semantics
  - uses world coordinates, closed on both sides
  MapGetPointSample
  - does what it says, uses GL_CLAMP_TO_EDGE semantics
  - similarly intuitive
  MapSetPointSample
  - sets the closest point to the world coordinates, inverse mapped through the range
  - does not set outside map, so no edge clamping
  - returns range mapped value set, clamped to [Min,Max]
  - identities:
    MapSetPointSample(m,x,y,v); MapGetPointSample(m,x,y) != v       (truncation error)
    MapSetPointSample(m,x,y,v) != v                                 (truncation error)
    v' = MapSetPointSample(m,x,y,v); MapGetPointSample(m,x,y) ~= v' (within eps, ~1e-5)
  MapGetX
  MapGetY
  - map from world space to closest map coordinates, point-sample-wise
  - use GL_CLAMP_TO_EDGE, so always valid to look up in map
  - identity:
    m->Data[MapGetY(m,y)*m->Width+MapGetX(m,x)] == MapUnmapValue(m,MapGetPointSample(m,x,y))
  MapMapValue
  - c is uint8 from map m, v is resulting float:
    v = Min + c/255. * (Max-Min)
  MapUnmapValue
  - c is uint8, v is float:
    c = 255*(v - Min)/(Max - Min)
*/
void MapSetRange( value_map *Map, float Min, float Max );
float MapGetBilerpSample( value_map *Map, float x, float y );
float MapGetPointSample( value_map *Map, float x, float y );
float MapSetPointSample( value_map *Map, float x, float y, float v );
uint MapGetX( value_map *Map, float x );
uint MapGetY( value_map *Map, float y );
uint8 MapUnmapValue( value_map *Map, float v );
float MapMapValue( value_map *Map, uint8 c );

/*----------------------------------------------------------------------------

  MapAllocateGuyDensity
  - distributes TotalGuys guys over map, allocating them at StartIndex, using Map's values as the density
  - exactly 0 means no guys will ever appear there, everything else is probability density based
  - SlowButAwesome means a true random distribution
  - !SlowButAwesome means a dithered distribution that bands in low density areas
  - RandomizeFrameOffsets makes the not be in lock step
  - @todo need Poisson distribution like Siggraph 01 dithering paper
*/
void MapAllocateGuyDensity( value_map *Map, struct Guy *Template, uint StartIndex, uint TotalGuys, bool RandomizeFrameOffsets = true, bool SlowButAwesome = false );

/*----------------------------------------------------------------------------

  MapAllocateBuildings
  - use a map to allocate axis-aligned rectangular buildings and set their heights and colors
  - CodeBased means every distinct value in the Positions map creates a new building
  - !CodeBased means every contiguous non-0 block creates a new building
  - by default, !CodeBased so that anti-aliasing in Positions doesn't create zillions of buildings
  - if you want abutting different-coded buildings, then set CodeBased and do the work to make solid colors
    in art tool
  - no color maps means random colors
  - easiest to just do MapAllocateBuildings(map,map) if you don't care about the height, or
    want the code to be the height
*/
void MapAllocateBuildings( value_map *Positions, value_map *Heights,
                           value_map *Red = 0, value_map *Green = 0, value_map *Blue = 0,
                           bool CodeBased = false );

/*----------------------------------------------------------------------------

  MapSetHeightField
  - exact HF size will exactly point sample the pixel values
  - else does bilerp
*/
void MapSetHeightField( value_map *Map );

/*----------------------------------------------------------------------------
  MapSetTerrainTexture
  - does what it says, loads the terrain texture, by default creating a new one

  MapSetTerrainTextureChannels
  - mostly for debugging maps used for game design stuff
  - individual channels are optional, pass 0
  - ResizeTexture loses original channels and sets them to black
  - defaults to just updating a channel in the current texture
  - defaults to not building mipmaps and just uploading to the biggest mip level
    which is fine for most debugging purposes, and it's much faster
  - if the map size matches the texture size it does a fast update
*/
void MapSetTerrainTexture( char const *FileName, bool ResizeTexture = true );
void MapSetTerrainTextureChannels( value_map *Red, value_map *Green, value_map *Blue, bool ResizeTexture = false, bool MipMap = false );

#define VALUE_MAP_H
#endif
